--- main data container
-- 
-- @file    DataModel.lua 
-- @author  Pavel Yadvichuk
-- @created 2015/04/08

local bticino_pb = require"bticino_pb"

--- @module DataModel
local DataModel = {} 

----------------------------------------------------------------
-- Module constants
----------------------------------------------------------------
DataModel.BT     = bticino_pb.BT_Message()
DataModel.XOPEN  = bticino_pb.X_Message()
DataModel.PBX    = bticino_pb.P_Message()
DataModel.FIAS   = bticino_pb.F_Message()
DataModel.SYS    = bticino_pb.M_Message()

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local BT         = DataModel.BT
local NAMES = {
		   			[BT.BT_XOPEN]   = "XOPEN",
		  	   		[BT.BT_FIAS]    = "FIAS",
		  	   		[BT.BT_WPBX]    = "PBX",
		  	   		[BT.BT_SYSTEM]  = "SYS"
		      }
local IDENTS = {}

-- print protobuf data set as debug messages
local function __printModel(model)
	if model and "table" == type(model) then
		local tag = ""
		if model.tag and "" ~= model.tag then
			tag = string.format("%s ==> \n", model.tag)
		end
		
		printdbg("%s%s", tag, tostring(model.inner))
	end
end

--- general comparator
local function __checkIdent(buf,ident)
	return buf.ident == ident
end

--- Converting BT Message type to string value
local function __type2string(m)
	if not m or not m.ident then
		return ""
	end
	return NAMES[m.ident] or ""
end

--- Itarate on items and pushing them to table
-- @param data	 A source protobuf data set
-- @param params A parameters for data collection: 
-- 				 false  - collect items to name=value:  evt_id: 451 =>> t[evt_id] = 451
-- 				 table	- collect items by table elements: {["name"] = "value"} =>> t[name] = value
-- 				 string - collect items in some way:
-- 				 		table - collect every item like separate table
local function __collect_repeated_items(data, params)
	if "table" ~= type(data) then 
		return false
	end
	
	local rv = {}
    
    if data.ListFields then     -- processing repeated data for MESSAGE type
        local field_descriptor = require("protobuf/descriptor").FieldDescriptor
        for field, value in data:ListFields() do
            if field_descriptor.LABEL_REPEATED == field.label then
                for idx, item in ipairs(value) do
                    if "table" == type(params) then
                        for i, v in pairs(params) do
                            if item[i] then
                                rv[item[i]] = item[v]
                            end
                        end
                    elseif "table" == params then
                        local iv = {}
                        for f, v in item:ListFields() do
                            if f.name and "allocated" ~= f.name then
                                iv[f.name] = v
                            end
                        end
                        table.insert(rv, iv)
                    else
                        for f, v in item:ListFields() do
                            if f.name and "allocated" ~= f.name then
                                rv[f.name] = v
                            end
                        end
                    end 
                end
            end
        end
    else    -- processing repeated data for simple types
        for key, value in ipairs(data) do
            rv[key]   = value 
            rv[value] = key
        end
    end
	
	return rv
end

--- Load data from string
local function __loadFromString(s)
	local data = nil
	local bt   = bticino_pb.BT_Message()

	-- deserialize data set
	bt:ParseFromString(s)
	if not bt or "table" ~= type(bt) then
		return nil
	end

	-- pre-parse incoming message
	if __checkIdent(bt,BT.BT_XOPEN) then
		data = bt.packedData and bticino_pb.X_Message() or bt.xopenMessage 
	elseif __checkIdent(bt,BT.BT_FIAS) then
		data = bt.packedData and bticino_pb.F_Message() or bt.fiasMessage
	elseif __checkIdent(bt,BT.BT_WPBX) then
		data = bt.packedData and bticino_pb.P_Message() or bt.wpbxMessage
	elseif __checkIdent(bt,BT.BT_SYSTEM) then
		data = bt.sysMessage
	else
		return nil
	end

	if bt.packedData and "" ~= bt.packedData then
		data:ParseFromString(bt.packedData)
	end
	
	return bt, data, "Incoming " .. __type2string(bt) .. " MSG"
end

--- Special handler for initializing of XOpen 
--         common data and endpoint value
local function __initXopenMessage(obj,data)
	if not obj then 
		return nil
	end
	
	obj.endpoint.allocated = 0
	obj.data.allocated = 0

	-- init it with defaults
	obj.endpoint.fcode = 0
	obj.endpoint.ucode = 0

	-- load from user defined config
	if data and "table" == type(data) then
		for i,k in pairs(data) do
			if i == "f" then
				obj.endpoint.fcode = k
			elseif i == "u" then
				obj.endpoint.ucode = k
			elseif i == "a" then
				obj.endpoint.address = k
			end			
		end
	end
	return obj
end

--- Special handler for initializing of
--  Fias common data and options
local function __initFiasMessage(obj,data)
	if not obj then 
		return nil
	end 
	
	obj.opts.allocated = 0
	obj.data.allocated = 0

	-- init it with defaults
	obj.opts.interface = 6
	obj.opts.delimiter = "|"

	-- load from user defined config
	if data and "table" == type(data) then
		for i,k in pairs(data) do
			if i == "ift" then
				obj.opts.interface = k
			elseif i == "delim" then
				obj.opts.delimiter = k
			end			
		end
	end
	return obj
end

--- Special handler for initializing of
--  Pbx common data
local function __initPbxMessage(obj, data)
	if not obj then 
		return nil
	end 
	
	obj.opts.allocated = 0
	obj.ident          = data
	
	return obj
end

--- Special handler for initializing of
--  system common data
local function __initSysMessage(obj, data)
	if not obj then 
		return nil
	end 
	
	obj.allocated = 0
	obj.ident     = data
	
	return obj
end

--- Load data from string
local function __initFromOptions(opts)
	if not opts["ident"] then
		return nil
	end

	local rv = bticino_pb.BT_Message()
		  rv.allocated = 0
		  rv.ident 	   = IDENTS[opts["ident"]]

	if not rv.ident then
		return nil
	end
	
	local inner = nil
	if __checkIdent(rv, BT.BT_XOPEN) then
		inner = __initXopenMessage(rv.xopenMessage, opts["endpoint"])
	elseif __checkIdent(rv, BT.BT_FIAS) then
		inner = __initFiasMessage(rv.fiasMessage, opts["options"])
	elseif __checkIdent(rv, BT.BT_WPBX) then
		inner = __initPbxMessage(rv.wpbxMessage, opts["ct"])
	elseif __checkIdent(rv, BT.BT_SYSTEM) then
		inner = __initSysMessage(rv.sysMessage, opts["ct"])
	end
	inner.allocated = 0

	return rv, inner, "Outgoing " .. __type2string(rv) .. " MSG"
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local model = {
	bt 	  = {},	-- processed BT message object
	inner = {},	-- concrete message data
	tag   = ""	-- tag of the message data
}

--- Serialization function
-- @return byte array for passing to native code
function model:serialize()
	return self.bt:SerializeToString()
end

--- Simple getter for message type 
function model:getType()
	return self.bt.ident
end

--- Helper for debug messages
function model:getTypeString()
	return __type2string(self.bt)
end

--- Simple checker
function model:isSystem()
	return __checkIdent(self.bt, BT.BT_SYSTEM)
end

--- Simple checker
function model:isXopen()
	return __checkIdent(self.bt, BT.BT_XOPEN)
end

--- Simple checker
function model:isFias()
	return __checkIdent(self.bt, BT.BT_FIAS)
end

--- Simple checker
function model:isPbx()
	return __checkIdent(self.bt, BT.BT_WPBX)
end

--- Simple checker
function model:isPeriodic()
	if __checkIdent(self.bt, BT.BT_SYSTEM) then
		return self.inner and DataModel.SYS.M_PERIODIC == self.inner.ident
	end
	return false
end

--- Simple getter for concrete message data
function model:getInner()
	return self.inner
end

--- Extract message field by it's name
-- @param p		incomming pattern
-- @return reference to requested field
-- @note format of incomming pattern is dot separated fields list
-- @example  "field1.field2.field3.wantedField"
function model:getItem(p)
	local node = self.inner
	if not p or "string" ~= type(p) then
		return node
	end
    
	for k in string.gmatch(p,"([^%.]+)") do
		node = node[k]
		if not node then
			break
		end
        
        -- Checking MESSAGE item type
        if "table" == type(node) and
          node.IsInitialized and 
          not node:IsInitialized() then
			node = nil
			break
		end
	end
	return node
end

--- Set value of corresponded field
-- @param p		incomming pattern
-- @param v		value to set to
-- @note format of incomming pattern is dot separated fields list
-- @example  "field1.field2.field3.wantedField"
function model:setItem(p, v)
	if not p or "string" ~= type(p) then
		return false
	end
	local node = self.inner
	
	for k in string.gmatch(p,"([^%.]+)") do
		if "table" ~= type(node[k]) then
			if v then
				node[k] = v
			end
			break
		else
			node = node[k]
			if "table" == type(node) then
				node.allocated = 0
			end
		end
	end
	return node
end

--- Add item to array and return reference to it 
-- @param p		incomming pattern
-- @return reference to requested field
-- @note format of incomming pattern is dot separated fields list
-- @example  "field1.field2.field3.wantedField"
function model:addRepeatedItem(p)
	local ap, ri = p:match("(.*)%.(%w+)$")
	local obj = self:setItem(ap)
	if not obj then
		return nil
	end
	
	obj = obj[ri]
	if not obj or "table" ~= type(obj) then
		return nil
	end
	
	local rv = obj:add()
		  rv.allocated = 0;
	return rv
end

--- Add repeated tuple
-- @param p			An incoming pattern
-- @param key 		A tuple key value
-- @param value		A tuple val value
-- @param fields	An used Tuple fields name, by default "[name]=>[value]"
-- @return reference to created tuple
-- @example "field1.field2.field3.wantedField"
-- @note Internal structure will be like following
-- 		 item = {item.name  = key, item.value = val}
function model:addRepeatedTuple(p, key, val, fields)
	if not key or "" == key then
		return nil
	end
	if not val or "" == val then
		return nil
	end
	
	local item = self:addRepeatedItem(p)
	if not item then
		return nil
	end
	
   	fields = "table" == type(fields) and fields or {["name"]="value"}
   	for k, v in pairs(fields) do
   		item[k] = tostring(key)
   		item[v] = tostring(val)
   		break
   	end
    	
	return item
end

--- Extract repeated field from item with last one
-- @param p		incomming pattern
-- @return requested field converted to new table of "[name]=>[value]" format 
-- @note format of incomming pattern is dot separated fields list
-- @example  "field1.field2.field3" where "wantedField" is repeated field
function model:getRepeatedItem(p)
	local obj = self:getItem(p)
	local params = not self:isXopen() and {["name"]="value"}
	
	return self:getRepeatedItemByInternalObj(obj, params)
end

--- Extract repeated field from itenal object
-- @param obj		internal object
-- @param params	a parameter for data collection
-- @return requested object contend converted to table
function model:getRepeatedItemByInternalObj(obj, params)
	if not obj or "table" ~= type(obj) then
		return nil
	end
	return __collect_repeated_items(obj, params)
end

--- Type identification for this module
function model:getIdentification()
	return "DataModel"
end

--- Constructor with parameters
-- @param opts			pattern for message creation
-- @note in case ("string"==type(opts)) returning object will be
--       deserialized from this string and internal variables
--		 will be protected from changing
-- @note in case ("table"==type(opts)) new empty object will be created
--       opts structure:
--       {ident="XOPEN",endpoint={f=<FCODE>,u=<UCODE>,a=<optional_ADDRESS>}}	<<< for xopen messages
--       {ident="FIAS",options={ift=<IF_TYPE>,delid=<DELIMITER>}}				<<< for fias messages
--       {ident="PBX",ct=<CONTENT_TYPE>}										<<< for pbx messages
--       {ident="SYS",ct=<CONTENT_TYPE>}										<<< for sys messages
function DataModel:new(opts)
	if not opts then 
		return nil, "Error data model options format"
	end

	-- prepare ident table
    for i, v in ipairs(NAMES) do
      IDENTS[v] = i
    end

	-- object oriented magic
	local obj = {}
	setmetatable(obj, {__index = model, __tostring = __printModel})

	-- initialize object
	if "string" == type(opts) then
		obj.bt, obj.inner, obj.tag = __loadFromString(opts)	-- from deserialized data 
	elseif "table" == type(opts) then
		obj.bt, obj.inner, obj.tag = __initFromOptions(opts)	-- with empty structure of corresponded type
	end
	local rv = obj.bt and "table" == type(obj.bt) and obj
	
	return rv, not rv and "Unable to allocate data packet"
end

return DataModel

