--- implementation of data synchronization scenario
-- 
-- @file	scrDataSync.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/19

--- @module scrDataSync
local scrDataSync = {}

----------------------------------------------------------------
-- implementation details
----------------------------------------------------------------
local SYS   = require"DataModel".SYS
local XOPEN = require"DataModel".XOPEN

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return true in case "p" is valid exemplar of "DataModel", false otherwise 
local function __validatePacket(p)
	if not p or "table" ~= type(p) then
		return false
	end
	if not p.getIdentification or "DataModel" ~= p:getIdentification() then
		return false
	end
	return true
end

--- Internal helper for checking type and content of incoming system packet
-- @param p			incoming data packet
-- @return true in case "p" is SystemMessage with type "M_PERIODIC"
local function __checkDataSyncPacket(p)
	local rv = nil
		
	if p:isFias() then
		local fiasType = p:getItem("data.ident")
		rv = 7 == fiasType or 8 == fiasType
	elseif p:isXopen() then
	   local xopenType = p:getItem("data.ident")
	   rv = XOPEN.X_EV_RET_BADGES == xopenType
	elseif p:isSystem() then
		local sysType = p:getItem("ident")
		rv = SYS.M_PERIODIC == sysType or SYS.M_DATABASE_SYNC == sysType
	end 
	return rv
end

----------------------------------------------------------------
-- public interface
----------------------------------------------------------------
local scenario = require"IScenario":new()        -- meta class

--- @see description in IScenario 
function scenario:initialize()
	self.controller	= require"DataSyncController":getInstance()
end

--- @see description in IScenario
function scenario:exec(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	
	if pkt:isFias() then
		local syncRequest = "" 
		local fiasType    = pkt:getItem("data.ident")
		if 7 == fiasType then
			syncRequest = "DS"
		elseif 8 == fiasType then
			syncRequest = "DE"
		end
		printdbg("Sync request for guest data with flag %q", syncRequest)
		self.controller.updSync(syncRequest) 
	elseif pkt:isXopen() then
		local xopenType = pkt:getItem("data.ident")
		if XOPEN.X_EV_RET_BADGES == xopenType then
			printdbg("Sync request for badges data ...")
			self.controller.updSync("RetBadges")
		end
	elseif pkt:isSystem() then
		local sysType = pkt:getItem("ident")
		if SYS.M_HOTEL_INFO == sysType then
			self.controller.initSync()			
		elseif SYS.M_PERIODIC == sysType then
			self.controller.runSync()
		elseif SYS.M_DATABASE_SYNC == sysType then
			printdbg("Sync request for badges data from system ...")
			if not self.controller.badgesExecSync() then
			    print("Failed/blocked by configuration!")
			end
		end
	end
end

--- Analize incoming packet and decide "whether 
--  it can be processed by current module"
-- @param pkt		incoming packet for analizing
-- @return boolean operation status
function scrDataSync.staticCheck(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	if __checkDataSyncPacket(pkt) then
		return true
	end
	return false
end


--- Constructor
function scrDataSync:new(name, storage)
	local obj = {}
	setmetatable(obj, {__index = scenario})
	obj.name    = name
	obj.storage = storage
	return obj
end

return scrDataSync