--- implementation of posting (PS) scenario
-- 
-- @file	scrPostingSimple.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/19

--- @module scrPostingSimple
local scrPostingSimple = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local SYS  = require"DataModel".SYS
local PBX  = require"DataModel".PBX

--- Checking FIAS interface
-- @return boolean value of the PBX iface state
local function __is_onlineFias()
	local modController = require"ModStatusController":getInstance()
	return "online" == modController.getStatus("FIAS")
end

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return true in case "p" is valid exemplar of "DataModel", false otherwise 
local function __validatePacket(p)
	if not p or "table" ~= type(p) then
		return false
	end
	if not p.getIdentification or "DataModel" ~= p:getIdentification() then
		return false
	end
	return true
end

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return TRUE in case "p":
-- 		   is FiasMessage with type 33 (Posting Answer (WR)) 
-- 		   is SystemMessage with type "M_PERIODIC"
-- 		   is PbxMessage with type "P_EVENT..516" (CDR Event) 
local function __checkPSPacket(p)
	local rv = nil
	
	if p:isFias() then	
		local fiasType = p:getItem("data.ident")
		rv = fiasType and 33 == fiasType 		-- FIAS PA message
	elseif p:isPbx() then
		local pbxType = p:getItem("ident")
		local subType = p:getItem("data.sub")
		rv = pbxType and subType and PBX.P_EVENT == pbxType and 516 == subType
	else
		rv = p:isPeriodic()
	end
	
	return rv
end

--- Parse PBX packet to internal data
local function __parsePSData(pkt)
	if not pkt then
		return false
	end
	
	local data 		= nil
	local CDRFields = pkt:getRepeatedItem("data")
	if CDRFields then
		local userField = CDRFields.UserField or ""
		if userField:find("WHOTEL%-MINIBAR") then
			local article, items = userField:match("WHOTEL%-MINIBAR:(%d+)#(%d+)")
			if article then
				data = {
					 type	 	= 2
					,startv  	= CDRFields.StartTime
					,roomIdent 	= CDRFields.Source
					,article 	= article
					,items 	 	= items
				}
			end
		elseif userField:find("WHOTEL%-CALL") then
			data = {
				 type 		= 1
				,startv 	= CDRFields.StartTime
				,roomIdent 	= CDRFields.Source
				,dst		= CDRFields.Destination
				,duration	= CDRFields.BillableSeconds
				,cost		= CDRFields["X-Wildix-Cost"]
			}
		end
	end
	
	return data
end

--- Parse FIAS packet to internal data
local function __parsePAData(pkt)
	if not pkt then
		return false
	end
	
	local dataItems = pkt:getRepeatedItem("data")
	local data 	= {
		 roomIdent	= dataItems["RN"]
		,id			= dataItems["P#"]
		,state		= dataItems["AS"]
	}
	
	return data
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local scenario = require"IScenario":new()        -- meta class

--- @see description in IScenario 
function scenario:initialize()
	self.controller = require"PostingSimpleController":getInstance()
	self.queue      = require"PostingQueue":new()
end

--- @see description in IScenario
function scenario:exec(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	
	if pkt:isPbx() then
		local data = self.controller.processPSData(__parsePSData(pkt), SYS) 
		if data then
			local rv, err = self.queue:push(data)
    		if not rv and err then
    			print("Error with pushing PS data to queue: %s", err)
    		else
    			printdbg("PS data was pushed to queue!")
    		end
		end
	elseif pkt:isFias() then
		local data = __parsePAData(pkt)
		self.controller.markPS(data)
	elseif pkt:isPeriodic() then
		if __is_onlineFias() then
    		local cacheState, cacheData = self.controller.updatePSCounter()
    		if cacheState then
        		local stateCallbacks = {}
        		stateCallbacks.NOTFOUND = function()
    				local psData = self.queue:top()
        			if psData then
        				printdbg("Get top element from the queue for new sending")
        				self.controller.createPS(psData)
        			end	
    			end
        		stateCallbacks.RESEND = function()
       				self.controller.markPS(cacheData)
        			local psData = self.queue:top()
        			if psData then
        				printdbg("Get top element from the queue for re-sending")
        				self.controller.createPS(psData)
        			end
    			end
    			stateCallbacks.TIMEOUT = function()
    				local queueSize = self.queue:size()
        			if 1000 <= queueSize then	
        				-- Queue size when we will make pop by TIMEOUT
        				self.controller.markPS(cacheData)
        			else
        				-- try resend Post Simple message
        				cacheData.state = "RESEND"
        				stateCallbacks[cacheData.state]()
        			end
        		end
        		
        		stateCallbacks[cacheState]()
        	end
        	if self.controller.clearMarkedPS() then
				printdbg("Pop top element from the queue")
				self.queue:pop()
			end
    	else
    		self.controller.clearAllPS()
    	end
	end
end

--- Analize incoming packet and decide "whether 
--  it can be processed by current module"
-- @param pkt		incoming packet for analizing
-- @return boolean operation status
function scrPostingSimple.staticCheck(pkt)
		if not __validatePacket(pkt) then
		return false
	end
	if __checkPSPacket(pkt) then
		return true
	end
	return false 
end


--- Constructor
function scrPostingSimple:new(name, storage)
	local obj = {}
	setmetatable(obj, {__index = scenario})
	obj.name    = name
	obj.storage = storage
	return obj
end

return scrPostingSimple
