--- implementation of Room Equipment (DND/MUR) scenario
-- 
-- @file	scrRoomEquipment.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/19

--- @module scrRoomEquipment
local scrRoomEquipment = {}

----------------------------------------------------------------
-- implementation details
----------------------------------------------------------------
local PBX   = require("DataModel").PBX
local XOPEN = require("DataModel").XOPEN
local SYS   = require("DataModel").SYS

--- Checking interface status
-- @return iface status value
local function __get_ifaceStatus(ifaceState)
	local modController = require"ModStatusController":getInstance()
	return modController.getStatus(ifaceState)
end

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return true in case "p" is valid exemplar of "DataModel", false otherwise 
local function __validatePacket(p)
	if not p or "table" ~= type(p) then
		return false
	end
	if not p.getIdentification or "DataModel" ~= p:getIdentification() then
		return false
	end
	return true
end

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return TRUE in case "p" :
-- 			is FiasMessage with ident Room equipment (RE)
-- 			is XopenMessage with Evt Room Info
-- 			is PbxMessage with CDR field marked as RE
local function __checkRoomEquipmentPacket(p)
	local rv = nil
	
	if p:isFias() then	
		local fiasType = p:getItem("data.ident")
		rv = fiasType and 25 == fiasType 	-- FIAS requrest with RE
	elseif p:isXopen() then
		local xopenType = p:getItem("data.ident")
		rv = xopenType and XOPEN.X_HSE_ROOM_DTLS_DATA == xopenType
	elseif p:isPbx() then
		local pbxType = p:getItem("ident")
		local subType = p:getItem("data.sub")
		rv = pbxType and subType and PBX.P_EVENT == pbxType and (516 == subType or 519 == subType)	-- CDR or feature event
	end
	
	return rv
end

--- Deserialize FIAS packet data to internal data
local function __parseRoomEquipmentData(pkt)
	if not pkt then
		return false
	end
	
	local dataIdent = pkt:getItem("data.ident")
	local dataItems = pkt:getRepeatedItem("data")
	local data = {
		 ifn		= "FIAS"
		,roomIdent 	= dataItems["RN"]
		,ACL		= dataItems["CS"]
		,DND 		= dataItems["DN"]
	}
	return data, dataIdent
end

--- Deserialize XOPEN packet data to internal data
local function __parseRoomEventDetails(pkt)
	if not pkt then
		return false
	end
	
	local dataIdent = pkt:getItem("data.ident")
	local roomDetails = pkt:getRepeatedItem("data.hseRoomDtls")
	local data = {
		 ifn		= "XOPEN"
		,roomIdent 	= roomDetails.aptid
		,MUR		= roomDetails.mur
		,DND 		= roomDetails.dnd	
	}
	return data, dataIdent
end

--- Deserialize PBX packet data to internal data
local function __parsePBXData(pkt)
	if not pkt then
		return false
	end
	
	local eqName    = nil 	-- can be DND, MUR
	local eqVal     = nil	-- can be 1 or 0
	local roomExten = nil	-- room extension
	local eventFields = pkt:getRepeatedItem("data")
	if eventFields then
		local userField = eventFields.UserField 
		if userField then
    		if userField:find("WHOTEL%-MUR") then
    			eqName, eqVal = userField:match("WHOTEL%-(%w+):(%-?%d)")
    			roomExten 	  = eventFields.Source
    		end
    	elseif "DND" == eventFields.Feature then
    		eqName 	  = eventFields.Feature
    		eqVal  	  = "true" == eventFields.State and "1" or "0"
    		roomExten = eventFields.Device
    	end
	end
	
	local data = nil
	if eqName then
		data = {
			 ifn		= "PBX"
			,roomIdent  = roomExten
			,MUR		= "MUR" == eqName and eqVal
			,DND		= "DND" == eqName and eqVal
		}
	end
	return data, eqName
end

----------------------------------------------------------------
-- public interface
----------------------------------------------------------------
local scenario = require"IScenario":new()        -- meta class

--- @see description in IScenario 
function scenario:initialize()
	self.controller = require("RoomEquipmentController"):getInstance()
end

--- @see description in IScenario
function scenario:exec(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	if "online" ~= __get_ifaceStatus("PBX") then
		print("Scenario %q was stopped. PBX interface is not online", self.name)
		return false
	end
	
	local xopenStatus = __get_ifaceStatus("XOPEN")
	if pkt:isFias() then
		local data = __parseRoomEquipmentData(pkt)
		print("Processing room equipment request (RE) ...")
		self.controller.setEquipment(data, SYS, "notInUse" ~= xopenStatus)
	elseif pkt:isXopen() then
		local data = __parseRoomEventDetails(pkt)
   		print("Processing event room info ...")
   		self.controller.setEquipment(data, SYS, "notInUse" ~= xopenStatus)
	elseif pkt:isPbx() then
		local data, eqName = __parsePBXData(pkt)
		if eqName then
			print("Processing PBX data (%s) ...", eqName)
			self.controller.setEquipment(data, SYS, "notInUse" ~= xopenStatus)			
		end
	end
end

--- Analize incoming packet and decide "whether 
--  it can be processed by current module"
-- @param pkt		incoming packet for analizing
-- @return boolean operation status
function scrRoomEquipment.staticCheck(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	if __checkRoomEquipmentPacket(pkt) then
		return true
	end
	return false
end


--- Constructor
function scrRoomEquipment:new(name, storage)
	local obj = {}
	setmetatable(obj, {__index = scenario})
	obj.name    = name
	obj.storage = storage
	return obj
end

return scrRoomEquipment

