--- implementation of RoomState (GI/GO/GC/KR/KD) scenario
-- 
-- @file	scrRoomState.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/19

--- @module scrRoomState
local scrRoomState = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local XOPEN = require"DataModel".XOPEN

--- Checking interface status
-- @return iface status value
local function __get_ifaceStatus(ifaceState, subIface)
	local modController = require"ModStatusController":getInstance()
	if subIface then
	    local subIfacesData = modController.getData(ifaceState, subIface)
	    return modController.getData(ifaceState, subIface) and modController.getStatus(ifaceState) or modController.DEFLT_STATUS
    else    
	    return modController.getStatus(ifaceState)
   end
end

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return true in case "p" is valid exemplar of "DataModel", false otherwise 
local function __validatePacket(p)
	if not p or "table" ~= type(p) then
		return false
	end
	if not p.getIdentification or "DataModel" ~= p:getIdentification() then
		return false
	end
	return true
end

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return TRUE in case "p":
-- 			is FiasMessage with ident GI/GO/GC and KR/KD
-- 			is XopenMessage with ident Evt Badges Created or Evt Badges Deleted
local function __checkGuestPacket(p)
	local rv = nil
	
	if p:isFias() then	
		local fiasType = p:getItem("data.ident")
		rv = fiasType and (11 == fiasType or 12 == fiasType or 13 == fiasType)			-- FIAS requrests for GI/GO/GC
		rv = rv or (fiasType and (34 == fiasType or 35 == fiasType))	                -- FIAS requrests for KR/KD
	elseif p:isXopen() then
		local xopenType = p:getItem("data.ident")
		rv = xopenType and (XOPEN.X_EV_BADGES_CREATED == xopenType or XOPEN.X_EV_BADGES_DELETED == xopenType)	-- XOPEN EvtBadgeCreate/EvtBadgeDelete
		rv = rv or (xopenType and XOPEN.X_EV_RET_BADGES == xopenType)											-- XOPEN RetBadges
	else
		rv = p:isPeriodic()
	end
	
	return rv
end

--- Parse FIAS packet to initernal data
local function __parseGuestData(pkt)
	if not pkt then
		return false
	end
	
	local function __getStringValue(val)
	   return "" ~= val and val
	end
	
	local dataIdent = pkt:getItem("data.ident")
	local dataItems = pkt:getRepeatedItem("data")
	
	local data = {
		-- common fields
		 ifn		= "FIAS"
		,rn 		= dataItems["RN"]
		,ro			= dataItems["RO"]
		,arrDate	= __getStringValue(dataItems["GA"])
		,depDate	= __getStringValue(dataItems["GD"])
		,wsId		= __getStringValue(dataItems["WS"])
		,surname	= __getStringValue(dataItems["GN"]) 
		,time		= dataItems["TI"]
		,date       = dataItems["DA"]
		
		-- guest request filelds
		,name		= __getStringValue(dataItems["GF"])
		,userId		= dataItems["ID"]			
		,lang		= __getStringValue(dataItems["GL"])
		,gtype  	= dataItems["GV"]
		,cos		= dataItems["CS"]
		,swap		= dataItems["SF"]
		
		-- key request fields
		,kCoder		= dataItems["KC"]
		,kType 		= dataItems["KT"]
		,depTime	= dataItems["DT"]
		,kOpt		= dataItems["KO"]
		,sInfo		= dataItems["SI"]
	}

	return data, dataIdent
end

--- Parse XOPEN packet to internal data
local function __parseBadgeData(pkt)
	if not pkt then
		return false
	end
	
	local dataIdent = pkt:getItem("data.ident")
	local data 		= nil
	
	if XOPEN.X_EV_BADGES_CREATED == dataIdent then
		local accessItems = pkt:getRepeatedItem("data.badgesData")
		data   = {
			 serial  	 = accessItems.badge.sn 
			,startv  	 = accessItems.badge.start_validity
			,endv	 	 = accessItems.badge.end_validity
			,name    	 = accessItems.badge.name
			,surname 	 = accessItems.badge.surname
			,type	 	 = accessItems.badge.type
			,apts	 	 = pkt:getRepeatedItemByInternalObj(accessItems.apts, "table")
			,workstation = {fc = pkt:getItem("endpoint.fcode") ,uc = pkt:getItem("endpoint.ucode")}
		}
	elseif XOPEN.X_EV_BADGES_DELETED == dataIdent then
		local bObj  = pkt:getItem("data.evtDelBadges.badges")
        data = {
        	badges = pkt:getRepeatedItemByInternalObj(bObj, "table")
        }
    elseif XOPEN.X_EV_RET_BADGES == dataIdent then
    	local aObj   	  = pkt:getItem("data.badgesData")
    	local accessItems = pkt:getRepeatedItemByInternalObj(aObj, "table")
    	if 0 ~= #accessItems then
    		data = {}
        	for idx, item in ipairs(accessItems) do
        		local badgeData = {
        			 serial  = item.badge.sn 
        			,startv  = item.badge.start_validity
        			,endv	 = item.badge.end_validity
        			,name    = item.badge.name
        			,surname = item.badge.surname
        			,type	 = item.badge.type
        			,apts	 = pkt:getRepeatedItemByInternalObj(item.apts, "table")
        		}
        		data[#data + 1] = badgeData
        	end
        end
	end
	
	return data, dataIdent
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local scenario = require"IScenario":new()        -- meta class

--- @see description in IScenario 
function scenario:initialize()
	self.controller = require"GuestDataController":getInstance()
	self.queue      = require"GuestDataQueue":new()
end

--- @see description in IScenario
function scenario:exec(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	if not pkt:isPeriodic() and "online" ~=  __get_ifaceStatus("PBX") then
		print("Scenario %q was stopped. PBX interface is not online", self.name)
		return false
	end

	local rv = nil
	if pkt:isFias() or pkt:isXopen() then
		local rv, err = self.queue:push(pkt)
		if rv then
			printdbg("Key/Guest data was pushed in queue with id %d", rv)
		else
			print("Error with pushing Key data to queue: %s", err or "unknown error")
		end
	elseif pkt:isPeriodic() then
		-- processing guest data queue
		if self.queue:isLocked() then
			-- avoid processing queue by different threads
			return rv
		end
		while not self.queue:isEmpty() do
			-- NOTE: do not use return inside while block with unlock method !!!
			if not self.queue:tryTimedLock(10) then  -- lock timeout is 10 seconds
				break
			end
			local gdQueueId, gdPkt = self.queue:pop()
			if not gdQueueId then
				break
			end
			print("Processing guest data with queue id %d", gdQueueId)

			local rData  = nil
			local rBadge = nil
			if gdPkt:isFias() then
				local data, dataIdent = __parseGuestData(gdPkt)
				if 11 == dataIdent then
					print("Processing guest check-in request (GI) ...")
					self.controller.guestCheckIn(data, "notInUse" ~= __get_ifaceStatus("XOPEN"))
				elseif 12 == dataIdent then
					print("Processing guest check-out request (GO) ...")
					self.controller.guestCheckOut(data, "notInUse" ~= __get_ifaceStatus("XOPEN"))
				elseif 13 == dataIdent then
					print("Processing guest data change request (GC) ...")
					self.controller.guestDataChange(data, "notInUse" ~= __get_ifaceStatus("XOPEN"))
				elseif 34 == dataIdent then
					print("Processing key request (KR) ...")
					rv, rData = self.controller.keyRequest(data, "notInUse" ~= __get_ifaceStatus("FIAS", "PB"))
				elseif 35 == dataIdent then
					print("Processing key delete (KD) ...")
					rv, rData = self.controller.keyDelete(data, "notInUse" ~= __get_ifaceStatus("FIAS", "PB"))
				end

			elseif gdPkt:isXopen() then
				local data, dataIdent = __parseBadgeData(gdPkt)
				if XOPEN.X_EV_BADGES_CREATED == dataIdent then
					print("Processing event badges created ...")
					rv, rData, rBadge = self.controller.badgesCreated(data, "notInUse" ~= __get_ifaceStatus("FIAS", "PB"))
				elseif XOPEN.X_EV_BADGES_DELETED == dataIdent then
					print("Processing event badges deleted ...")
					rv, rData, rBadge = self.controller.badgesDeleted(data)
				elseif XOPEN.X_EV_RET_BADGES == dataIdent then
					print("Processing event badges with all guest data ...")
					self.controller.badgeListProcessing(data, "notInUse" ~= __get_ifaceStatus("FIAS", "PB"))
					rv = {"dataSync"}
					break
				end
			end

			-- processing returned data
			if rData then
				self.controller.KeyAnswer(rData)
			end
			if rBadge then
				self.controller.badgeResponse(rBadge)
			end
		end
		self.queue:unlock()

		-- checking timeouted key requests
		local keysTimeout = self.controller.checkKeysTimeout()
		for i, keyData in pairs(keysTimeout) do
			self.controller.KeyAnswer(keyData)
		end
	end
	return rv
end

--- Analize incoming packet and decide "whether 
--  it can be processed by current module"
-- @param pkt		incoming packet for analizing
-- @return boolean operation status
function scrRoomState.staticCheck(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	if __checkGuestPacket(pkt) then
		return true
	end
	return false 
end


--- Constructor
function scrRoomState:new(name, storage)
	local obj = {}
	setmetatable(obj, {__index = scenario})
	obj.name    = name
	obj.storage = storage
	return obj
end

return scrRoomState
