--- implementation of starting system up scenario
-- 
-- @file	scrStartUp.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/19

--- @module scrStartUp
local scrStartUp = {}

----------------------------------------------------------------
-- implementation details
----------------------------------------------------------------
local SYS = require("DataModel").SYS

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return true in case "p" is valid exemplar of "DataModel", false otherwise 
local function __validatePacket(p)
	if not p or "table" ~= type(p) then
		return false
	end
	if not p.getIdentification or "DataModel" ~= p:getIdentification() then
		return false
	end
	return true
end


--- Internal helper for checking type and content of incoming system packet
-- @param p			incoming data packet
-- @return true in case "p" is SystemMessage of type "M_HOTEL_INFO", false otherwise 
local function __checkSysPacket(p)
	if not p:isSystem() then
		return false
	end
	local sysType = p:getItem("ident")
	return sysType and SYS.M_HOTEL_INFO == sysType
end

--- Write rooms to the memory
-- @param t     			item type 
-- @param provider			data provider
-- @param tbl				rooms table
-- @param dflt_addr			default XOpen universal address  (address of HSE engine)
-- @return number of rooms added
local function __store_items(t, provider, tbl, dflt_addr)
	local rv 	  	= 0
	local errList 	= {}
	local roomsList = {}
	if "table" == type(tbl) then
		for _, item in pairs(tbl) do
			if item then
				item["FC"] = item["FC"] or dflt_addr["FC"]	-- keep XOpen address actual even if not configured
				item["UC"] = item["UC"] or dflt_addr["UC"]
				local ident  = nil
				local errMsg = nil
				if "room" == t then
					ident, errMsg = provider.pushRoom(item)
					if ident then
						table.insert(roomsList, ident)
						rv = 1 + rv
					end
				elseif "ca" == t then
					ident, errMsg = provider.pushCommArea(item)
					if ident then
						rv = 1 + rv
					end
				end
				table.insert(errList, errMsg)
			end
		end
		provider.saveRoomsList(roomsList)
	end
	return rv, errList
end

--- Save virtual rooms list to the memory
-- @param provider  data provider
-- @param tbl       virtual rooms table
-- @return number of the saved virtual rooms
local function __store_virtualRooms(provider, tbl)
    local rv               = 0
    local virtualRoomsList = {}
    local errMsgList       = {}
    if "table" == type(tbl) then
        for _, item in pairs(tbl) do
            local ident, errMsg = provider.pushVirtualRoom(item)
            if ident then
                table.insert(virtualRoomsList, ident)
                rv = rv + 1
            else
                table.insert(errMsgList, errMsg)
            end
        end
        provider.saveVirtualRoomsList(virtualRoomsList)
    end
    return rv, errMsgList
end

--- Write mapping table to the memory
-- @param provider			data provider
-- @param tbl				rooms table
-- @return number of mappings added
local function __store_mapping(provider, tbl)
	local rv = 0
	if "table" == type(tbl) then
		-- clearing random workstation data
		provider.delRandomWSMapping()
		
		-- loading workstations with set up random one
		for _, item in pairs(tbl) do
			if item then
				local rm = provider.pushMapping(item) 
				if 0 ~= rm then
					rv = rv + rm 
					local rs, msg = provider.pushRandomWSMapping(item)
				end
			end
		end
		
		-- displaying result random workstation data
		local randWS = provider.getRandomWSMapping()
		if randWS and randWS["fc"] and randWS["uc"] then 
			printdbg(string.format("RandomClientWS - %s/%s", randWS["fc"], randWS["uc"]))
		else
			print("RandomClientWS was not set by some errors!")
		end
	end
	return rv
end

--- Sending to manager flag about initialization finish state
local function __set_initialized_flag()
	local flagSrvPkt, err = require"DataModel":new({ident = "SYS", ct = SYS.M_FLAGS_SERVICE})
	if not flagSrvPkt then
		exit_msg(err)
	end
	
	flagSrvPkt:setItem("flags", SYS.M_FLAG_INITIALIZED)
	local rv = require"appHelper":startTransaction(flagSrvPkt)
	if not rv then
		exit_msg("Unable to set initialized flag!")
	end
end

----------------------------------------------------------------
-- public interface
----------------------------------------------------------------
local scenario = require"IScenario":new()        -- meta class

--- @see description in IScenario 
function scenario:initialize()
	self.dataProvider = require("DataProvider"):getInstance()
	self.roomProvider = require("RoomProvider"):getInstance()
end

--- @see description in IScenario
function scenario:exec(pkt)
	if not __validatePacket(pkt) then
		return false
	end

	local initial = pkt:getItem("hotelInfo.isInitial")
	if 1 ~= initial then
		return false
	end
	
	-- save PID of process
	self.dataProvider.setProcPid()

	---------------------------------------------------------
	--- here we absolutelly sure that event "StartUp" come!!
	---------------------------------------------------------
	
	-- save run-time config
	local cfg = pkt:getItem("hotelInfo")
	self.dataProvider.setHotelConfig(cfg)

	-- load hsprj file
	local hsprj, errMsg = require("xmlprj").xmlprj_load(cfg.hotelConfig)
	if not hsprj then
		exit_msg("Unable to parse %q file: %s", cfg.hotelConfig, errMsg)
	end
	local ginfo  = {
		["FC"] = hsprj.configurator.fcode,
		["UC"] = hsprj.configurator.ucode,
	}
	if 1 ~= self.roomProvider.pushGlobalWSMapping(ginfo) then
		print("Error in saving hotel server mapping data!")
	end
	local roomsNbr, errList = __store_items("room", self.roomProvider, hsprj.configurator.rooms, ginfo)
	print("Added %d room(s) to storage", roomsNbr)
	for _, errMsg in pairs(errList) do printdbg(errMsg) end

	local caNbr, errList = __store_items("ca", self.roomProvider, hsprj.configurator.commonareas, ginfo)
	print("Added %d common area(s) to storage", caNbr)
	for _, errMsg in pairs(errList) do printdbg(errMsg) end
	
	-- load virtual rooms list
	local vRoomsCfg, errMsg = require("xmlprj").vrlist_load(cfg.VRoomsConfig)
	if vRoomsCfg then
    	local vrNbr, errMsgList = __store_virtualRooms(self.roomProvider, vRoomsCfg.virtualrooms)
    	print("Added %d virtual room(s) with suites to storage", vrNbr)
    	for _, errMsg in pairs(errMsgList) do printdbg(errMsg) end
	else
	    printdbg("Unable to parse %q file: %s", cfg.VRoomsConfig, errMsg)
	end

	-- load workstation mapping file
	local wsMap, errMsg = require("xmlprj").wsmap_load(cfg.wsMapping)
	if wsMap then
		local wsNbr = __store_mapping(self.roomProvider, wsMap.workstations)
		print("Stored %d workstation(s) to storage", wsNbr)
	else
		printdbg("Unable to parse %q file: %s", cfg.wsMapping, errMsg)
	end
    	
    -- initialize memcache queues
    require"GuestDataQueue":new():init()
    require"WakeUpQueue":new():init()
	
	-- set iface status to default value
	local ifaceController = require("ModStatusController"):getInstance()
	ifaceController.setStatus("PBX", "offline")
	ifaceController.setStatus("FIAS", "offline")
	ifaceController.setStatus("XOPEN", "offline")
	
	-- set flag about manager initialized state
	__set_initialized_flag()
	
	return {"dataSync"}
end

--- Analize incoming packet and decide "whether 
--  it can be processed by current module"
-- @param pkt		incoming packet for analizing
-- @return boolean operation status
function scrStartUp.staticCheck(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	if __checkSysPacket(pkt) then
		return true
	end
	return false 
end


--- Constructor
function scrStartUp:new(name, storage)
	local obj = {}
	setmetatable(obj, {__index = scenario})
	obj.name    = name
	obj.storage = storage
	return obj
end

return scrStartUp
