--- implementation of WakeUp scenario
-- 
-- @file	scrWakeUp.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/19

--- @module scrWakeUp
local scrWakeUp = {}

----------------------------------------------------------------
-- implementation details
----------------------------------------------------------------
local SYS  = require"DataModel".SYS
local PBX  = require"DataModel".PBX

--- Checking PBX interface
-- @return boolean value of the PBX iface state
local function __is_onlinePbx()
	local modController = require"ModStatusController":getInstance()
	return "online" == modController.getStatus("PBX")
end

--- Internal helper for checking type and content of incoming packet
-- @param p			incoming data packet
-- @return true in case "p" is valid exemplar of "DataModel", false otherwise 
local function __validatePacket(p)
	if not p or "table" ~= type(p) then
		return false
	end
	if not p.getIdentification or "DataModel" ~= p:getIdentification() then
		return false
	end
	return true
end

--- Internal helper for checking type and content of incoming system packet
-- @param p			incoming data packet
-- @return true in case "p":
-- 		   is FiasMessage with type 27 (Wakeup request (WR)) or 29 (Wakeup clear (WC)) 
-- 		   is SystemMessage with type "M_PERIODIC"
-- 		   is PbxMessage with type "P_EVENT..516" (CDR Event) 
local function __checkWKPacket(p)
	local rv = false
	if p:isFias() then
		local fiasType = p:getItem("data.ident")
		rv = fiasType and (27 == fiasType or 29 == fiasType)
	elseif p:isPbx() then
		local pbxType = p:getItem("ident")
		local subType = p:getItem("data.sub")
		rv = pbxType and subType and PBX.P_EVENT == pbxType and 516 == subType  
	elseif p:isSystem() then
		local sysType = p:getItem("ident")
		rv = sysType and (SYS.M_PERIODIC == sysType or SYS.M_WAKEUP_SERVICE == sysType)
	end
	
	return rv
end

--- Parse FIAS packet to internal data
local function __parseWKData(pkt)
	if not pkt then
		return false
	end
	
	local ident = pkt:getItem("data.ident")
	local items = pkt:getRepeatedItem("data")
	local data  = {
		 ifn  = "FIAS"
		,treq =	27 == ident and "WR" or "WC"
		,rn   = items["RN"]
		,da   = items["DA"]
		,ti   = items["TI"]
	}
	
	return data
end

--- Parse PBX packet to internal data
local function __parsePBXWKData(pkt)
	if not pkt then
		return false
	end
	
	local data		= nil
	local CDRFields = pkt:getRepeatedItem("data")
	if CDRFields then
		local userField = CDRFields.UserField or ""
		if userField:find("REQUEST") then
			local rn, wkTime, wkDate = userField:match("WHOTEL%-WAKEUP%-REQUEST:(%w+)!(%d+)!(%d+)")
			if rn and wkTime and wkDate then
				data = {
					 treq = "WR"
					,rn   = rn
					,da	  = wkDate
					,ti	  = wkTime
				}
			end
		elseif userField:find("CLEAR") then
			local rn, wkTime, wkDate = userField:match("WHOTEL%-WAKEUP%-CLEAR:(%w+)!(%d+)!(%d+)")
			if rn and wkTime and wkDate then
				data = {
					 treq = "WC"
					,rn   = rn
					,da	  = wkDate
					,ti	  = wkTime
				}
			end
		else
    		local rn, wkId = userField:match("WHOTEL%-WAKEUP:(%w+)!(%w+)")
    		if rn and wkId then
    			data = {
    				 treq = "CDR"
    				,rn   = rn
    				,wkId = wkId
    			}
    		end
    	end
	end
	
	if data then
		data.ifn = "PBX"
	end
	
	return data
end

--- Parse PBX packet to internal data
local function __parseSystemWKResponce(pkt)
	if not pkt then
		return false
	end
	
	local data     = nil
	local action   = pkt:getItem("WakeUpObject.action")
	local tag 	   = pkt:getItem("WakeUpObject.tag") or "" 
	local rn, wkId = tag:match("(%w+)!(%w+)")
	if rn and wkId then
    	data = {
    		 ifn  = "SYS"	-- we push to manager room exten
    		,treq = "SYS"
    		,rn   = rn 
    		,wkId = wkId
    	}
    end
	
	return data, action
end

----------------------------------------------------------------
-- public interface
----------------------------------------------------------------
local scenario = require"IScenario":new()        -- meta class

--- @see description in IScenario 
function scenario:initialize()
	self.controller = require"WakeUpController":getInstance()
	self.queue      = require"WakeUpQueue":new()
end

--- @see description in IScenario
function scenario:exec(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	if not pkt:isPeriodic() and not __is_onlinePbx() then
		print("Scenario %q was stopped. PBX interface is not online", self.name)
		return false
	end
	
	local function __pushDataToQueue(data)
		if "table" == type(data) then
    		local rv, err = self.queue:push(data)
    		if not rv then
    			print("Error with pushing WK data to queue: %s", err or "Unknown error")
    		else
    			printdbg("WakeUp data was pushed in queue with id %d", rv)
    		end
    	end
	end
	
	if pkt:isFias() then -- push request to wakeUp queue
		__pushDataToQueue(__parseWKData(pkt))
	elseif pkt:isPbx() then
		__pushDataToQueue(__parsePBXWKData(pkt))
	elseif pkt:isSystem() then
		local sysType = pkt:getItem("ident")
		if SYS.M_PERIODIC == sysType then
    		while not self.queue:isEmpty() do
    			if self.controller.getSyncFlag() then
    				printdbg("Wait when WakeUps sync procedure will be finished!")
    				break
    			end
    			
    			local rs    = false
    			local rData = nil
    			local wkQueueId, wkData = self.queue:pop()
    			if wkQueueId then
        			printdbg("Processing wakeUp data with queue id %d", wkQueueId)
        			
        			if "WR" == wkData.treq then
        				print("Processing wake up request (WR) ...")
        				rs, rData = self.controller.create(wkData)
        			elseif "WC" == wkData.treq then
        				print("Processing wake up clear request (WC) ...")
        				self.controller.delete(wkData, "WakeUpClear")
        			elseif "CDR" == wkData.treq then
        				print("Processing wake up responce event ...")
                   		rs, rData = self.controller.delete(wkData, "CDREvent")
        			elseif "SYS" == wkData.treq then
        				print("Processing system command for removing wake up by timeout ...")
        				rs, rData = self.controller.delete(wkData, "SYSTimeout")
        			end
        			
        			if rData then
    					self.controller.sendAnswer(rData[1], rData[2])
    				end
    			end
    		end
    	elseif SYS.M_WAKEUP_SERVICE == sysType then
    		local data, dataAction = __parseSystemWKResponce(pkt)
    		if SYS.M_WA_RING == dataAction then
    			self.controller.ring(data)
	    	elseif SYS.M_WA_DEL == dataAction then
	    		__pushDataToQueue(data)
	    	end
    	end
	end
end

--- Analize incoming packet and decide "whether 
--  it can be processed by current module"
-- @param pkt		incoming packet for analizing
-- @return boolean operation status
function scrWakeUp.staticCheck(pkt)
	if not __validatePacket(pkt) then
		return false
	end
	if __checkWKPacket(pkt) then
		return true
	end
	return false 
end


--- Constructor
function scrWakeUp:new(name, storage)
	local obj = {}
	setmetatable(obj, {__index = scenario})
	obj.name    = name
	obj.storage = storage
	return obj
end

return scrWakeUp