--- Main extension file with processing procedure for scenarios 
-- 
-- @file    extensions.lua 
-- @author  Vyacheslav Khudan
-- @created 2014/10/14
-- @note global variable StateUID indicates thread by integer value

local cfgDir = "/etc/wildix/whoteld"
package.path = package.path ..";"..cfgDir.."/?.lua"
package.path = package.path ..";"..cfgDir.."/utils/?.lua"
package.path = package.path ..";"..cfgDir.."/rawDrivers/?.lua"
package.path = package.path ..";"..cfgDir.."/modules/?.lua"
package.path = package.path ..";"..cfgDir.."/storage/?.lua"
package.path = package.path ..";"..cfgDir.."/dataModel/?.lua"
package.path = package.path ..";"..cfgDir.."/eventModel/?.lua"

----------------------------------------------------------------
-- Helper interface
----------------------------------------------------------------
-- destroy Lua state with error message
function exit_msg(fmt, ...)
	if fmt then
		error(string.format(fmt, ...), 2)
	end
	return false
end

----------------------------------------------------------------
-- Implementation details
---------------------------------------------------------------- 
local function __loadHotelInfo()
	local dModel 	   = require"DataModel"
	local hotelInfoPkt = dModel:new({ident = "SYS", ct = dModel.SYS.M_HOTEL_INFO_SERVICE})
	
	local rs, dsID = require"appHelper":startTransaction(hotelInfoPkt, true)
	if not dsID then
		exit_msg("Couldn't get Hotel data!")
	end
	
	local hInfoPkt = dModel:new(bt_data(dsID))
	if not hInfoPkt then
		eixt_msg("Error parsing Hotel info data Set!")
	end
	
	-- create prefix for memc storage
	local family 	 = hInfoPkt:getItem("hotelInfo.astDbFamily") or ""
	local dataSuffix = hInfoPkt:getItem("hotelInfo.dataSuffix") or ""
	
	local hotelInfo = {
		["memc"]  = {
			host   = hInfoPkt:getItem("hotelInfo.MemCacheHost"),
			port   = hInfoPkt:getItem("hotelInfo.MemCachePort"),
			prefix = family.."-"..dataSuffix,
		},
		["ast"]   = {
			prefix = hInfoPkt:getItem("hotelInfo.astDbFamily"),
		},
		["hotel"] = {
		    file = hInfoPkt:getItem("hotelInfo.PersistDataBase"),
		}
	}
	
	return hotelInfo
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Main runnable function 
function run()
	-- Parsing incoming package
    local dsMain = bt_data()
    if not dsMain then
    	exit_msg("Main data set is empty")
    end
    local pkt, err = require"DataModel":new(dsMain)
    if not pkt then
    	exit_msg("Couldn't parse incoming packet: %s", err)
    end
    if not pkt:isPeriodic() then
    	print(tostring(pkt)) --TODO: remove tostring after fix in luaRunner
    end
    
    -- Initialize data storage
    local hInfo    = __loadHotelInfo()
    local stor_cfg = {
    	["mem"]   = { ident = "memc_simple", options = hInfo.memc },
    	["hotel"] = hInfo.hotel,
    	["ast"]   = hInfo.ast 
    }
    require("rawInterfaces"):initializeSingletons(stor_cfg)
    local storage, err = require"DataStorage":new(stor_cfg)
    if not storage then
    	exit_msg(err)
    end
    
    -- Initialize singletons (DataSync should be initialized last)
    local dataProvider = require"DataProvider":getInstance(storage)
	local roomProvider = require"RoomProvider":getInstance(storage)
	require"ModStatusController":getInstance(dataProvider)
	require"WakeUpController":getInstance(dataProvider, roomProvider)
	require"GuestDataController":getInstance(dataProvider, roomProvider)
	require"RoomEquipmentController":getInstance(dataProvider, roomProvider)
	require"PostingSimpleController":getInstance(dataProvider, roomProvider)
	require"DataSyncController":getInstance(dataProvider, roomProvider)
	
	-- processing scenarios
	local queue   = require("fifo"):new()
	local factory = require("ScenariosFactory"):new(storage)
	
	-- Go trough events detector and collect
    -- scenarios which can process current packet
    for item, stat in factory:iterator(pkt, "staticCheck") do
    	if item and stat then
    		queue:push(item)
    	end
    end
    
    if queue:isEmpty() then
    	print("Not found scenario for this message!!!")
    else
    	while not queue:isEmpty() do
    		local scr = factory:instantiate(queue:pop())
    		if scr then
    			if not pkt:isPeriodic() then
    				printdbg("Attempt to execute scenario => %q", scr:getName())
    			end
    			scr:create()
    			local candidates = scr:exec(pkt)
    			if "table" == type(candidates) then
    				for i=1, #candidates do
    					queue:push(candidates[i])
    				end
    			end
    			scr:destroy()
    		end
    	end
    end
end
