--- Data sync controller for executing synchronize procedure
-- 
-- @file	DataSyncController.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/06/29

--- @module DataSyncController
local DataSyncController = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Checking interface status
local function __get_ifaceStatus(ifaceName)
	local modController = require"ModStatusController":getInstance()
	return modController.getStatus4Sync(ifaceName)
end

--- Sending Data Request message to FIAS interface
local function __send_DataRequest(syncData)
	if not syncData then
		return false
	end
	
	local psAnsPkt, err = require"DataModel":new({ident = "FIAS"})
	if not psAnsPkt then
		return false, err
	end
	
	psAnsPkt:setItem("data.ident", 6)	-- DR message
	psAnsPkt:addRepeatedTuple("data.fields", "DA", syncData.date)
	psAnsPkt:addRepeatedTuple("data.fields", "TI", syncData.time)
	
	return require"appHelper":startTransaction(psAnsPkt)
end

--- Sending request for retrieving badges info to XOPEN interface
local function __send_GetBadges(wsInfo)
	if "table" ~= type(wsInfo) or not wsInfo.fc or not wsInfo.uc then
		return false
	end
	
	local dModel			= require"DataModel"
	local XOPEN				= dModel.XOPEN
	local guestDataPkt, err = dModel:new({ident = "XOPEN", endpoint = {f = wsInfo.fc, u = wsInfo.uc}})
	if not guestDataPkt then
		return false, err
	end
	
	guestDataPkt:setItem("data.ident", XOPEN.X_GET_BADGES)
	
	local __handleStatusValue = function(pkt)
    	local rs  = false
    	local err = "We got error during guest data sending!"
    	
    	local ident  = pkt:getItem("data.ident")
    	if XOPEN.X_STATUS == ident then
    		local status = tonumber(pkt:getItem("data.status"))
    		if 200 == status or 210 == status then
    			rs = true
    			err = nil
    		else
    			err = string.format("%s (RC: %d)", err, status)
    		end
    	end
    	
    	return rs, err
    end

    return require"appHelper":startTransactionWithHandleResponce(guestDataPkt, __handleStatusValue)
end

local function __start_syncFromFIAS()
	if "online" ~= __get_ifaceStatus("FIAS") then
		return false, "FIAS interface is not online! Sync request aborted!"
	end
	
	local syncDate, syncTime = require"dateConverter".unix2fias(os.time())
	return __send_DataRequest({date = syncDate, time = syncTime})
end

local function __start_syncFromXOPEN(provider)
	if not provider then
		return false, "Not found provider data"
	end
	if "online" ~= __get_ifaceStatus("XOPEN") then
		return false, "XOPEN interface is not online! Sync request aborted!"
	end
	
	local wsInfo = provider.getGlobalWSMapping()
	
	return __send_GetBadges(wsInfo)
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
DataSyncController._instance = nil -- Singletone instance

function DataSyncController:getInstance(dataProvider, roomProvider)
	if self._instance then
		return self._instance
	end
	if not dataProvider or not roomProvider then
		return false
	end
	
	self._instance = DataSyncController 
	local o = self._instance
	
	o.dataProvider = dataProvider
	o.roomProvider = roomProvider
	
	o.SYS          = require"DataModel".SYS
	
	o.roomEqController 		= require"RoomEquipmentController":getInstance()
	o.wkController 			= require"WakeUpController":getInstance()
	o.guestDataController 	= require"GuestDataController":getInstance()
	
	
	o.initSync = function()
		-- check sync level
    	if o.SYS.M_RESYNC_FORBIDDEN == o.dataProvider.getReSyncLevel() then  
    	    -- resynchronization is forbidden by user  
    		printdbg("Sync procedure is blocked by configuration!")
    		return false
    	end
    	
    	-- set sync state to wait online iface
    	printdbg("Set initial flag for data sync procedure ...")
		o.dataProvider.setSyncState("WaitIfaces")
	end
	
	-- Sync procedure after restart/reboot
	o.runSync = function()
		local syncState = o.dataProvider.getSyncState()
		if not syncState then
			return false
		end
		local syncLevel = o.dataProvider.getReSyncLevel()
		
		if "WaitXOpenStage" == syncState then	-- check sync continious stage (remain RESYNC = Full)
			-- Send sync request to XOPEN
			if o.badgesExecSync() then
				o.dataProvider.setSyncState("SyncXOpenStage")
			end
		elseif "WaitIfaces" == syncState then	-- check iface status for sync
    		local PBXIfaceStatus4Sync   = __get_ifaceStatus("PBX")
            local FIASIfaceStatus4Sync  = __get_ifaceStatus("FIAS")
            local XOPENIfaceStatus4Sync = __get_ifaceStatus("XOPEN")
            
    		if "offline" == PBXIfaceStatus4Sync or
               "offline" == FIASIfaceStatus4Sync or
               "offline" == XOPENIfaceStatus4Sync then
    			printdbg("Data sync procedure wait ready status for all ifaces...")
    			return false
    		end
            
        	-- setting sync mode
        	local syncMode = "RESTART"
    		local magicNumber = o.dataProvider.getMagicNumber()
        	if not magicNumber then
        		syncMode = "REBOOT"
        		o.dataProvider.setMagicNumber()
        	end
        	local initFlag = o.dataProvider.getInitializationFlag()
        	if "1" == initFlag then
        		syncMode = "REINIT"
        	end
        	
        	-- loading sync data
        	local roomsList = nil
        	if "REINIT" ~= syncMode then
        		-- get list of the loaded rooms
            	roomsList = o.roomProvider.getRoomsList()
            	if not roomsList then
            		print("List of the loading room is empty! Sync procedure aborted...")
            		o.dataProvider.delSyncState()
            		return false
            	end
        	end
        	
        	-- processing sync mode
    		if "RESTART" == syncMode then
    			print("Starting data sync after daemon restart ...")
        		o.dataProvider.setSyncState("SyncAfterRestartInProgress")
        		o.wkController.sync(roomsList)					-- wakeUp's sync
        		o.dataProvider.delSyncState()
        		print("Data sync procedure has been finished!")
    		elseif "REBOOT" == syncMode then
    			print("Starting data sync after pbx restart ...")
    			o.dataProvider.setSyncState("SyncAfterRebootInProgress")
    			o.guestDataController.syncRoomState(roomsList) 	-- room guest state sync
        		o.roomEqController.syncEquipment(roomsList)		-- room equipment sync (ACL, MUR, DND)
    			o.wkController.sync(roomsList)					-- wakeUp's sync
        		o.dataProvider.delSyncState()
        		print("Data sync procedure has been finished!")
        	elseif "REINIT" == syncMode then
                -- All data was removed. We have to get them again...
                if o.SYS.M_RESYNC_FULL ~= o.dataProvider.getReSyncLevel() then
                    o.dataProvider.delSyncState()
                    o.dataProvider.setInitializationFlag("0")
                    print("Data sync procedure has been finished!")
                    printdbg("Sending sync requests are denied by configuration!")
                elseif "online" == FIASIfaceStatus4Sync then  -- Send sync request to FIAS
                    if o.masterExecSync() then
                        o.dataProvider.setSyncState("SyncFiasStage")
                        print("Starting data sync by initial system flag via FIAS ...")
                    end
                elseif "online" == XOPENIfaceStatus4Sync then -- Send sync request to XOPEN
                    if o.badgesExecSync() then
                        o.dataProvider.setSyncState("SyncXOpenStage")
                        print("Starting data sync by initial system flag via XOPEN ...")
                    end
                end
    		end
    	end
	end
	
	--- Update data sync procedure. Used only in "REINIT" mode.
	-- @param flag	A flag value
	o.updSync = function(flag)
		if not flag then
			return false
		end
		
		local isSyncFinalized = false
    	if "SyncFiasStage" == o.dataProvider.getSyncState() and "DE" == flag then
    		if "online" == __get_ifaceStatus("XOPEN") then
    			o.dataProvider.setSyncState("WaitXOpenStage")
    			
    			if o.badgesExecSync() then  -- Send sync request to XOPEN
					o.dataProvider.setSyncState("SyncXOpenStage")
				end
    		else
    			isSyncFinalized = true
    		end
    	elseif "SyncXOpenStage" == o.dataProvider.getSyncState() and "RetBadges" == flag then
    		isSyncFinalized = true
    	end
    	
    	if isSyncFinalized then
    		o.dataProvider.delSyncState()				-- clear synchronization state
    		o.dataProvider.setInitializationFlag("0")	-- clear initialization flag
    		print("Data sync procedure has been finished!")
    	end
	end
	
	--- Send sync badges request only to XOPEN interface
	o.badgesExecSync = function()
	    if o.SYS.M_RESYNC_FULL ~= o.dataProvider.getReSyncLevel() then
	       return false
	    end
		local rs, errMsg = __start_syncFromXOPEN(o.roomProvider)
		if not rs then
			print(errMsg)
		end
		return rs
	end
    
    --- Send sync to PMS interface for data as from master backend
    o.masterExecSync = function()
        if o.SYS.M_RESYNC_FULL ~= o.dataProvider.getReSyncLevel() then
           return false
        end
        local rs, errMsg = __start_syncFromFIAS()
        if not rs then
            print(errMsg)
        end
        return rs
    end
end

return DataSyncController