--- Guest data controller for processing Guest and Key actions
-- 
-- @file	GuestDataController.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/05/29

--- @module GuestDataController
local GuestDataController = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Key answer codes:
	--	BY - Telephone / Encoder Busy
	--	DE - Wakeup/Key has been deleted
	--	NA - Night Audit
	--	OK - Command or request completed successfully
	--	RY - Retry
	--	UR - Unprocessable request, this request cannot be carried out, no retry
local _keyRetCodes = {
	["REQUEST"] = {
    	 NO_ROOM		= {code = "UR", desc = "Room configuration error for Wildix Hotel System"}
    	,NO_WSINFO		= {code = "UR", desc = "Workstation configuration error for Wildix Hotel System"}
    	,KEY_IN_CACHE	= {code = "BY", desc = "Processing another key for the room"}
    	,NO_GUESTDATA	= {code = "UR", desc = "Guest was not found in the room"}
    	,NO_APTS		= {code = "UR", desc = "Room/common area configuration error for Wildix Hotel System"}
    	,ERR_SEND		= {code = "UR", desc = "Key system communication error"}
    	,CONFIG_DENY    = {code = "UR", desc = "Key request denied by configuration of Wildix Hotel System"}
    	,NOT_SUPPORTED  = {code = "UR", desc = "Not supported key type"}
    },
    ["DELETE"] = {
    	 TIMEOUT		= {code = "RY", desc = "Canceled by timeout"}
    	,NO_ROOM		= {code = "UR", desc = "Room configuration error for Wildix Hotel System"}
    	,NO_WSINFO		= {code = "UR", desc = "Workstation configuration error for Wildix Hotel System"}
    	,KEY_IN_CACHE	= {code = "BY", desc = "Processing another key for the room"}
    	,NO_GUESTDATA	= {code = "OK", desc = "Guest was not found in the room"}
    	,EMPTY_DATA		= {code = "UR", desc = "Room does not have any keys"}
    	,CONFIG_DENY    = {code = "UR", desc = "Key request denied by configuration of Wildix Hotel System"}
    },
    ["CREATED"]	= {
    	 OK				= {code = "OK", desc = "Key created successfully"}
    	,RETRY			= {code = "RY", desc = "Key system communication error"}
    	,FAIL			= {code = "UR", desc = "Key system occurred error"}
    },
    ["DELETED"] = {
    	OK				= {code = "OK", desc = "Key deleted successfully"}
    }
}

--- processing code set conversion from extended ASCII to UTF8
-- UTF8 supports by XOPEN and PBX and FIAS - extended ASCII.
local function __convertASCII2UTF(str)
    local function _isUTF8String(str)
        local rv = false
        local i   = 0
        if not str then
            return rv, i
        end
        local len = #str
        if 0 == len then
            return rv, i
        end
        
        i = i + 1
        while i <= len do
            if i == str:find("[%z\1-\127]", i) then
                i = i + 1
            elseif i == str:find("[\194-\223][\128-\191]", i) then 
                i = i + 2
            elseif i == str:find("\224[\160-\191][\128-\191]", i) or
              i == str:find("[\225-\236][\128-\191][\128-\191]", i) or
              i == str:find("\237[\128-\159][\128-\191]", i) or
              i == str:find("[\238-\239][\128-\191][\128-\191]", i) then 
                i = i + 3
            elseif i == str:find("\240[\144-\191][\128-\191][\128-\191]", i) or
              i == str:find("[\241-\243][\128-\191][\128-\191][\128-\191]", i) or
              i == str:find("\244[\128-\143][\128-\191][\128-\191]", i) then 
                i = i + 4
            else
                return false, i
            end
        end
        
        return true
    end
    
    local rv = str
    if not str or "" == str then
        return rv
    end
    
    if _isUTF8String(str) then  -- checking UTF8 characters
        return rv
    end
    
    local codeSet       = require"codeSet":new("iso-8859-1", "utf8")
    local fixedStr, err = codeSet:conv(str)
    if fixedStr then
        rv = fixedStr
    else
        printdbg("String code set conversion error: %s", err or "unknown error")
    end
    
    return rv
end

--- apply guest data to WMS user
local function __set_wmsUser(newInfo, oldInfo)
	if "table" ~= type(newInfo) then
		return false
	end
	
	local userOldExten = "table" == type(oldInfo) and oldInfo[1]
	local userExten    = newInfo[1]
	local userData 	   = {}
	if "table" == type(newInfo[2]) then
		userData = {
			 name    = newInfo[2].name
			,surname = newInfo[2].surname
			,lang	 = newInfo[2].lang
		}
	end  
	
	local errMsg = nil
	local rs	 = true
	if userOldExten then
		printdbg("Try to update wmsUser info for old room extension %q", userOldExten)
		rs, errMsg = require"wmsUser":setInfo(userOldExten, {name = "room_"..userOldExten, lang = "en"})
	end
	if userData.name or userData.surname or userData.lang then
		printdbg("Try to update wmsUser info for new room extension %q: %s %s (%s)", userExten, userData.name or "", userData.surname or "", userData.lang or "")
		rs, errMsg = require"wmsUser":setInfo(userExten, userData)
	end
	
	return rs, errMsg
end

--- processing room list retrieved in badge event
local function __processingRoomListByAptID(provider, roomList)
    local rv = {}
    if not provider or "table" ~= type(roomList) then
        return rv
    end
    
    local commonArea = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,max=0} -- 20 available common areas
    local guestRoom  = {}
    for idx, aptId in ipairs(roomList) do
        -- proceed common area information
        local commonAreaInfo = provider.getCommonArea(aptId, "XOPEN")
        if commonAreaInfo then
            local commonAreaIndex       = tonumber(commonAreaInfo.number) or 0
            commonArea.max              = commonAreaIndex > commonArea.max and commonAreaIndex or commonArea.max
            commonArea[commonAreaIndex] = 1
        end
        
        local guestRoomInfo = provider.getRoom(aptId, "XOPEN")
        if guestRoomInfo then
            guestRoom[#guestRoom + 1] = guestRoomInfo
        end
    end
    
    if 1 == #guestRoom then
        rv.guest = guestRoom[1].ident
    elseif 1 < #guestRoom then
        local suitesList = provider.getVirtualRoomsList() or {} 
        for _, vrid in ipairs(suitesList) do
            local vRoomInfo = provider.getVirtualRoom(vrid)
            local suiteInfo = vRoomInfo and vRoomInfo.suite
            if #suiteInfo == #guestRoom then
                for idx, room in ipairs(guestRoom) do
                    if not suiteInfo[room.number] then
                        break
                    end
                    if idx == #suiteInfo then
                        rv.suite   = table.concat(suiteInfo, ";")
                        rv.virtual = vrid 
                    end 
                end
            end
            if rv.virtual and rv.suite then
                break
            end
        end
    end
    
    if 0 < commonArea.max then
        rv.common = table.concat(commonArea):sub(1, commonArea.max)
    end
    
    return rv
end

--- retreiving apparmtent ids from FIAS type
local function __make_aptIdArray(provider, roomNumber, commonAreas, suiteRooms)
	if not provider or not roomNumber then
		return false, "invalid arguments"
	end
	local aptArray = {}
	commonAreas    = commonAreas or ""
	suiteRooms     = suiteRooms or "" 
	
	-- process room information
	local areaInfo = provider.getRoom(roomNumber, "FIAS")
	if areaInfo then
		aptArray[#aptArray + 1] = {aptid = areaInfo.aptid, desc = areaInfo.desc}
	end
	
	-- process common area information
	local comIdx = 0
	for commonArea in commonAreas:gmatch(".") do
		comIdx = comIdx + 1
		commonArea = " " ~= commonArea and commonArea or 0	-- the defaults in the keycard system
		if 0 < tonumber(commonArea) then
			areaInfo = provider.getCommonArea(comIdx, "FIAS")
			if not areaInfo then
			    return false, string.format("not found configured common area [%d]", comIdx)
			end
			aptArray[#aptArray + 1] = {aptid = areaInfo.aptid, desc = areaInfo.desc}
		end
	end
	
	-- process suite area information
	local vRoomInfo = provider.getVirtualRoom(roomNumber)
	local suiteInfo = vRoomInfo and vRoomInfo.suite 
	if suiteInfo and "" ~= suiteRooms then
	   local reqSuiteInfo = {}
	   for suiteRoom in suiteRooms:gmatch("(%d+);?") do
	       if not suiteInfo[suiteRoom] then
	           reqSuiteInfo = {}
	           break
           end
           reqSuiteInfo[suiteRoom] = suiteRoom
           table.insert(reqSuiteInfo, suiteRoom) 
	   end
	   
	   if not reqSuiteInfo[1] or #suiteInfo ~= #reqSuiteInfo then
	       return false, "suite info does not correspond to virtual rooms configuration"
	   end
	   for _, suiteRoom in ipairs(reqSuiteInfo) do
	       areaInfo = provider.getRoom(suiteRoom, "FIAS")
	       if not areaInfo then
	           return false, string.format("not found the configured room from suite info [%s]", suiteRoom)
	       end
           aptArray[#aptArray + 1] = {aptid = areaInfo.aptid, desc = areaInfo.desc} 
	   end
	end
	
	return 0 ~= #aptArray and aptArray
end

-- Preparing guest data for sending "Set/Upd/Del guest data" request to XOPEN
local function __make_guestData(provider, kData, gInfo, bData)
	if not provider or "table" ~= type(kData) then
		return false
	end
	local gInfo = "table" == type(gInfo) and gInfo or {}
	local bData = "table" == type(bData) and bData or {}

	local function __fix_surname(val)  -- WMS-5765, workaround
		if val then
			val = val:gsub("%s+", "")
			if #val > 16 then
				val = val:sub(1, 16)
			end
		end
		return val
	end

	-- processing guest data section
	local bType = ({N = 0, D = 1, O = 2})[bData.type or kData.kType or "N"]
    
    local bGuestName     = kData.rn   -- as undefined in key request
    local bGuestType     = gInfo.gtype	
	local bGuestSurname	 = kData.surname or gInfo.surname or "Undefined"
	if not bGuestSurname then
	   return false, "Unknown badge guest surname!"
	end
	bGuestSurname = __fix_surname(bGuestSurname)
	bGuestSurname = __convertASCII2UTF(bGuestSurname)
	
	local dateConv		 = require"dateConverter"
	local bCheckInDate   = dateConv.fias2unixUTC(kData.arrDate, "000000")
	if 0 == bCheckInDate then
		bCheckInDate = bData.startV or gInfo.arrDate
	end
	local bCheckOutDate	 = dateConv.fias2unixUTC(kData.depDate, kData.depTime or kData.time)
	if 0 == bCheckOutDate then
		bCheckOutDate = bData.endV or gInfo.depDate
	end
	if not bCheckOutDate then
	   return false, "Invalid/not found check-out date!"
	end
	
	-- collect all apts ID
	local aptsList, errMsg = __make_aptIdArray(provider, kData.rn, kData.kOpt or bData.common, kData.sInfo or bData.suite)
	if not aptsList then
	   return false, errMsg
	end
	
	-- collect old apts ID (for notification in case room moving)
	local aptsOldList = __make_aptIdArray(provider, kData.ro)
	
	-- push badge serials
	local badgesList = nil
	if "table" == type(bData.serial) then
		 badgesList = bData.serial
	elseif bData.serial then
		badgesList = {bData.serial}
	end
	
	local data = { -- result guest data message
		 checkInDate  = bCheckInDate
		,checkOutDate = bCheckOutDate
		,guestName	  = bGuestName
		,guestSurname = bGuestSurname
		,guestType	  = bGuestType
		,bageType	  = bType
		,aptsList	  = aptsList
		,aptsOldList  = aptsOldList
		,badgesList	  = badgesList
	}
    
	return data
end

--- making composite code from badge return codes
local function __make_compositeCode(compositeCode, aptCode)
	local codesTable = {
		 OK    = {0, 4}
		,RETRY = {1, 2, 5, 6, 9}
		,FAIL  = {3, 7, 8, 10} 
	}
	for codeName, codesPool in pairs(codesTable) do
		if "table" == type(codesPool) then
    		for _, numCode in pairs(codesPool) do
    			codesTable[numCode] = codeName
			end
		end
	end
	
	local status = codesTable[aptCode] or "FAIL"
	if "FAIL" == status then
		compositeCode = status
	elseif "RETRY" == status and "FAIL" ~= compositeCode then
		compositeCode = status
	elseif "FAIL" ~= compositeCode and "RETRY" ~= compositeCode then
		compositeCode = status
	end
	
	return compositeCode
end

--- creating date for badge response (like badge delete)
local function __make_respBadge(bData, compositeCode)
	local rv = nil
	if "table" ~= type(bData) then
		return rv
	end
	if "OK" ~= compositeCode then
		return rv
	end
	
	-- processing workstation info
	local ws = bData.workstation
	if "table" ~= type(ws) or not ws.fc or not ws.uc then
		return rv
	end
	
	-- processing guest info
	local checkInDate  = bData.startv
	local checkOutDate = bData.endv
	if not checkOutDate then
		return rv
	end
	local guestName    = bData.name
	local guestSurname = bData.surname
	if not guestName or not guestSurname then
		return rv
	end
	
	-- processing badge info
	local badgeSerial = bData.serial
	if not badgeSerial then
		return rv
	end
	
	if "table" ~= type(bData.apts) then
		return rv
	end
	local aptsList = {}
	for _, aptInfo in pairs(bData.apts) do
		aptsList[#aptsList + 1] = {aptid = aptInfo.aptID}
	end
	
	-- building return badge data
	rv = {
		 ws 	   = ws
		,guestData = {
			 checkInDate  = checkInDate
    		,checkOutDate = checkOutDate
    		,guestName	  = guestName
    		,guestSurname = guestSurname
    		,aptsList	  = aptsList
    		,badgesList	  = {badgeSerial}
		}
	}
		
	return rv
end

--- send guestData request to XOPEN
local function __send_guestData(guestData, msgType, endPoint)
	if not guestData or not msgType or not endPoint then
		return false, "Invalid arguments"
	end
	
	local dModel			= require"DataModel"
	local XOPEN				= dModel.XOPEN
	local guestDataPkt, err = dModel:new({ident = "XOPEN", endpoint = endPoint})
	if not guestDataPkt then
		return false, err
	end
	
	local ident = XOPEN.X_INVALID_IDENT
	if "SET" == msgType then
		ident = XOPEN.X_SET_GUEST_DATA
	elseif "UPD" == msgType then
		ident = XOPEN.X_UPD_GUEST_DATA
	elseif "DEL" == msgType then
		ident = XOPEN.X_DEL_GUEST_DATA
	end
	
	guestDataPkt:setItem("data.ident", ident)
	guestDataPkt:setItem("data.guestData.guest.guest_check_in_date", tonumber(guestData.checkInDate))
	guestDataPkt:setItem("data.guestData.guest.guest_check_out_date", tonumber(guestData.checkOutDate))
	guestDataPkt:setItem("data.guestData.guest.guest_surname", guestData.guestSurname)
    guestDataPkt:setItem("data.guestData.guest.guest_name", guestData.guestName)
	if "DEL" ~= msgType then
		guestDataPkt:setItem("data.guestData.guest.guest_type", tonumber(guestData.guestType))
	end

	if "table" == type(guestData.aptsList) then
       	for _, aptItem in pairs(guestData.aptsList) do
       		if "table" == type(aptItem) and aptItem.aptid and "" ~= aptItem.aptid then 
       			local apts    = guestDataPkt:addRepeatedItem("data.guestData.apts.items")
       			apts["aptID"] = aptItem.aptid
       			if aptItem.desc and "" ~= aptItem.desc then
       				apts["description"] = aptItem.desc
       			end
       		end
       	end
    end
    
    if "UPD" == msgType and "table" == type(guestData.aptsOldList) then
        for _, aptOldItem in pairs(guestData.aptsOldList) do
            if "table" == type(aptOldItem) and aptOldItem.aptid and "" ~= aptOldItem.aptid then 
                local apts    = guestDataPkt:addRepeatedItem("data.guestData.oldApts.items")
                apts["aptID"] = aptOldItem.aptid
                if aptOldItem.desc and "" ~= aptOldItem.desc then
                    apts["description"] = aptOldItem.desc
                end
            end
        end
    end
    
	if "SET" ~= msgType and "table" == type(guestData.badgesList) then
    	for _, badgeSN in pairs(guestData.badgesList) do
    		local badges = guestDataPkt:addRepeatedItem("data.guestData.badges.items")
    		badges["sn"] = badgeSN
    	end
    end
    
    local __handleStatusValue = function(pkt)
    	local rs  = false
    	local err = "We got error during guest data sending!"
    	
    	local ident  = pkt:getItem("data.ident")
    	if XOPEN.X_STATUS == ident then
    		local status = tonumber(pkt:getItem("data.status"))
    		if 200 == status or 210 == status then
    			rs = true
    			err = nil
    		else
    			err = string.format("%s (RC: %d)", err, status)
    		end
    	end
    	
    	return rs, err
    end

    return require"appHelper":startTransactionWithHandleResponce(guestDataPkt, __handleStatusValue)
end

--- Send key answer message to FIAS
local function __send_keyAnswer(keyData)
	if not keyData then
		return false
	end
	
	local keyAnsPkt, err = require"DataModel":new({ident = "FIAS", options = {ift = 2}})
	if not keyAnsPkt then
		return false, err
	end
	
	keyAnsPkt:setItem("data.ident", 36)	-- KA message
	keyAnsPkt:addRepeatedTuple("data.fields", "KC", keyData.kCoder)
	keyAnsPkt:addRepeatedTuple("data.fields", "WS", keyData.wsId)
	keyAnsPkt:addRepeatedTuple("data.fields", "CT", keyData.text)
	keyAnsPkt:addRepeatedTuple("data.fields", "AS", keyData.status)
	keyAnsPkt:addRepeatedTuple("data.fields", "$2", keyData.trackID)
	keyAnsPkt:addRepeatedTuple("data.fields", "DA", keyData.timestamp.date)
	keyAnsPkt:addRepeatedTuple("data.fields", "TI", keyData.timestamp.time)
	
	return require"appHelper":startTransaction(keyAnsPkt)
end

local function __notifyRoomState(provider, msgType, guestData)
    if not msgType or "table" ~= type(guestData) then
        return false, "Invalid data for room state notification"
    end
    
    -- check room number data
    if not guestData.rn then
        return false, "Not found room number in the guest data"
    end
    local roomInfo = provider.getRoom(guestData.rn, "FIAS")
    if not roomInfo then
        return false, string.format("Not found room %q for check-in request!", guestData.rn)
    end
    
    -- processing workstation data
    local wsInfo = provider.getMapping(guestData.wsId)
    if not wsInfo then  -- let's get random WS ID if it was not got
        wsInfo = provider.getRandomWSMapping()
    end
    
    -- preparing data for guest data message
    local guestInfo = {
         name    = guestData.name
        ,surname = guestData.surname or guestData.rn
        ,gtype   = tonumber(guestData.gtype) or 0
    }
    
    local curDate, curTime = require"dateConverter".unix2fias(os.date("%s"))
    local keyData   = {
         rn      = guestData.rn
        ,ro      = guestData.ro
        ,arrDate = guestData.arrDate or guestData.date or curDate
        ,depDate = guestData.depDate or guestData.date or curDate
        ,time    = guestData.time or curTime
    }
    local bInfo     = {}
    if "SET" ~= msgType then
        bInfo.serial = "FFFFFFFF"
    end
    local guestDataMsg, errMsg = __make_guestData(provider, keyData, guestInfo, bInfo)
    if not guestDataMsg then
        return false, string.format("Not found mandatory guest data from request for room %q: %s!", guestData.rn, errMsg)
    end

    -- sending guest data to XOPEN 
    local rs, err = __send_guestData(guestDataMsg, msgType, {f = wsInfo.fc, u = wsInfo.uc})
    return rs, err    
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
GuestDataController._instance = nil -- Singletone instance

function GuestDataController:getInstance(dataProvider, roomProvider)
	if self._instance then
		return self._instance
	end
	if not dataProvider or not roomProvider then
		return false
	end
	
	self._instance = GuestDataController 
	local o = self._instance
	
	o.dataProvider = dataProvider
	o.roomProvider = roomProvider
	
	--- Processing guest check-in request
	-- @param guestData			A data with guest information
	-- @param useXOpenIface		A flag for disabling sending messages to XOPEN interface
	-- @return boolean value
	o.guestCheckIn = function(guestData, useXOpenIface)
		if "table" ~= type(guestData) then
			printdbg("Invalid guest data")
			return false
		end
		
		-- check room number data
		if not guestData.rn then
			printdbg("Not found room number in the guest data")
			return false
		end
		local roomInfo = o.roomProvider.getRoom(guestData.rn, "FIAS")
		if not roomInfo then
			print("Not found room %q for check-in request!", guestData.rn)
			return false
		end
		local roomIdent = roomInfo.ident
		
		-- check guest in the room and remove them without notification
		local guestId = o.roomProvider.getRoomGuest(roomIdent)
		if guestId then
			printdbg("Processing quiet check-out (by check-in to occupied room) ...")
			o.guestCheckOut(guestData, useXOpenIface)
		end
		
		-- restore initial state for room regarding extension voicemails
		local vCnt, errMsg = require"userVoicemail".removeAll(roomInfo.exten)
		if errMsg then print(errMsg) end
		
		-- processing guest data information
		guestData.name 	  = __convertASCII2UTF(guestData.name)
		guestData.surname = __convertASCII2UTF(guestData.surname or guestData.rn)
		
		local guestId = o.dataProvider.setGuestData(guestData, roomInfo)
		if not guestId then
			print("Error in the guest data saving (for room - %s)!", guestData.rn)
			return false
		end
		
		-- move in guest to the room
		o.roomProvider.setRoomGuest(roomIdent, guestId)
		
		-- processing room attributes
		local rEqController = require"RoomEquipmentController":getInstance()
		local eqData = {
			 ifn 	  	= "FIAS"
			,roomIdent 	= guestData.rn
			,ACL		= guestData.cos or 0
		}  
		rEqController.setEquipment(eqData)
		
		-- update WMS user data by guest data information
		local rs, errMsg = __set_wmsUser({roomInfo.exten, guestData})
		if errMsg then print(errMsg) end
		
		-- send notification about room state
		if 1 == o.dataProvider.getRoomStateNotification() and useXOpenIface then
            printdbg("Sending room state notification for room %q", guestData.rn)
            local rs, errMsg = __notifyRoomState(o.roomProvider, "SET", guestData)
            if errMsg then print(errMsg) end  
		end
		
		local swapInfo = guestData.swap and " (Swap Flag)" or ""
		print("Guest CheckIn operation for room %q was made successfully%s!", guestData.rn, swapInfo)
		return true
	end
	
	--- Processing guest check-out request
	-- @param guestData			A data with guest information
	-- @param useXOpenIface		A flag for disabling sending messages to XOPEN interface
	-- @return boolean value
	o.guestCheckOut = function(guestData, useXOpenIface)
		if "table" ~= type(guestData) then
			printdbg("Invalid guest data")
			return false
		end
		
		-- get room information
		if not guestData.rn then
			print("Not found room number in the guest data")
			return false
		end
		local roomInfo = o.roomProvider.getRoom(guestData.rn, "FIAS")
		if not roomInfo then
			print("Not found room %q for check-out request!", guestData.rn)
			return false
		end
		local roomIdent = roomInfo.ident
		
		-- get guest id
		local guestId = o.roomProvider.getRoomGuest(roomIdent)
		if not guestId then
			print("Room %q has already been empty", guestData.rn)
			return true
		end
		
		-- remove all wakeups for this room (without removing from ast storage)
		local wkController = require"WakeUpController":getInstance()
		wkController.delete({ifn = "FIAS", rn = roomInfo.number, da = "000000", ti = "000000"}, "GuestOut")
		
		-- remove all voicemails for this room
		local vCnt, errMsg = require"userVoicemail".removeAll(roomInfo.exten)
		if not vCnt then
			if errMsg then print(errMsg) end
		elseif 0 < vCnt then
			printdbg("Removed %d voicemail(s) for room %q!", vCnt, guestData.rn)
		end
		
		-- remove all room equipment
		local rEqController = require"RoomEquipmentController":getInstance()
		rEqController.delEquipment(roomInfo, useXOpenIface)
		
		if useXOpenIface then
            if 1 == o.dataProvider.getRoomStateNotification() then
                printdbg("Sending room state notification for room %q", guestData.rn)
                local rs, errMsg = __notifyRoomState(o.roomProvider, "DEL", guestData)
                if errMsg then print(errMsg) end
            elseif not guestData.swap then
                -- make key deleting for all guests without Key Answer
                o.internalKeyDelete({wsId = guestData.wsId, rn = guestData.rn})
            end
		end
		
		-- remove guest from room and hotel storage
		local rv, err = o.dataProvider.delGuestData(guestId)
		if not rv then
		    printdbg("Error for deleting guest data: %s", err or "unknown error")
		end
		o.roomProvider.delRoomGuest(roomIdent)
		
		-- update WMS user data to default values
		local rs, errMsg = __set_wmsUser({roomInfo.exten, {name = "room_"..roomInfo.exten, lang = "en"}})
		if errMsg then print(errMsg) end

		local swapInfo = guestData.swap and " (Swap Flag)" or ""
		print("Guest CheckOut operation for room %q was made successfully%s!", guestData.rn, swapInfo)		
		return true
	end
	
	--- Processing guest data change request
	-- @param guestData			A data with guest information
	-- @param useXOpenIface		A flag for disabling sending messages to XOPEN interface
	-- @return boolean value 
	o.guestDataChange = function(guestData, useXOpenIface)
		if "table" ~= type(guestData) then
			printdbg("Invalid guest data")
			return false
		end
		
		local srcRoomInfo = nil
		local dstRoomInfo = nil
		
		-- get room information
		if not guestData.rn then
			print("Not found room number in the guest data")
			return false
		end
		local roomInfo = o.roomProvider.getRoom(guestData.rn, "FIAS")
		if not roomInfo then
			print("Not found room %q for guest data change request!", guestData.rn)
			return false
		end
		srcRoomInfo = roomInfo
		dstRoomInfo = roomInfo
		
		-- check destination room information
		if guestData.ro then
			local roomInfo = o.roomProvider.getRoom(guestData.ro, "FIAS")
			if not roomInfo then
				print("Not found room %q for guest data change!", guestData.ro)
				return false
			end
			srcRoomInfo = roomInfo
		end
		
		-- create/update guest data
		local guestId = o.roomProvider.getRoomGuest(srcRoomInfo.ident)
		if not guestId then
			print("Source room %q does not have guests!", srcRoomInfo.number)
			return o.guestCheckIn(guestData, useXOpenIface)
		end
		
		-- update guest info into WMS and hotel storage
		local isUpdateWMSUser = false
		if guestData.ro or guestData.name or guestData.surname or guestData.lang then
			-- fill empty guest data in the request
			local presentGuestData = {}	
			if not guestData.name or not guestData.surname or not guestData.lang then
				presentGuestData = o.dataProvider.getGuestDataById(guestId)	-- read full guest data
			end
			guestData.name 	  = __convertASCII2UTF(guestData.name or presentGuestData.name)
			guestData.surname = __convertASCII2UTF(guestData.surname or presentGuestData.surname)
			guestData.lang 	  = guestData.lang or presentGuestData.lang
			isUpdateWMSUser	  = true
		end
		local guestId = o.dataProvider.setGuestData(guestData, dstRoomInfo, guestId)
		if not guestId then
			print("Error in the guest data updating (ID - %s)!", guestId)
			return false
		end
		
		-- updating WMS user information
		if isUpdateWMSUser then
			local wmsUserInfo = { name = guestData.name, surname = guestData.surname ,lang = guestData.lang}
			local rs, errMsg = __set_wmsUser({dstRoomInfo.exten, wmsUserInfo}, guestData.ro and {srcRoomInfo.exten})
			if errMsg then print(errMsg) end
		end
		
		if useXOpenIface then
            if 1 == o.dataProvider.getRoomStateNotification() then
                printdbg("Sending room state notification for room %q", guestData.rn)
                local rs, errMsg = __notifyRoomState(o.roomProvider, "UPD", guestData)
                if errMsg then print(errMsg) end
			end
		end
		
		if guestData.ro then	-- moving to new room
			-- checking destination room for check-out
			o.guestCheckOut({rn = dstRoomInfo.number}, useXOpenIface)
			local vCnt, errMsg = require"userVoicemail".removeAll(dstRoomInfo.exten)
			if errMsg then print(errMsg) end
			
			-- make key deleting for all guests without Key Answer
            o.internalKeyDelete({wsId = guestData.wsId, rn = guestData.ro})
			
			-- change guest attribure for rooms
			o.roomProvider.delRoomGuest(srcRoomInfo.ident)
			o.roomProvider.setRoomGuest(dstRoomInfo.ident, guestId)
			
			-- update wakeUp values for rooms
			local wkController = require"WakeUpController":getInstance()
			wkController.move(srcRoomInfo.ident, dstRoomInfo.ident)
			
			-- remove all related voicemails for old room
			local vCnt, errMsg = require"userVoicemail".removeAll(srcRoomInfo.exten)
			if not vCnt then
				if errMsg then print(errMsg) end
			elseif 0 < vCnt then
				printdbg("Removed %d voicemail(s) for old room %q!", vCnt, guestData.rn)
			end
			
			-- update attribute values for rooms
			local rEqController = require"RoomEquipmentController":getInstance()
			local eqData        = rEqController.getEquipment(srcRoomInfo.ident)
			eqData.ifn 		 	= "FIAS"
			eqData.roomIdent 	= dstRoomInfo.number
			eqData.ACL			= guestData.cos or eqData.ACL
			rEqController.setEquipment(eqData, nil, useXOpenIface)			-- set equipment to new room	
			rEqController.delEquipment(srcRoomInfo, useXOpenIface)			-- remove equipment from old room
			
			print("Guest change from room %q to %q was made successfully!", srcRoomInfo.number, dstRoomInfo.number)
		else 					-- update info for current room
			if guestData.cos then	-- let's update only user ACL
        		local rEqController = require"RoomEquipmentController":getInstance()
        		local eqData = {
        			 ifn 	  	= "FIAS"
        			,roomIdent 	= guestData.rn
        			,ACL		= guestData.cos or 0
        		}  
        		rEqController.setEquipment(eqData)
        	end
			
			print("Guest information from room %q was updated successfully!", srcRoomInfo.number)
		end
		
		return true
	end
	
	--- Sync guest state in the rooms from Hotel storage to Mem storage
	-- @param roomsList		A list of the rooms idents that must be synced
	-- @return boolean value
	o.syncRoomState = function(roomsList)
		if "table" ~= type(roomsList) then
			return false
		end
		
		local roomGuestCnt = 0
		for _, roomIdent in ipairs(roomsList) do
			local gInfo = o.dataProvider.getGuestDataByRoom(roomIdent)
			if gInfo.gid then
				o.roomProvider.setRoomGuest(roomIdent, gInfo.gid)
				roomGuestCnt = 1 + roomGuestCnt
			end
		end
		
		printdbg("Synced check-in guest for %d rooms", roomGuestCnt)
		return true
	end
	
	--- Processing key request data
	-- @param keyData		    A data with key information
	-- @param isRequireGuest    The flag indicated configuration for PBX FIAS interface
	-- @return boolean value and answered code
	o.keyRequest = function(keyData, isRequireGuest)
		if "table" ~= type(keyData) then
			printdbg("Invalid key data")
			return false
		end
		
		local rv 	   		= false
		local rData	   		= nil
		local _keyRetCodes 	= _keyRetCodes["REQUEST"]
		
		repeat
			if not keyData.kCoder or not keyData.rn or not keyData.wsId then
				print("Key data does not contain mandatory fields")
				break
			end
			if "N" ~= keyData.kType and "D" ~= keyData.kType then
				print("Key type %q is not valid!", keyData.kType)
				rData = {keyData = keyData, status = _keyRetCodes["NOT_SUPPORTED"]}
				break
			end
			
			if 1 == o.dataProvider.getRoomStateNotification() then
			    print("Can't be processed by configuration settings for badge processing")
			    rData = {keyData = keyData, status = _keyRetCodes["CONFIG_DENY"]}
                break
            end
			
			-- get workstation information
			local wsId   = keyData.wsId
			local wsInfo = o.roomProvider.getMapping(wsId) 
			if "table" ~= type(wsInfo) or not wsInfo.fc or not wsInfo.uc then
				print("Not found workstation info for WSID: %s !", wsId)
				rData = {keyData = keyData, status = _keyRetCodes["NO_WSINFO"]}
				break
			end
			
			-- get room information
			local vRoomInfo = nil
    		local roomInfo  = o.roomProvider.getRoom(keyData.rn, "FIAS")
    		if not roomInfo then
    		    vRoomInfo = o.roomProvider.getVirtualRoom(keyData.rn)
    		    if not vRoomInfo then
        			print("Not found room %q for key request!", keyData.rn)
	       			rData = {keyData = keyData, status = _keyRetCodes["NO_ROOM"]}
			     	break
		     	end
    		end
    		local roomIdent = roomInfo and roomInfo.ident
    		local vRoomIdent = vRoomInfo and vRoomInfo.number
    		if vRoomIdent then    -- remove requirements regarding guest in the room
    		   isRequireGuest = nil
    		   roomIdent      = vRoomIdent
    		end
    		
    		-- checking key cache
            local cacheUID = "CR"..(roomIdent or "")
            if o.dataProvider.checkKeyCache(cacheUID) then
                print("Key request for room %q was found in the cache! (Dropping key request ...)", keyData.rn)
                rData = {keyData = keyData, status = _keyRetCodes["KEY_IN_CACHE"]}
                break
            end
            
            -- retrieving guest info
            local guestId = o.roomProvider.getRoomGuest(roomIdent)
            if not guestId and isRequireGuest then  -- making check-in for empty room (PBX iface active)
                printdbg("Making quiet check-in procedure for room %q", keyData.rn)
                local reqData = {
                     rn      = keyData.rn
                    ,name    = keyData.depDate
                    ,surname = keyData.surname or keyData.userId or keyData.arrDate or keyData.depDate
                    ,arrDate = keyData.arrDate
                    ,depDate = keyData.depDate
                }
                local rv, err = o.guestCheckIn(reqData, true)
                if not rv then
                    rData = {keyData = keyData, status = _keyRetCodes["NO_GUESTDATA"]}
                    break
                end
            end
            local guestInfo = o.dataProvider.getGuestDataById(guestId)
                		
    		-- processing key type value
            if "N" == keyData.kType then    -- remove current keys in room for NEW key type
                printdbg("Try to removing all existing keys in the room %q", keyData.rn)
                -- sending delete command without Key Answer
                o.internalKeyDelete({wsId = keyData.wsId, rn = keyData.rn}, isRequireGuest) 
            end
            if "D" == keyData.kType then
                if not o.dataProvider.getBadgeByRoomIdent(roomIdent) then
                    keyData.kType = "N"
                end
            end
    		
			-- preparing data for guest data message
			local guestDataMsg, errMsg = __make_guestData(o.roomProvider, keyData, guestInfo)
			if not guestDataMsg then
				print("Not found mandatory guest data from key request for room %q: %s!", keyData.rn, errMsg)
				rData = {keyData = keyData, status = _keyRetCodes["NO_APTS"]}
				break
			end
			
			-- sending guest data message
			local rs, err = __send_guestData(guestDataMsg, "SET", {f = wsInfo.fc, u = wsInfo.uc})
			if not rs then
				print(err)
				rData = {keyData = keyData, status = _keyRetCodes["ERR_SEND"]}
				break
			end
			
			-- save key request for checking by some answer with timeout
			local cachedData 	= {
				 reqType   = "SET" 
				,keyCoder  = keyData.kCoder
				,ws        = keyData.wsId
				,rid       = roomIdent
				,badgeType = guestDataMsg.bageType
			}
			o.dataProvider.setKeyCache(cacheUID, cachedData)
			
			print("Key request for room %q is processing ...", keyData.rn)
			rv = true 
		until true
		
		return rv, rData
	end
	
	--- Processing key delete data
	-- @param keyData		    A data with key information
	-- @param isRequireGuest    The flag indicated configuration for PBX FIAS interface
	-- @param forceFlag        A flag for force removing by internal command
	-- @return boolean value and answered code
	o.keyDelete = function(keyData, isRequireGuest, forceFlag)
		if "table" ~= type(keyData) then
			printdbg("Invalid key data")
			return false
		end
		
		local rv 	   		= false
		local rData	   		= nil
		local _keyRetCodes 	= _keyRetCodes["DELETE"]
		
		repeat
			-- We will not save key to cache if keyCoder is not found
			if not keyData.rn then
				print("Key data does not contain mandatory fields")
				break
			end
			
			if 1 == o.dataProvider.getRoomStateNotification() then
                print("Can't be processed by configuration settings for badge processing")
                rData = {keyData = keyData, status = _keyRetCodes["CONFIG_DENY"]}
                break
            end
			
			-- get workstation information
			local wsId   = keyData.wsId
			local wsInfo = o.roomProvider.getMapping(wsId) 
			
			-- get room information
			local vRoomInfo = nil
    		local roomInfo  = o.roomProvider.getRoom(keyData.rn, "FIAS")
    		if not roomInfo then
    		    vRoomInfo = o.roomProvider.getVirtualRoom(keyData.rn)
                if not vRoomInfo then
                    print("Not found room %q for key request!", keyData.rn)
                    rData = {keyData = keyData, status = _keyRetCodes["NO_ROOM"]}
                    break
                end
    		end
    		local roomIdent  = roomInfo and roomInfo.ident
    		local vRoomIdent = vRoomInfo and vRoomInfo.number
            if vRoomIdent then    -- remove requirements regarding guest in the room
               isRequireGuest = nil
               roomIdent      = vRoomIdent
            end
    		
    		-- checking key cache
    		local cacheUID = "DL" .. (roomIdent or "")
    		if not forceFlag then
        		if o.dataProvider.checkKeyCache(cacheUID) then
        			print("Key request for room %q was found in the cache! (Dropping key delete ...)", keyData.rn)
        			rData = {keyData = keyData, status = _keyRetCodes["KEY_IN_CACHE"]}
        			break
        		end
    		end
    		
    		-- get guest information from room
    		-- try to remove badges even room is empty (for some failure cases)
    		local guestId   = o.roomProvider.getRoomGuest(roomIdent)
            local guestInfo = o.dataProvider.getGuestDataById(guestId)
            
            -- processing badges info
            local badges = o.dataProvider.getBadgeByRoomIdent(roomIdent) or {}
            
			-- creating guest data for requests
			local badgeCount  	= 0
			local guestDataMsgs = {}
			for _, bInfo in pairs(badges) do
				if "table" == type(bInfo) then
    				local guestDataMsg = __make_guestData(o.roomProvider, keyData, guestInfo, bInfo)
    				if guestDataMsg then
       					guestDataMsgs[bInfo.serial] = guestDataMsg
       					badgeCount 					= badgeCount + 1	
        			end
        		end
			end
			
			-- check count of guest data items
			if 0 == badgeCount then
				print("Not found badges data for current key delete!")
				rData = {keyData = keyData, status = _keyRetCodes["EMPTY_DATA"]}
    			break
			end
			
   			-- let's get random WS ID if it was not got
   			if not wsInfo then
   				wsInfo = o.roomProvider.getRandomWSMapping()
   				printdbg("Try to use random WSID - %s/%s", wsInfo.fc or "", wsInfo.uc or "")
   			end
   			if "table" ~= type(wsInfo) or not wsInfo.fc or not wsInfo.uc then
   				print("Not found workstation info for WSID: %s !", wsId)
   				rData = {keyData = keyData, status = _keyRetCodes["NO_WSINFO"]}
				break
   			end
   			
			-- sending guest data to XOPEN for every badge
			for _, data in pairs(guestDataMsgs) do
				local rv, err = __send_guestData(data, "DEL", {f = wsInfo.fc, u = wsInfo.uc})
				if not rv then
				    if forceFlag then
				        print("!!!! Error during badge removing - %s !!!!!", err)
			        else
				        print("!!!! Badge was not removed - %s !!!!", err)
			        end
			    end
    		end
    		
    		if not forceFlag then
        		-- save the delete request to cache with count of the keys.
        		-- We will skip this record for empty keyCoder
       			local cachedData = {
       				 reqType  = "DEL"
       				,keyCoder = keyData.kCoder	
       				,ws  	  = keyData.wsId
       				,rid      = roomIdent
       				,badgeCnt = badgeCount
       			}
       			o.dataProvider.setKeyCache(cacheUID, cachedData)
   			end
   			
			print("Key delete for room %q is processing ...", keyData.rn)
			rv = true
		until true
		
		return rv, rData
	end
	
	--- Processing system internal key delete data
    -- @param keyData           A data with key information
    -- @param isRequireGuest    The flag indicated configuration for PBX FIAS interface
    -- @note Used in some cases to force badge removing without response (KA).
	o.internalKeyDelete = function(keyData, isRequireGuest)
	   local isRequireGuest = nil == isRequireGuest and true or isRequireGuest
	   return o.keyDelete(keyData, isRequireGuest, true)
	end
	
	--- Sending key answer to FIAS
	-- @param ansData		A data with key answer information
	-- @return boolean value 
	o.KeyAnswer = function(ansData)
		if "table" ~= type(ansData) then
			printdbg("Invalid key data")
			return false
		end
        
		-- try to clear chached key request
		if ansData.cacheUID then
			o.dataProvider.delKeyCache(ansData.cacheUID)
		end
        
        -- processing key coder
        local keyData = ansData.keyData
		if not keyData.kCoder then
			-- KA does not needed, because request was generated without KR/KD
			printdbg("Skip sending key answer, because initial key request was not sent!")
			return true
		end
		
		if not keyData.wsId or "table" ~= type(ansData.status) then
			print("Key data does not contain mandatory fields")
			return false
		end
		keyData.status = ansData.status.code
		keyData.text   = ansData.status.desc
		
        -- checking workstation ID
		local wsId   = keyData.wsId
        local wsInfo = o.roomProvider.getMapping(wsId) 
        if not wsInfo or not wsInfo.fc or not wsInfo.uc then
            wsInfo = o.roomProvider.getRandomWSMapping()
            printdbg("Try to use random WSID - %s/%s", wsInfo.fc or "", wsInfo.uc or "")
        end
		if "table" ~= type(wsInfo) or not wsInfo.fc or not wsInfo.uc then
			print("Not found workstation info for WSID: %s !", wsId)
			return false
		end
		
		-- retrieve local date and time
		keyData.timestamp = {
		     date = os.date("%y%m%d")
		    ,time = os.date("%H%M%S") 
		} 
		
		local rv, err = __send_keyAnswer(keyData)
		if not rv then print(err) end
		return rv
	end
	
	--- Processing cached key requests by timeout
	-- @return table with data for timeouted keys
	o.checkKeysTimeout = function()
		local keyTimeout = o.dataProvider.getBadgesResponseTimeout()	-- periodic times (t x 10 (periodic) seconds)
		local ansKeyData = {}
		
		for cacheUID in o.dataProvider.iterateKeyCacheMap() do
			local cacheVal = o.dataProvider.getKeyCache(cacheUID)
			if cacheVal and cacheVal.divider then
				local divider = cacheVal.divider
				if keyTimeout <= divider then
					local keyData = {
						 kCoder = cacheVal.keyCoder
						,wsId	= cacheVal.ws
					}
					table.insert(ansKeyData, {keyData = keyData, status = _keyRetCodes["DELETE"]["TIMEOUT"], cacheUID = cacheUID})
					printdbg("Found cached key request with expiration time: %s", cacheUID or "")
				else
					cacheVal.divider = divider + 1
					o.dataProvider.setKeyCacheValueOnly(cacheUID, cacheVal) 
				end					
			end
		end
		
		return ansKeyData
	end
	
	--- Processing XOPEN event of a created badges
	-- @param bData			   A badges data
	-- @param isRequireGuest   A flag for additional room state checking
	-- @param syncExecuting    Indicate about synchronization procedure is running
	-- @return boolean value, badge data and answered data
	-- @note Return badge data sends in case some error. Expected that key answer has alerady sent
	o.badgesCreated = function(bData, isRequireGuest, syncExecuting)
		local rv = false
		if "table" ~= type(bData) then
			printdbg("Invalid badges data")
			return rv
		end
		
		if not bData.serial or "table" ~= type(bData.apts) then
			print("Badges event does not contain mandatory fields")
			return rv
		end
		if 0 ~= bData.type then
		    print("Unknown badge type in the message")
		    return rv
		end
		
		-- get room list from badge message
		local roomInfoList      = {}
		local compositeRetCode 	= "OK"
		for _, aptData in pairs(bData.apts) do
			if "table" == type(aptData) then
			    -- retrieve room information
				roomInfoList[#roomInfoList + 1] = aptData.aptID
				
				-- makeing composite return code
                compositeRetCode = __make_compositeCode(compositeRetCode, aptData.ret_code)
    		end
		end
		-- processing room list by devide to common area, suite and guest rooms
		roomInfoList = __processingRoomListByAptID(o.roomProvider, roomInfoList)
		
		-- determine internal mandatory values
		local roomIdent = roomInfoList.guest or roomInfoList.virtual
		if roomInfoList.virtual and roomInfoList.suite then   -- remove requirements regarding guest in the room  
		    isRequireGuest = false
		end
		
		-- retrieve cache identifier
		local cacheUID = roomIdent and "CR" .. roomIdent
		
		-- look up cache for created badge
        local cacheVal = o.dataProvider.getKeyCache(cacheUID)
        if syncExecuting then
            print("Processing event badge created via synchronization procedure")
            cacheVal = { reqType = "SET" }
        end
		
		local rData    = nil
		local rBadge   = nil
		repeat
    		if not roomIdent then
    			print("Intenral error with not valid room identifier")
    			rBadge = __make_respBadge(bData, compositeRetCode)
    			break
    		end
    		if not cacheVal then
                print("Not found waiting key response for event badge created!")
                rBadge = __make_respBadge(bData, compositeRetCode)
                break
            end
    		if not cacheVal.reqType then
    			print("Couldn't detect request message type!")
    			rBadge = __make_respBadge(bData, compositeRetCode)
    			break
    		end
    		if isRequireGuest and not o.roomProvider.getRoomGuest(roomIdent) then
    		   print("Guest not found for room %q!", roomIdent)
    		   rBadge = __make_respBadge(bData, compositeRetCode)
    		   break
		    end
    		
    		if "SET" == cacheVal.reqType then	-- processing response on "Set guest data" message
    			-- check already present badgeSerial
        		local badgeInfo = o.dataProvider.getBadgeBySn(bData.serial)
        		if badgeInfo then
        		    -- update currently present badge by new information
        		    print("Badge with serial number %q has already present for room %q and will be updated!", bData.serial, badgeInfo.rid or "")
        		    o.dataProvider.delBadge(bData.serial)
        		end
        		
        		if "OK" == compositeRetCode then	-- processing by composite code
            		-- add badge to hotel database
            		local data = {	
            			 serial = bData.serial
            			,startv = bData.startv	 
            			,endv	= bData.endv	
            			,type   = cacheVal.badgeType
            			,common = roomInfoList.common
            			,suite  = roomInfoList.suite
            			,rid	= roomIdent
            		}
            		o.dataProvider.setBadge(data)
            	end
            	print("Badge create event [%s] has composite code - %q", bData.serial or "NULL", compositeRetCode)
            	
            	-- sending answer message to FIAS from got return code
            	rData = {
            		 keyData  = {
            		      kCoder  = cacheVal.keyCoder 
            		     ,wsId    = cacheVal.ws
            		     ,trackID = "OK" == compositeRetCode and bData.serial  
        		     }
            		,status   = _keyRetCodes["CREATED"][compositeRetCode]
            		,cacheUID = cacheUID
            	}
    		end
		
			print("Event badges created (%s) has been processed successfully for room %q!", cacheVal.reqType, roomIdent)
			rv = true
		until true
		
		return rv, rData, rBadge
	end
	
	--- Processing XOPEN event of a deleted badges
	-- @param bData		A badges data
	-- @return boolean value and answered data
	o.badgesDeleted = function(bData)
		if "table" ~= type(bData) then
			printdbg("Invalid badges data")
			return false
		end
		
		if "table" ~= type(bData.badges) or 0 == #bData.badges then
			print("Badges event does not contain mandatory fields")
			return false
		end
		
		-- get badges info from hotel storage
		local badgesInfoList = {}
		for _, badge in pairs(bData.badges) do
			local badgeInfo = o.dataProvider.getBadgeBySn(badge.sn)
			if badgeInfo then
				table.insert(badgesInfoList, badgeInfo)
			end
		end
		if 0 == #badgesInfoList then
			print("Couldn't got badges info for badge serials in the event. It can be sync procedure/internal badge data command!")
			return false
		end
		
		-- check content of anyone badge
		local oneBadgeInfo = badgesInfoList[1]
		if not oneBadgeInfo.serial or not oneBadgeInfo.rid then
			print("Invalid parameters of the badges info!")
			return false
		end
		
		-- processing cached key request or simple badge delete
		local rData    = nil
		local cacheUID = "DL"..(oneBadgeInfo.rid or "")
		local cacheVal = o.dataProvider.getKeyCache(cacheUID) 
		if cacheVal then 
			local badgeCount = cacheVal.badgeCnt
			badgeCount 		 = badgeCount - 1
			if badgeCount <= 0 then
				rData = {
					 keyData  = {kCoder = cacheVal.keyCoder ,wsId = cacheVal.ws}
					,status   = _keyRetCodes["DELETED"]["OK"]
					,cacheUID = cacheUID
				}
			else
				cacheVal.badgeCnt = badgeCount
				o.dataProvider.setKeyCacheValueOnly(cacheUID, cacheVal)
			end
		end
		
		-- removing badges from hotel storage
		for _, badgeInfo in pairs(badgesInfoList) do
			if badgeInfo then
				o.dataProvider.delBadge(badgeInfo.serial)
			end
		end
		
		print("Event badges deleted has been processed successfully for room %q!", oneBadgeInfo.rid)
		return true, rData
	end
	
	--- Sending badge response to XOPEN (like badge delete)
	-- @param bData		A badges data
	-- @return boolean value 
	o.badgeResponse = function(bData)
		if "table" ~= type(bData) then
			printdbg("Invalid badges data")
			return false
		end
		
		local data    = bData.guestData
		local wsInfo  = bData.ws
		if "table" ~= type(data) or "table" ~= type(wsInfo) then
			printdbg("Invalid data for badge response")
			return false
		end 
		
		local rv, err = __send_guestData(data, "DEL", {f = wsInfo.fc, u = wsInfo.uc})
		if not rv then printdbg(err) end
		
		return rv
	end
	
	--- Processing badges array during sync procedure
	-- @param bList		       A list with all badges info
	-- @param isRequireGuest   An flag for additional room state checking
	-- @return boolean value
	o.badgeListProcessing = function(bList, isRequireGuest)
		if "table" ~= type(bList) then
			printdbg("Invalid badges list data !!!")
			return false
		end
		
		-- retrieve random workstation ID
		local ws = o.roomProvider.getRandomWSMapping()
		if "table" ~= type(ws) or not ws.fc or not ws.uc then
			print("Couldn't retrieve workstation data!")
			return false
		else
			print("Sync is using workstation with data - %s/%s", ws.fc, ws.uc)
		end
		
		-- processing badges list
		local totalBadges = #bList
		local syncBadges  = 0
		local delBadges   = 0
		for idx, bData in ipairs(bList) do
            if bData and 0 == bData.type then -- processing only Guest badges type  
    			bData.workstation = ws						-- extend badge data by workstation ID
				-- executing handling event for badge created
				local rs, _, rBadge = o.badgesCreated(bData, isRequireGuest, true)
				if rs then
					syncBadges = syncBadges + 1
				elseif rBadge then
					if o.badgeResponse(rBadge) then
						delBadges = delBadges + 1
					end
				end
			end
		end
		
		local unknownBadges = totalBadges - syncBadges - delBadges
		printdbg("Badge sync report: %d - synced, %d - deleted, %d - unknown!", syncBadges, delBadges, unknownBadges)
	end
end

return GuestDataController
