--- Posting controller for processing call and MiniBar CDR records
-- 
-- @file	PostingSimpleController.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/06/22

--- @module PostingSimpleController
local PostingSimpleController = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
-- Send posting simple data to FIAS
local function __send_postSimple(psData)
	if not psData then
		return false
	end
	
	local psAnsPkt, err = require"DataModel":new({ident = "FIAS"})
	if not psAnsPkt then
		return false, err
	end
	
	psAnsPkt:setItem("data.ident", 30)	-- PS message
	psAnsPkt:addRepeatedTuple("data.fields", "RN", psData.roomNumber)
	psAnsPkt:addRepeatedTuple("data.fields", "P#", psData.id)
	psAnsPkt:addRepeatedTuple("data.fields", "DA", psData.date)
	psAnsPkt:addRepeatedTuple("data.fields", "TI", psData.time)
	psAnsPkt:addRepeatedTuple("data.fields", "PT", psData.type)
	psAnsPkt:addRepeatedTuple("data.fields", "DD", psData.dst)
	psAnsPkt:addRepeatedTuple("data.fields", "DU", psData.duration)
	psAnsPkt:addRepeatedTuple("data.fields", "TA", psData.cost)
	psAnsPkt:addRepeatedTuple("data.fields", "MA", psData.article)
	psAnsPkt:addRepeatedTuple("data.fields", "M#", psData.items)
	psAnsPkt:addRepeatedTuple("data.fields", "SO", psData.so)
	
	return require"appHelper":startTransaction(psAnsPkt)
end
----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
PostingSimpleController._instance = nil -- Singletone instance

function PostingSimpleController:getInstance(dataProvider, roomProvider)
	if self._instance then
		return self._instance
	end
	if not dataProvider or not roomProvider then
		return false
	end
	
	self._instance = PostingSimpleController 
	local o = self._instance
	
	o.dataProvider = dataProvider
	o.roomProvider = roomProvider

	--- Validation and processing data from CDR
	-- @param psData	A data of the Post Simple message
	-- @param SYS		Environment of the Manager variables
	-- @return postSimple data with FIAS structure
	o.processPSData = function(psData, SYS)
		if "table" ~= type(psData) then
			return false
		end
		
		local rs   = false
		local data = {} -- sending to FIAS data
		repeat
    		if not psData.roomIdent or not psData.type or not psData.startv then
    			print("Not found mandatory parameters in PS!")
    			break
    		end
    		data.id = psData.id
    		
    		-- get room information
    		local roomInfo = o.roomProvider.getRoom(psData.roomIdent, "PBX")
    		if not roomInfo then
    			print("Not found room %q for room equipment request!", psData.roomIdent)
    			break
    		end
    		data.roomNumber = roomInfo.number
    		
    		-- convertind data and time
    		local date, time = require"dateConverter".string2fias(psData.startv)
    		if "000000" == date or "000000" == time then
    			print("Error conversion of the date value!")
    			break
    		end
    		data.date = date
    		data.time = time
    		
    		if 1 == psData.type then		-- Call PS
    			if not SYS then
    				print("Not found billing variables")
    				break
    			end
    			
    			if not psData.dst or "" == psData.dst then
    				print("Destination parameter is not found in Call PS!")
    				break
    			end
    			
    			-- processing call data
    			local duration    = require"dateConverter".second2fias(psData.duration)
    			local callCost 	  = psData.cost or ""
    			local billingType = o.dataProvider.getBillingType()
    			if SYS.M_BILLING_DU == billingType then
    				if "000000" == duration then
    					print("Zero duration in call PS for current billing type!")
    					break
            		end
            		callCost = ""	-- clear CDR Cost by using only Duration flag
            	end
            	
    			if SYS.M_BILLING_DU_COST == billingType and ("0.00" == callCost or "" == callCost) then
            		print("Zero cost in call PS for current billing type!")
            		break
            	end
            	
            	-- set post simple type and appropriate variables
            	local psType = "T"	-- Telephone charge (Costing make by PMS)
            	if "" == callCost then
            		if 0 == duration then
            			print("Null duration in call PS with type 'telephone charge (T)'")
            			break
            		end
            		callCost = nil
            	else
            		psType 	 = "C"	-- Direct charge (must contain TA. Consting make by other system)
            		callCost = callCost:gsub("%.", "")	-- remove decimal point
            	end
            	
            	data.type 		= psType
            	data.dst  		= psData.dst
            	data.duration 	= duration
            	data.cost		= callCost
            	
            	rs = true
            	print("Procsssing Call posting from room %q to destination number %q", roomInfo.ident, data.dst)
    		elseif 2 == psData.type	then	-- MiniBar PS
    			psData.article = tonumber(psData.article)
    			psData.items   = tonumber(psData.items)
    			if not psData.article then
    				print("Not found mandatory paramters in MiniBar PS!")
    				break
    			end
				
				-- fix up some PS data records
				data.duration = "NULL"
    			if not psData.items or 0 > psData.items then
    				psData.items = 1
    			end
    			
    			local minibarType = o.dataProvider.getMinibarType()
    			if SYS.M_MINIBAR_ARTICLE == minibarType then 
    				data.type 	 = "M"
    				data.article = psData.article
    				data.items 	 = psData.items
    				print("Processing MiniBar posting from room %q with article %q (#%d)", roomInfo.ident, data.article, data.items)
    			else
    				data.type = "C"
    				data.cost = psData.article * psData.items
    				print("Processing MiniBar posting from room %q with total amount %q", roomInfo.ident, data.cost)	
    			end
    			
    			rs = true
    		else
    			print("Wrong PS data type!")
    		end
    	until true
    	
		return rs and data		
	end
	
	--- Create Posting Simple from data in the 'Mem' storage and send to FIAS
	-- @param psData	A data of the Post Simple message
	-- @return boolean value
	o.createPS = function(psData)
		if "table" ~= type(psData) then
			printdbg("Invalid data for creating PS!")
			return false
		end
		
		local rs   = false
		repeat
    		-- check active simple in the cache
    		local cachePS = o.dataProvider.getCachePS() 
    		if cachePS and ("DEL" == cachePS.state or "SENT" == cachePS.state) then
    			printdbg("Found PS in processing state")
    			rs = true
    			break
    		end
    		
    		if not psData.type or "" == psData.type then
    			print("Unknown type of the PS message!")
    		end
    		if not psData.id or not psData.roomNumber or not psData.id or not psData.date or not psData.time then
    			print("Not found mandatory parameters in PS!")
    			break
    		end
    		
    		-- update cached value
    		psData.pulse 	= 0 ~= psData.items and psData.pulse
    		psData.article  = 0 ~= psData.article and psData.article
    		psData.items 	= 0 ~= psData.items and psData.items
    		if "C" == psData.type then
    			psData.so = "NULL" == psData.duration and "2" or "1" -- SalesOutlet: 1 - phone call, 2 - minibar 
    		end
    		if "" == psData.duration then
    			psData.duration = "000000"
    		elseif "NULL" == psData.duration then
    			psData.duration = ""
    		end

    		-- sending PS message to FIAS interface
    		if __send_postSimple(psData) then
    			-- set tag for cached PS
    			o.dataProvider.setCachePS(psData.id, psData.roomNumber, "SENT")
    			rs = true 
    		end
    	until true
    	
    	if not rs then
    		o.dataProvider.setCachePS(psData.id, psData.roomNumber, "DEL")
    	end
		
		return rs 
	end
	
	--- Remove Posting Simple by some reason
	-- @param psData	A data of the Post Answer message
	-- @return boolean value
	o.markPS = function(psData)
		if "table" ~= type(psData) then
			printdbg("Invalid data for removing PS!")
			return false
		end
		
		if not psData.roomIdent or not psData.id or not psData.state then
			print("Not found mandatory data in PA!")
			return false
		end
		
		local cachePS = o.dataProvider.getCachePS()
		if not cachePS or cachePS.id ~= psData.id or cachePS.rn ~= psData.roomIdent then
			print("Got post simple not found in the cache")
			return false
		end
		
		--[[ proceed answer status
    		CO - Posting denied because overwriting the CreditLimit is not allowed (not used with Suite8)
    		DM - Sum of subtotals doesn't match TotalAmount
    		NA - Night Audit
    		NP - Posting denied for this guest (NoPost flag has been set)
    		OK - Command or request completed successfully
    		RY - Retry
    		UR - Unprocessable request, this request cannot be carried out , no retry
    	]]--
    	repeat
    	   if "RESEND" == psData.state then    -- set data chache as require update
               o.dataProvider.setCachePS(psData.id, psData.roomIdent, "RESEND")
               printdbg("Try to re-send Post Simple message for room %q and sequence number '%d' by timeout", psData.roomIdent, psData.id)
               break
           end
           if "OK" == psData.state or "TIMEOUT" == psData.state then   -- remove data from cache
               o.dataProvider.setCachePS(psData.id, psData.roomIdent, "DEL")
               print("Post Answer for room %q and sequence number '%d' was processed (%s)", psData.roomIdent, psData.id, psData.state)
               break
           end
           if "UR" == psData.state then
               o.dataProvider.setCachePS(psData.id, psData.roomIdent, "DEL")
               print("WARNING: unsupported posting message for room %q and sequence number '%d'!!!", psData.roomIdent, psData.id)
               break
           end
           
           print("Unknown status in PA - %s", psData.state)
           return false
    	until true

    	return true
	end
	
	--- Remove marked Post Simple from cache
	-- @return boolean value
	o.clearMarkedPS = function()
		local cachePS = o.dataProvider.getCachePS()
		if cachePS and "DEL" == cachePS.state then
			o.dataProvider.delCachePS()
			return true
		end
		
		return false
	end
	
	--- Clear all posting simple cache data
	o.clearAllPS = function()
		if o.dataProvider.getCachePSCounter() then
			o.dataProvider.delCachePSCounter()
			o.dataProvider.delCachePS()
		end
	end
	
	--- Updating Posting Simple data counter
	-- @return action for active Posting Simple message with PS data
	o.updatePSCounter = function()
		local cachePS = o.dataProvider.getCachePS()
		if cachePS then
			local cachePSCnt = o.dataProvider.getCachePSCounter()
			if cachePSCnt then
    			if cachePS.id == cachePSCnt.id then
    				cachePSCnt.cnt = cachePSCnt.cnt + 1
    			else
    				cachePSCnt = {cnt = 1}
    			end
    		else
    			cachePSCnt = {cnt = 1}
    		end
			o.dataProvider.setCachePSCounter(cachePSCnt.cnt, cachePS.id)
			
			if "SENT" == cachePS.state then
    			local data 		= {roomIdent = cachePS.rn, id = cachePS.id}
    			local dividerPS = require"timerHelper".fromTimeCounter(90)	-- time for resend PS message (90 sec)
    			local timeoutPS = dividerPS * 3                             	-- limit of the PS re-sending (3 attempt)
    			if 0 == cachePSCnt.cnt % dividerPS then
    			    if cachePSCnt.cnt >= timeoutPS then
    			        data.state = "TIMEOUT"
                        return data.state, data
                    else
                        data.state = "RESEND"
                        return data.state, data
    			    end
    			end
        	end
        else
        	return "NOTFOUND"
		end
	end
end

return PostingSimpleController
