--- Room equipment controller
-- 
-- @file	RoomEquipmentController.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/06/16

--- @module RoomEquipmentController
local RoomEquipmentController = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Build and send XOPEN message MaidStatus
local function __send_maidStatus(eqData, endPoint)
	if "table" ~= type(eqData) or "table" ~= type(endPoint) then
		return false, "Invalid arguments"
	end
	
	local type   = tonumber(eqData.type)
	local detail = tonumber(eqData.detail)
	if not eqData.aptid or not type or not detail then
		return false, "Not found mandatory parameters"
	end
	
	local dModel			= require"DataModel"
	local XOPEN				= dModel.XOPEN
	local eqDataPkt, err = dModel:new({ident = "XOPEN", endpoint = endPoint})
	if not eqDataPkt then
		return false, err
	end
	
	eqDataPkt:setItem("data.ident", XOPEN.X_MAID_STATUS)
	
	local room  = eqDataPkt:addRepeatedItem("data.maidStat.rooms")
	room.aptID  = eqData.aptid
	room.type   = tonumber(eqData.type)
	room.detail = tonumber(eqData.detail)
	
	return require"appHelper":startTransaction(eqDataPkt)
end

--- Build and send FIAS message with room equipment content
local function __send_roomEquipment(eqData)
	if "table" ~= type(eqData) then
		return false
	end
	
	local eqDataPkt, err = require"DataModel":new({ident = "FIAS"})
	if not eqDataPkt then
		return false, err
	end
	
	eqDataPkt:setItem("data.ident", 25)	-- RE message
	eqDataPkt:addRepeatedTuple("data.fields", "RN", eqData.rn)
	eqDataPkt:addRepeatedTuple("data.fields", "DN", eqData.DND)
	eqDataPkt:addRepeatedTuple("data.fields", "RS", eqData.MUR)
	
	return require"appHelper":startTransaction(eqDataPkt)
end

--- Build and send PBX messsage with originate to DND extension
-- @note We can set this feature via Originate or Command
local function __set_DNDfeature(exten, dndVal)
	if not dndVal or not exten then
		return false, "DND feature data is not specified!"
	end

    local rv = nil
    
    repeat
        -- try update data via user status module
        rv, err = require"wmsUser":setInfo(exten, {status = 1 == dndVal and "DND" or "AVAILABLE"})
        if rv then
            break
        end
    
        -- sending simple DND event to device
        local dModel 		 = require"DataModel"
        local PBX			 = dModel.PBX
        local eqDNDPkt, err = dModel:new({ident = "PBX", ct = PBX.P_COMMAND})
        if not eqDNDPkt then
            return false, err
        end

        local syncCommand = string.format("sip sync feature %s DND %s", exten, 1 == dndVal and "true" or "false")
        eqDNDPkt:addRepeatedTuple("opts.items", "Command", syncCommand)

        require"appHelper":startTransaction(eqDNDPkt)
    until true
	
	return rv
end

--- Internal XOpen equipment object
local xopenEqObj = { 
	 data = {
	 	 items 		= 0
	 	,room 		= {aptid = nil, fcode = nil, ucode = nil}
	 	,useXopen 	= false
	 }
	,setRoom = function(self, aptid, fcode, ucode)
		self.data.room.aptid = aptid
		self.data.room.fcode = fcode
		self.data.room.ucode = ucode
	end
	,set  = function(self, name, val)
		if name and val then
			self.data[name] = val
			self.data.items = self.data.items + 1
		end
	end
	,exec = function(self)
		if 0 < self.data.items and self.data.room.aptid then
			if self.data.useXopen then
				local endPoint = {f = self.data.room.fcode, u = self.data.room.ucode} 
				local data 	   = {aptid = self.data.room.aptid}
				if self.data.DND then
					data.strType = "DND"
					data.type    = 3			-- Evt DND
					data.detail  = self.data.DND
				elseif self.data.MUR then
					data.strType = "MUR"
					data.type    = 4			-- Evt MUR
					data.detail  = self.data.MUR
				end
				
				local rv, err = __send_maidStatus(data, endPoint) 
				if not rv then
					print(err)
				else
					print("Room equipment was sent to XOPEN: %s - %d", data.strType, data.detail)
				end
				return rv
			else
				printdbg("XOpen interface is not configured! Room equipment message will be dropped!")	
			end
		end
	end
}
		
--- Internal FIAS equipment object
local fiasEqObj = {
	 data = {
	 	 items 		= 0
	 	,rn 		= nil
	 	,roomState 	= nil
	 }
	,def = {
		DND = {
			 [1] = "Y"
			,[0] = "N"
		},
		MUR = {	-- list of the MUR statuses
			 [0] = "Unknown"
    		,[1] = "Dirty/Vacant"	     -- 1 + false
    		,[2] = "Dirty/Occupied"		 -- 1 + true
    		,[3] = "Clean/Vacant"		 -- 0 + false
    		,[4] = "Clean/Occupied"		 -- 0 + true
    		,[5] = "Inspected/Vacant"	 -- 2 + false
    		,[6] = "Inspected/Occupied"	 -- 2 + true
		}
	}
	,setRoom = function(self, rn, roomState)
		self.data.rn 		= rn
		self.data.roomState = roomState
	end
	,set  = function(self, name, val)
		if name and val then
			if "DND" == name then
				val = self.def.DND[tonumber(val) or 0] 
			elseif "MUR" == name then
				local absVal = math.abs(val)
            	local murVal = (0 == absVal and 3) or (2 == absVal and 5) or 1
            	local roomSt = 0 <= tonumber(val) and self.data.roomState and 1 or 0
            	val 		 = murVal + roomSt
            end
            if val then
				self.data[name] = val
				self.data.items = self.data.items + 1
			end
		end
	end
	,exec = function(self)
		if 0 < self.data.items and self.data.rn then
			local execItems = {}
			if self.data.DND then
				table.insert(execItems, string.format("DND - %s", self.data.DND))
			end
			if self.data.MUR then
				table.insert(execItems, string.format("MUR - %s", self.def.MUR[tonumber(self.data.MUR) or 0]))
			end
			
			local rv, err = __send_roomEquipment(self.data)
			if not rv then
				print(err)
			else 
				print("Room equipment was sent to FIAS: %s", table.concat(execItems, ", "))
			end
			return rv 
		end
	end
}

--- Internal PBX equipment object
-- @note limited by execution only for one item
local pbxEqObj = {
	 data = {
	 	 items 	= 0
	 	,exten 	= nil
	 }
	,setRoom = function(self, exten)
		self.data.exten = exten
	end
	,set  = function(self, name, val)
		if name and val then
			self.data[name] = val
			self.data.items = self.data.items + 1
		end
	end
	,exec = function(self)
		if 1 == self.data.items and self.data.exten then
			local rv, err = __set_DNDfeature(self.data.exten, self.data.DND)
			if not rv then
				print(err)
			else
				printdbg("DND sync command was sent to PBX(%s) with status %q", self.data.exten, self.data.DND)
			end
		elseif 1 < self.data.items then
			printdbg("We can't make PBX exec for more that one item")
		end
	end
}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
RoomEquipmentController._instance = nil -- Singletone instance

function RoomEquipmentController:getInstance(dataProvider, roomProvider)
	if self._instance then
		return self._instance
	end
	if not roomProvider then
		return false
	end
	
	self._instance = RoomEquipmentController 
	local o = self._instance
	
	o.dataProvider = dataProvider
	o.roomProvider = roomProvider
	
	--- init ACL values table
    local aclDefVal = {
    	 BLOCKED	= -1
    	,INTERNAL	= 0
    	,LOCAL		= 1
    	,NATIONAL	= 2
    	,ALL		= 3
    }
    
    local aclValues = {}
    for val, idx in pairs(aclDefVal) do
    	aclValues[val] = idx
    	aclValues[idx] = val
    end
	
	--- Set equipment for room
	-- @param eqData			A room equipment data
	-- @param SYS				Environment of the Manager variables
	-- @param useXOpenIface		A flag for disabling sending messages to XOPEN interface
	-- @return boolean value
	o.setEquipment = function(eqData, SYS, useXOpenIface)
		if "table" ~= type(eqData) then
			printdbg("Invalid room equipment data")
			return false
		end
		
		if not eqData.roomIdent or not eqData.ifn then
			printdbg("Not found room identification in the equipment data")
			return false
		end
		local srcIface = eqData.ifn
		
		-- get room info
		local roomInfo = o.roomProvider.getRoom(eqData.roomIdent, srcIface)
		if not roomInfo then
			print("Not found room %q for room equipment request!", eqData.roomIdent)
			return false
		end
		local roomIdent = roomInfo.ident
		local guestId   = o.roomProvider.getRoomGuest(roomIdent)
		if not guestId then
			printdbg("Room %q is empty. We cannot set CoS or/and DND values!", roomIdent)
		end
		
		-- init related objects
		xopenEqObj.data.useXopen = useXOpenIface
		xopenEqObj:setRoom(roomInfo.aptid, roomInfo.fcode, roomInfo.ucode)
		fiasEqObj:setRoom(roomInfo.number, guestId)
		pbxEqObj:setRoom(roomInfo.exten)
		
		-- process incoming data
		local data  = {}
		if eqData.ACL then
			if guestId then
    			data.ACL = {STORAGE = aclValues[tonumber(eqData.ACL) or -1]}
    			print("Try to set ACL for room %q to value %q", roomIdent, data.ACL.STORAGE)
    		end	
		end
		if eqData.DND then
			local dndValues    = {N = 0, Y = 1}	-- convert to storage format
			eqData.DND	  	   = dndValues[eqData.DND] or tonumber(eqData.DND)
			if guestId then
				local roomDNDState = o.roomProvider.getRoomDND(roomIdent)
				local DNDBehaviour = o.dataProvider.getDNDBehaviour()
				if "PBX" == srcIface and roomDNDState == eqData.DND then
					-- do not send equipment in case we got event from PBX with the same DND status
					printdbg("Do not processing this DND event as setting to the same value!")
				elseif SYS and DNDBehaviour == SYS.M_DND_STD and ("XOPEN" == srcIface or "PBX" == srcIface) then
					printdbg("Do not processing DND room state by DND behaviour settings!")
				else
					data.DND = {STORAGE = eqData.DND, XOPEN = eqData.DND, FIAS = eqData.DND, PBX = eqData.DND}
					print("Try to set DND for room %q to value %q", roomIdent, eqData.DND)
        		end
    		end
		end
		if eqData.MUR then
			local murValues = {N = 0, Y = 1}	-- convert to storage format
			eqData.MUR 		= murValues[eqData.MUR] or tonumber(eqData.MUR)
			local absMUR    = eqData.MUR and math.abs(eqData.MUR)	-- module from MUR value
			data.MUR 		= {STORAGE = eqData.MUR, FIAS = eqData.MUR, XOPEN = 2 ~= absMUR and absMUR}
			print("Try to set MUR for room %q to value %q", roomIdent, eqData.MUR)
		end
		
		-- save to data to storage and sending to appropriate interface
		for eqName, val in pairs(data) do
			if o.roomProvider["setRoom"..eqName] then
				o.roomProvider["setRoom"..eqName](roomIdent, val.STORAGE, "PBX" == srcIface)
			end
			if "XOPEN" ~= srcIface then
				xopenEqObj:set(eqName, val.XOPEN)
			end
			if "FIAS" ~= srcIface then
				fiasEqObj:set(eqName, val.FIAS)
			end
			if "PBX" ~= srcIface then
				pbxEqObj:set(eqName, val.PBX)
			end
		end
		
		-- sending prepared data
		xopenEqObj:exec()
		fiasEqObj:exec()
		pbxEqObj:exec()
		
		return true
	end
	
	--- Get room equipment
	-- @param roomIdent		A room ident
	-- @return table with room equipment status
	-- @note MUR attribute exception as individual room
	o.getEquipment = function(roomIdent)
		if not roomIdent then
			return {}
		end
		
		return {
			 DND = o.roomProvider.getRoomDND(roomIdent)
			,ACL = aclValues[o.roomProvider.getRoomACL(roomIdent)] or -1
		}
	end
	
	--- Remove all room equipment from "mem" and "ast" storage (exeption MUR value)
	-- @param roomInfo		A room information
	-- @return boolean value
	-- @note This function also remove all WakeUp records from "ast" storage 
	o.delEquipment = function(roomInfo, useXOpenIface)
		if "table" ~=  type(roomInfo) then
			return false
		end
		
		-- retrieve current values of the room equipment
		local dndVal = o.roomProvider.getRoomDND(roomInfo.ident)
		local murVal = o.roomProvider.getRoomMUR(roomInfo.ident)
		
		-- remove all room equipment
		o.roomProvider.delRoomData(roomInfo.ident)
		
		-- set saved MUR value
		o.roomProvider.setRoomMUR(roomInfo.ident, murVal)
		
		-- notify related interfaces about event (same as from FIAS iface)
		if 0 ~= dndVal then
            local eqData = {DND = {PBX = 0, XOPEN = 0}}
            
            xopenEqObj.data.useXopen = useXOpenIface
            xopenEqObj:setRoom(roomInfo.aptid, roomInfo.fcode, roomInfo.ucode)
            pbxEqObj:setRoom(roomInfo.exten)
            for eqName, val in pairs(eqData) do
                xopenEqObj:set(eqName, val.XOPEN)
                pbxEqObj:set(eqName, val.PBX)
            end
            xopenEqObj:exec()
            pbxEqObj:exec()
        end
		
		return true
	end
	
	--- Sync equipment for rooms from Hotel storage to Mem storage
	-- @param roomsList		A list of the rooms idents that must be synced
	-- @return boolean value
	o.syncEquipment = function(roomsList)
		if "table" ~= type(roomsList) then
			return false
		end
		
		local syncRoomCnt = 0
		for _, roomIdent in ipairs(roomsList) do
			local syncRes = 0
			syncRes = syncRes + o.roomProvider.syncRoomACL(roomIdent)
			syncRes = syncRes + o.roomProvider.syncRoomDND(roomIdent)
			syncRes = syncRes + o.roomProvider.syncRoomMUR(roomIdent)
			if 0 < syncRes then
				syncRoomCnt = 1 + syncRoomCnt
			end 
		end
		printdbg("Synced equipment for %d rooms", syncRoomCnt)
		
		return true
	end
end

return RoomEquipmentController
