--- add-on for simplify access to room database 
-- 
-- @file	RoomProvider.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/21

----------------------------------------------------------------
-- public interface
----------------------------------------------------------------
local RoomProvider = {
	_instance  = nil,
}

function RoomProvider:getInstance(storage)
    if self._instance then
    	return self._instance
    end
	if not storage then
		return nil
	end

	self._instance  = RoomProvider
	local o         = self._instance
    	  o.storage = storage

	----------------------------------------------------------------
	--- Internal helpers
	----------------------------------------------------------------
    local function __setTempByKey(k, v)
    	return o.storage:putTemp(k, v)
    end
    local function __getTempByKey(k)
    	return o.storage:getTemp(k)
    end
    local function __delTempByKey(k)
    	return o.storage:delTemp(k)
    end
    
    local function __setAstByKey(k, v)
    	return o.storage:put("ast", k, v)
    end
    local function __getAstByKey(k)
    	return o.storage:get("ast", k)
    end
    local function __delAstByKey(k)
    	return o.storage:del("ast", k)
    end
    local function __delTreeAstByKey(k)
    	return o.storage:del("ast", {ident="tree", val=k})
    end

	--- place new item to database
    local function __pushItem(t, r)
    	if "table" ~= type(r) then
    		return false, "Invalid input argumetns"
    	end
    	
    	local number = r["rn"]
		local exten  = r["extension"]
		local desc   = r["descr"] or ""
		local aptid  = string.match(r["id"] or "", "^{?([^{}]*)}?$")
		local fc     = r["FC"] or "0"
		local uc     = r["UC"] or "0"
		
		-- input parameters validation
		local emptyMandElmts = {}
		if not number then
			table.insert(emptyMandElmts, "RoomNumber")
		end
		if not exten then
			table.insert(emptyMandElmts, "Extension")
		end
		if "0" == fc or "0" == uc or not aptid then
			table.insert(emptyMandElmts, "AptId or Options")
		end
		if 0 ~= #emptyMandElmts then
			local errMsg = string.format("Not found mandatory parameters for %q: (%s)", number or aptid or exten or desc, table.concat(emptyMandElmts, ","))
			return false, errMsg
		end
		
		local ident  = exten
		local tag    = "ROOM"
		if "ca" == t then
			ident 	 = number
			tag 	 = "CA"
		end
		
		if "" == ident then
			return false, "Ident does not found"
		end
		desc = require"strConverter".toHex(desc)	-- encoding description text
		
		local areaInfo = string.format("%d|%s!%s!%s!%d!%d!%s", ident, number, exten, aptid, fc, uc, desc)
		__setTempByKey(tag.."/CONF/"..(ident or "0"), areaInfo)

		__setTempByKey(tag.."/IDENT/FIAS/"..(number or "0"), tostring(ident))
		__setTempByKey(tag.."/IDENT/PBX/"..(exten or "0"), tostring(ident))
		__setTempByKey(tag.."/IDENT/XOPEN/"..(aptid or "0"), tostring(ident))
		return ident
    end
    
    --- get item from database
    local function __getItem(t, ident, ifn)
    	if not ident then
    		return nil
    	end
    	
    	if "ca" ~= t and "room" ~= t then
    		return nil
    	end
    	local tag = "ca" == t and "CA" or "ROOM"
    	
    	local id  = ident
    	if ifn and "SYS" ~= ifn then	-- in case request from manager we always use room ident value
    		id = __getTempByKey(tag.."/IDENT/"..string.upper(ifn).."/"..ident)
    	end
    	local ret = id and __getTempByKey(tag.."/CONF/"..id) 
       	if not ret then
       		return nil
       	end
       	
		local rv = {["ident"] = id}
		local cid, rn, exten, appid, fcode, ucode, desc = string.match(ret, "^(%d+)|(.*)!(.*)!(.*)!(%d+)!(%d+)!(.*)$")
		if not cid or tonumber(cid) ~= tonumber(id) then
			print("!!!! integrity constraint broken while room loading!!!")
			return nil
		end
		
		rv["number"] = rn
		rv["exten"]  = exten
		rv["aptid"]  = appid
		rv["fcode"]  = tonumber(fcode)
		rv["ucode"]  = tonumber(ucode)
		rv["desc"]   = require"strConverter".fromHex(desc)
		return rv
    end
    
    --- delete item from database
    local function __delItem(t, ident, ifn)
    	if not ident or not ifn then
    		return nil
    	end
    	
    	if "ca" ~= t and "room" ~= t then
    		return nil
    	end
    	local tag = "ca" == t and "CA" or "ROOM"
    	
    	local item = __getItem(t, ident, ifn)
    	if not item then
    		return false
    	end
    	
    	__delTempByKey(tag.."/CONF/"..item["ident"])

		__delTempByKey(tag.."/IDENT/FIAS/"..item["rn"])
		__delTempByKey(tag.."/IDENT/PBX/"..item["exten"])
		__delTempByKey(tag.."/IDENT/XOPEN/"..item["aptid"])
    end
    
    --- save virtual room data to the memory
    local function __pushVirtualRoom(r)
        if "table" ~= type(r) then
            return false, "Invalid arguments"
        end
        
        local number = r.number
        local suite  = {}
        for key, val in pairs(r) do
            if key:find("^suite") and val and val.room then
                table.insert(suite, val.room)
            end
        end
        
        if not number then
            return false, "Not found virtual room number"
        end
        if 0 == #suite then
            return false, string.format("Virtual number %q has empty suite rooms list", number)
        end
        
        local virtualRoomData = string.format("%s|%s", number, table.concat(suite, "!"))
        __setTempByKey("VROOM/CONF/".. number, virtualRoomData)
        
        return number
    end
    
    --- retrieve virtual room data
    local function __getVirtualRoom(num)
        if not num then
            return nil
        end
        
        local rv = __getTempByKey("VROOM/CONF/" .. num)
        if not rv then
            return nil
        end
        
        local vrid, suiteStr = rv:match("^(%w+)|([^|]+)$") 
        if not vrid or vrid ~= num then
            print("!!!! integrity constraint broken while room loading!!!")
            return nil
        end
        
        local suite = {}
        if suiteStr then
            for val in suiteStr:gmatch("(%w+)", "!") do
                suite[val]        = val
                suite[#suite + 1] = val
            end
        end
        
        return {
             number = vrid
            ,suite  = suite
        } 
    end
    
    --- clear virtual room data from the memory
    local function __delVirtualRoom(num)
        if not num then
            return nil
        end
        
        return __delTempByKey("VROOM/CONF/" .. num)
    end
    
    -- save workstation info to memory
    local function __pushWSMap(m)
    	if "table" ~= type(m) then
    		return 0, "Invalid input workstation data type"
    	end
    	
    	local ws = m["ws"] or ""
		local fc = m["fc"] or ""
		local uc = m["uc"] or ""
		local ip = m["ip"] or ""
		
		-- check workstation info
		local ws = m["ws"] or ""
		local fc = m["fc"] or ""
		local uc = m["uc"] or ""
		local ip = m["ip"] or ""

		local mandp = {}	
		if "" == ws then table.insert(mandp, "WSName")	end
		if "" == fc then table.insert(mandp, "WSFCode")	end
		if "" == uc then table.insert(mandp, "WSUCode")	end
		if 0 ~= #mandp then
			local errMsg = string.format("Not found mandatory parameters for workstation %q: (%s)", ws or m["description"] or "", table.concat(mandp, ","))
			return 0, errMsg 
		end
		
    	__setTempByKey("MAPPING/WS/"..(ws or ""), string.format("%s|%d!%d!%s",ws,fc,uc,ip))
		__setTempByKey(string.format("MAPPING/FCUC/%d!%d",fc,uc), string.format("%s|%d!%d!%s",ws,fc,uc,ip))
		return 1
    end
    
    -- retrieve info about workstation
    local function __getWSMap(v1, v2)
    	if not v1 then 
			return nil
		end

		local loc = nil
		if "string" == type(v1) then
			loc = string.format("WS/%s",v1)
		elseif "number" == type(v1) and v2 then
			loc = string.format("FCUC/%d!%d",v1,v2)
		end

       	local ret = __getTempByKey("MAPPING/"..loc)
       	if not ret then 
       		return nil
       	end

		local rv = {}
		local ws, fc, uc, ip = string.match(ret,"^(.*)|(%d+)!(%d+)!(.*)$")
		rv["ws"] = ws
		rv["fc"] = tonumber(fc)
		rv["uc"] = tonumber(uc)
		rv["ip"] = ip or ""
		return rv
    end
    
    -- delete workstation info
    local function __delWSMap(v1, v2)
    	if not v1 then 
			return nil
		end

		local loc = nil
		if "string" == type(v1) then
			loc = string.format("WS/%s",v1)
		elseif "number" == type(v1) and v2 then
			loc = string.format("FCUC/%d!%d",v1,v2)
		end
		
		return __delTempByKey("MAPPING/"..loc)
    end

	----------------------------------------------------------------
	--- In-memory room database
	---------------------------------------------------------------
    o.pushCommArea = function(r)
    	return __pushItem("ca", r)
    end
    
    o.getCommonArea = function(ident, ifn)
    	return __getItem("ca", ident, ifn)
    end
    
    o.delCommonArea = function(ident, ifn)
    	return __delItem("ca", ident, ifn)
    end
    
    o.pushRoom = function(r)
    	return __pushItem("room", r)
    end

    o.getRoom = function(ident, ifn)
    	return __getItem("room", ident, ifn)
    end

    o.delRoom = function(ident, ifn)
    	return __delItem("room", ident, ifn)
    end
    
    o.saveRoomsList = function(list)
    	if "table" ~= type(list) or 0 == #list then
    		return false
    	end
    	
    	__setTempByKey("ROOMS-LIST", table.concat(list, "!"))
    end
    
    o.getRoomsList = function()
    	local rv   = {}
    	local list = __getTempByKey("ROOMS-LIST")
    	if "string" == type(list) then
    		for roomIdent in list:gmatch("(%w+)!?") do
    			table.insert(rv, roomIdent)
    		end
    	end
    	
    	return 0 ~= #rv and rv or nil
    end
    
    o.delRoomsList = function()
    	return __delTempByKey("ROOMS-LIST")
    end
    
    ---------------------------------------------------------------
    --- Virtual rooms 
    ---------------------------------------------------------------
    o.pushVirtualRoom = function(r)
        return __pushVirtualRoom(r)
    end
    
    o.getVirtualRoom = function(r)
        return __getVirtualRoom(r)
    end
    
    o.delVirtualRoom = function(r)
        return __delVirtualRoom(r)
    end
    
    o.saveVirtualRoomsList = function(list)
        if "table" ~= type(list) or 0 == #list then
            return false
        end
        
        __setTempByKey("VIRTUALROOMS-LIST", table.concat(list, "!"))
    end
    
    o.getVirtualRoomsList = function()
        local rv   = {}
        local list = __getTempByKey("VIRTUALROOMS-LIST")
        if "string" == type(list) then
            for suiteIdent in list:gmatch("(%w+)!?") do
                table.insert(rv, suiteIdent)
            end
        end
        
        return 0 ~= #rv and rv or nil
    end
    
    o.delVirtualRoomsList = function()
        return __delTempByKey("VIRTUALROOMS-LIST")
    end
    
    ---------------------------------------------------------------
    --- Room attributes (ACL, DND, MUR, Guests)
	---------------------------------------------------------------
	o.setRoomGuest = function(rid, v)
	    if rid and v then
		    return __setTempByKey("ROOM/"..(rid or "000") .. "/ATTR/GUEST", v)
	    end
	end
	o.getRoomGuest = function(rid)
	    if rid then
		    return __getTempByKey("ROOM/"..(rid or "000") .. "/ATTR/GUEST")
	    end
	end
	o.delRoomGuest = function(rid)
	    if rid then
		   return __delTempByKey("ROOM/"..(rid or "000") .. "/ATTR/GUEST")
	    end
	end

    o.setRoomMUR = function(rid, v, onlyCache)
    	if v then
        	__setTempByKey("ROOM/"..(rid or "000") .. "/ATTR/MUR", v)
        	if not onlyCache then
    			__setAstByKey((rid or "000").."/MUR", v)
    		end
    	end
    end
    o.getRoomMUR = function(rid)
       	return __getTempByKey("ROOM/"..(rid or "000") .. "/ATTR/MUR") 
    end
    o.delRoomMUR = function(rid)
    	__delTempByKey("ROOM/"..(rid or "000") .. "/ATTR/MUR")
       	__delAstByKey((rid or "000").."/MUR")
    end
    o.syncRoomMUR = function(rid)
    	local v = __getAstByKey((rid or "000").."/MUR")
    	if v then
    		__setTempByKey("ROOM/"..(rid or "000") .. "/ATTR/MUR", v)
    		return 1
    	end
    	return 0
    end
    
    o.setRoomDND = function(rid, v)
    	if v then
        	__setTempByKey("ROOM/"..(rid or "000") .. "/ATTR/DND", v)
   			__setAstByKey((rid or "000").."/DND", v)
    	end
    end
    o.getRoomDND = function(rid)
    	local val = __getTempByKey("ROOM/"..(rid or "000") .. "/ATTR/DND") 
       	return tonumber(val) or 0
    end
    o.delRoomDND = function(rid)
    	__delTempByKey("ROOM/"..(rid or "000") .. "/ATTR/DND")
       	__delAstByKey((rid or "000").."/DND")
    end
    o.syncRoomDND = function(rid)
    	local v = __getAstByKey((rid or "000").."/DND")
    	if v then
    		__setTempByKey("ROOM/"..(rid or "000") .. "/ATTR/DND", v)
    		return 1
    	end
    	return 0
    end
    
    o.setRoomACL = function(rid, v, onlyCache)
    	if v then
        	__setTempByKey("ROOM/"..(rid or "000") .. "/ATTR/ACL", v)
        	if not onlyCache then
    			__setAstByKey((rid or "000").."/ACL", v)
    		end
    	end
    end
    o.getRoomACL = function(rid)
       	return __getTempByKey("ROOM/"..(rid or "000") .. "/ATTR/ACL") 
    end
    o.delRoomACL = function(rid)
    	__delTempByKey("ROOM/"..(rid or "000") .. "/ATTR/ACL")
       	__delAstByKey((rid or "000").."/ACL")
    end
    o.syncRoomACL = function(rid)
    	local v = __getAstByKey((rid or "000").."/ACL")
    	if v then
    		__setTempByKey("ROOM/"..(rid or "000") .. "/ATTR/ACL", v)
    		return 1
    	end
    	return 0
    end
    
    -- Remove all room data with exception: not remove ATTR/GUEST and WK from memcached
    o.delRoomData = function(rid)
    	__delTempByKey("ROOM/"..(rid or "000") .. "/ATTR/MUR")
    	__delTempByKey("ROOM/"..(rid or "000") .. "/ATTR/DND")
    	__delTempByKey("ROOM/"..(rid or "000") .. "/ATTR/ACL")
    	__delTreeAstByKey(rid or "000")
    end

	----------------------------------------------------------------
	--- In-memory workstations mapping table
	----------------------------------------------------------------
	o.getMapping = function(v1)
		return __getWSMap(v1)
    end
    
    o.pushMapping = function(m)
		-- update database
    	return __pushWSMap(m)
    end

	o.getGlobalWSMapping = function()
		return __getWSMap("HotelServerMappingData")
	end
	
	o.pushGlobalWSMapping = function(m)
		if "table" ~= type(m) or not m["FC"] or not m["UC"] then
			return 0
		end
		
		local globalWSMap = {
			 ws = "HotelServerMappingData"
			,fc = m["FC"]
			,uc = m["UC"]
		}
		
		return __pushWSMap(globalWSMap)
	end
	
	o.getRandomWSMapping = function()
		return __getWSMap("HotelRandomClientMappingData")
	end
	
	o.pushRandomWSMapping = function(m)
		if "table" ~= type(m) then
			return 0
		end
		
		if o.getRandomWSMapping() then
			return 0
		end
		
		-- save random as first workstation
		m["ws"] = "HotelRandomClientMappingData"
		return __pushWSMap(m) 
	end
	
	o.delRandomWSMapping = function()
		return __delWSMap("HotelRandomClientMappingData")
	end
	
    return o
end

return RoomProvider
