--- WakeUp controller
-- 
-- @file	WakeUpController.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/05/26

--- @module WakeUpController
local WakeUpController = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- WakeUp answer codes:
	--	BY - Telephone / Encoder Busy
	--	DE - Wakeup/Key has been deleted
	--	NA - Night Audit
	--	NR - No Response
	--	OK - Command or request completed successfully
	--	RY - Retry
	--	SV - Wakeup has been sent to external system
	--	UR - Unprocessable request, this request cannot be carried out, no retry
local _WKRetCodes = {
	["CREATE"] = {
    	NO_ROOM			= "UR",
    	EMPTY_ROOM 		= "UR",
    	INVALID_TIME	= "UR",
    	FOUND_ITEM		= "UR",
    },
    ["DELETE"] = {
    	ANSWERED		= "OK",
    	TIMEOUT			= "NR"
    }
}

--- Update wakeUp map
local function __update_wkMap(provider, act, rid, wkId)
	if not rid or not provider then
		return false
	end
	
	local wkMap = {}
	if "add" == act then
    	wkMap.old = provider.getWakeUpMap(rid) or ""
    	if "" == wkMap.old then
    		wkId  	  = "1"
    		wkMap.new = wkId.."!"
    	else
    		wkId  = wkMap.old:match("(%d+)!$")
    		if not wkId then
    			return false
    		end
    		wkId 	  = wkId + 1
    		wkMap.new = wkMap.old..wkId.."!"
    	end
	elseif "del" == act then
		if not wkId then
			return false
		end
		wkMap.old = provider.getWakeUpMap(rid) or ""
		if not wkMap.old:find(wkId.."!") then
			return false
		end
		wkMap.new = wkMap.old:gsub(wkId .. "!", "")
	end
	if "" == wkMap.old then wkMap.old = nil end
	if "" == wkMap.new then wkMap.new = nil end
	
	return wkMap, wkId
end

--- sending request to MANAGER with wakeUp action
local function __exec_wkService(act, wkData)
	if "table" ~= type(wkData) or not wkData.tag then
		return false, "Invalid wakeUp data"
	end
	if "SET" ~= act and "DEL" ~= act then
		return false, "Unknown action"
	end
	
	local dModel     	= require"DataModel"
	local SYS	 	    = dModel.SYS
	local wkSrvPkt, err = dModel:new({ident = "SYS", ct = SYS.M_WAKEUP_SERVICE})
	if not wkSrvPkt then
		return false, err
	end
	
	local action = SYS.M_WA_DEL
	if "SET" == act then
		action = SYS.M_WA_SET
	end
	wkSrvPkt:setItem("WakeUpObject.action",    action)
	wkSrvPkt:setItem("WakeUpObject.tag",       tostring(wkData.tag))
	wkSrvPkt:setItem("WakeUpObject.timevalue", tonumber(wkData.timeval))
	wkSrvPkt:setItem("WakeUpObject.tries", 	   tonumber(wkData.tries))
	wkSrvPkt:setItem("WakeUpObject.period",    tonumber(wkData.timeout))
	
	return require"appHelper":startTransaction(wkSrvPkt)
end

--- Sending wakeUp message to FIAS
local function __send_wkMessage(wkData)
	if "table" ~= type(wkData) then
		return false, "Invalid arguments"
	end
	
	local wkMsgPkt, err = require"DataModel":new({ident = "FIAS"})
	if not wkMsgPkt then
		return false, err
	end
	
	wkMsgPkt:setItem("data.ident", "WR" == wkData.treq and 27 or 29)	-- WR/WC message
	wkMsgPkt:addRepeatedTuple("data.fields", "RN", wkData.rn)
	wkMsgPkt:addRepeatedTuple("data.fields", "DA", wkData.da)
	wkMsgPkt:addRepeatedTuple("data.fields", "TI", wkData.ti)
	
	return require"appHelper":startTransaction(wkMsgPkt)
end

--- Sending wakeUp anwer message to FIAS
local function __send_wkAnswer(wkData)
	if "table" ~= type(wkData) then
		return false, "Invalid arguments"
	end
	
	local wkAnswerPkt, err = require"DataModel":new({ident = "FIAS"})
	if not wkAnswerPkt then
		return false, err
	end
	
	wkAnswerPkt:setItem("data.ident", 28)	-- WA message
	wkAnswerPkt:addRepeatedTuple("data.fields", "RN", wkData.rn)
	wkAnswerPkt:addRepeatedTuple("data.fields", "DA", wkData.da)
	wkAnswerPkt:addRepeatedTuple("data.fields", "TI", wkData.ti)
	wkAnswerPkt:addRepeatedTuple("data.fields", "AS", wkData.status)
	
	return require"appHelper":startTransaction(wkAnswerPkt)
end

--- Sending originate event to PBX
local function __send_wkRing(ringData)
	if "table" ~= type(ringData) then
		return false, "Invalid ring data"
	end
	
	local dModel 		 = require"DataModel"
	local PBX			 = dModel.PBX
	local wkRingPkt, err = dModel:new({ident = "PBX", ct = PBX.P_ORIGINATE})
	if not wkRingPkt then
		return false, err
	end
	
	wkRingPkt:addRepeatedTuple("opts.items", "Channel",  ringData.channel)
	wkRingPkt:addRepeatedTuple("opts.items", "Exten",    ringData.exten)
	wkRingPkt:addRepeatedTuple("opts.items", "Context",  ringData.context)
	wkRingPkt:addRepeatedTuple("opts.items", "Timeout",  ringData.timeout)
	wkRingPkt:addRepeatedTuple("opts.items", "CallerID", ringData.callerid)
	wkRingPkt:addRepeatedTuple("opts.items", "Variable", ringData.variable)
	wkRingPkt:addRepeatedTuple("opts.items", "Variable", ringData.userfield)
	
	return require"appHelper":startTransaction(wkRingPkt)
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
WakeUpController._instance = nil -- Singletone instance

function WakeUpController:getInstance(dataProvider, roomProvider)
	if self._instance then
		return self._instance
	end
	if not dataProvider or not roomProvider then
		return false
	end
	
	self._instance = WakeUpController
	local o = self._instance
	
	o.dataProvider = dataProvider
	o.roomProvider = roomProvider
	
	--- Create wakeUp item
	-- @param wkData	A wakeUp data with WK time and room number
	-- @return TRUE with WK answer code if present otherwise FALSE followed error message
	o.create = function(wkData)
		if "table" ~= type(wkData) then
			printdbg("WakeUp data is not found")
			return false
		end
		if not wkData.rn or not wkData.da or not wkData.ti or not wkData.ifn then
			printdbg("Not found mandatory wakeUp data")
			return false
		end
		
		local rs		= false
		local rData 	= nil
		
		local wkMap     = {}
		local wkId      = nil
		local roomInfo  = nil
		local WKRetCodes		= _WKRetCodes["CREATE"]
		
		repeat
			-- get room information
    		roomInfo = o.roomProvider.getRoom(wkData.rn, wkData.ifn)
    		if not roomInfo then
    			print("Not found room %q for WakeUp settings!", wkData.rn)
    			break
    		end
    		
    		-- processing message from PBX
    		if "PBX" == wkData.ifn then
    			local wkMsgData = {rn = roomInfo.number, da = wkData.da, ti = wkData.ti, treq = wkData.treq}
    			__send_wkMessage(wkMsgData)
    		end
    		
    		-- check room state
    		if not o.roomProvider.getRoomGuest(roomInfo.ident) then
    			print("Room %q is empty. Dropping wakeUp request ...", wkData.rn)
    			rData = {wkData, WKRetCodes["EMPTY_ROOM"]}
    			break
    		end
    		
    		-- some date/time transformations
    		local wkTime = require"dateConverter".fias2unix(wkData.da, wkData.ti)
    		if 0 == wkTime then
    			print("Couldn't process WakeUp time!")
    			rData = {wkData, WKRetCodes["INVALID_TIME"]}
    			break
    		end
    		
    		-- check wakeup in current list
    		if o.dataProvider.getWakeUpDataByTimeVal(roomInfo.ident, wkTime) then
    			print("Found the same wakeUp %q for room %q", os.date("%T %F", wkTime), wkData.rn)
    			rData = {wkData, WKRetCodes["FOUND_ITEM"]}
    			break
    		end
    		
    		-- update WK map by new id
    		wkMap, wkId = __update_wkMap(o.dataProvider, "add", roomInfo.ident)
    		if "table" ~= type(wkMap) or not wkId then
    			print("WakeUp Map is empty!")
    			rData = {wkData, WKRetCodes["MAP_ERROR"]}
    			break
    		end
    		
    		-- get wakeUp defines
    		local wkTries   = o.dataProvider.getWakeUpTries()
    		local wkTimeout = tonumber(o.dataProvider.getWakeUpTimeout())
    		if not wkTimeout or not wkTries then
    			print("Define variables are empty!")
    			rData = {wkData,  WKRetCodes["NO_DEFINES"]}
    			break
    		end
    		wkTimeout = wkTimeout + 60 	-- increase timeout for waiting CDR event 
    		
    		-- push WK map an item to storage
    		o.dataProvider.setWakeUpMap(roomInfo.ident, wkMap.new)
			o.dataProvider.setWakeUpData(roomInfo.ident, wkId, wkTime)
			
			-- push WK request to manager
			local tag = roomInfo.ident.."!"..wkId
			__exec_wkService("SET", {tag = tag, timeval = wkTime, tries = wkTries, timeout = wkTimeout})
			
			rs 		= true
			print("WakeUp has been set for room %q at %q", roomInfo.number, os.date("%T %F", wkTime))
		until true
		
		return rs, rData
	end
	
	--- Delete wakeUp item
	-- @param wkData	A wakeUp data with WK time and room number
	-- @param reason	A reason which makes this action
	-- @return TRUE with WK answer code if present otherwise FALSE followed error message
	-- @note We will not remove data from astDB in case "GuestOut". It must be done by RoomEquipmentController
	o.delete = function(wkData, reason)
		if "table" ~= type(wkData) or not reason then
			printdbg("WakeUp data is not found")
			return false
		end
		
		if not wkData.rn or not wkData.ifn then
			printdbg("Not found mandatory wakeUp data")
			return false
		end
		
		local status = nil
		if "CDREvent" == reason then
			if not wkData.wkId then
				printdbg("Not found wakeUp id in case PBX event")
				return false	-- User did not hear wakeUp ringer. Manager will try again
			end
			status = "ANSWERED"
		elseif "SYSTimeout" == reason then
			if not wkData.wkId then
				printdbg("Not found wakeUp id in case system timeout")
				return false	-- User did not hear wakeUp ringer. Manager will try again
			end
			status = "TIMEOUT"
		elseif "WakeUpClear" == reason or "GuestOut" == reason then
			status = "CLEAR"
		end
		
		local rs  	= false
		local rData = nil
		local WKRetCodes = _WKRetCodes["DELETE"]
		
		repeat
			local wkId      = nil
			local wkDate    = nil
			local wkTime    = nil
			local wkTimeVal = nil
			
			-- get room information
			local roomInfo  = o.roomProvider.getRoom(wkData.rn, wkData.ifn)
			local roomIdent = roomInfo and roomInfo.ident
			
			-- get wakeUp data
			if "CLEAR" == status then	-- we got FIAS/PBX pkt with WK RN,DA,TI
				wkDate	  = wkData.da
    			wkTime    = wkData.ti
				
				if "000000" ~= wkData.da and "000000" ~= wkData.ti then
    				wkTimeVal = require"dateConverter".fias2unix(wkData.da, wkData.ti)
					wkId 	  = o.dataProvider.getWakeUpDataByTimeVal(roomIdent, wkTimeVal)
				else
					wkTimeVal = "000000"
					wkId 	  = "000000"
				end
			else			-- we got WK ID, RID
				wkId	  	   = wkData.wkId 
				wkTimeVal      = o.dataProvider.getWakeUpData(roomIdent, wkId)
				wkDate, wkTime = require"dateConverter".unix2fias(wkTimeVal)
			end
			
			if not roomInfo then
				print("Not found room %q for WakeUp item!", wkData.rn)
				break
			end
			if "CLEAR" == status and not wkId then
				print("Not found wakeUp %q for room %q!", os.date("%T %F", wkTimeVal), wkData.rn)
				break
			end
        	if not wkId or not wkTimeVal or not wkDate or not wkTime then
        		print("Error during wakeUp data retrieving!")
        		break
    		end
    		
    		-- check wakeUp item for room in map
    		if "000000" == wkId then	-- remove wakeUps for all room
    			local wkItems = o.dataProvider.getAllwakeUp(roomIdent)
    			if wkItems then
    				for wkId, timeVal in pairs(wkItems) do
    					__exec_wkService("DEL", {tag = roomIdent.."!"..wkId})
    					o.dataProvider.delWakeUpData(roomIdent, wkId, timeVal, "GuestOut" == reason)
    				end
        			o.dataProvider.delWakeUpMap(roomIdent, "GuestOut" == reason)
        		end
            else
        		local wkMap = __update_wkMap(o.dataProvider, "del", roomIdent, wkId)
        		if "table" ~= type(wkMap) then
        			print("WakeUp item does not found in the map!")
        			break
        		end
        		
        		-- remove wakeUp item from manager
        		__exec_wkService("DEL", {tag = roomIdent.."!"..wkId})
    			
    			-- remove WK item from astDB and memory
    			o.dataProvider.delWakeUpData(roomIdent, wkId, wkTimeVal)
    			
    			-- update wakeUp map in astDB and memory by deleted id
    			if wkMap.new then
            		o.dataProvider.setWakeUpMap(roomIdent, wkMap.new)
            	else
            		o.dataProvider.delWakeUpMap(roomIdent) 
        		end
        	end
        	
        	-- fill wakeUp data by new variables for wakeUp Answer
        	local wkAnsData = {
        		 rn = roomInfo.number
        		,da = wkDate
        		,ti = wkTime
        	}
        	
        	-- processing message from PBX
    		if "PBX" == wkData.ifn and "WakeUpClear" == reason then
    			local wkMsgData = {rn = roomInfo.number, da = wkData.da, ti = wkData.ti, treq = wkData.treq}
    			__send_wkMessage(wkMsgData)
    		end
			
			-- set final return variables
			if "ANSWERED" == status then
				rData = {wkAnsData, WKRetCodes["ANSWERED"]}
			elseif "TIMEOUT" == status then
				rData = {wkAnsData, WKRetCodes["TIMEOUT"]}
			end
			rs 	   = true
			print("WakeUp %q has been removed for room %q successfully!", wkId, wkData.rn)
		until true
		
		return rs, rData
	end
	
	--- Move wakeUp items for one room to other
	-- @param srcRoomIdent		A identitification of the source room with wakeups
	-- @param dstRoomIdent		A destination room identification
	-- @return boolean value
	-- @note wakeUp data from AstDB must be removed by RoomEquipmentController
	o.move = function(srcRoomIdent, dstRoomIdent)
		if not srcRoomIdent or not dstRoomIdent then
			return false
		end
		
		-- get wakeUp defines
   		local wkTries   = o.dataProvider.getWakeUpTries()
   		local wkTimeout = tonumber(o.dataProvider.getWakeUpTimeout()) + 60
   		 
		local wkItems, wkMap = o.dataProvider.getAllwakeUp(srcRoomIdent)
		if not wkItems then
			return true
		end
		
		-- set up wakeUp for destination room
		for wkId, timeVal in pairs(wkItems) do
			local tag = dstRoomIdent.."!"..wkId
			__exec_wkService("SET", {tag = tag, timeval = timeVal, tries = wkTries, timeout = wkTimeout})
			o.dataProvider.setWakeUpData(dstRoomIdent, wkId, timeVal)
		end
		o.dataProvider.setWakeUpMap(dstRoomIdent, wkMap)
		
		-- remove wakeUp for source room
		for wkId, timeVal in pairs(wkItems) do
			__exec_wkService("DEL", {tag = srcRoomIdent.."!"..wkId})
			o.dataProvider.delWakeUpData(srcRoomIdent, wkId, timeVal, true)
		end
		o.dataProvider.delWakeUpMap(srcRoomIdent, true)
		
		return true
	end
	
	--- Send wakeUp answer to FIAS interface
	-- @param wkData	A wakeUp data with WK time and room number
	-- @param status	An answer status (see description in the header)
	-- @return boolean value followed by error message if present
	o.sendAnswer = function(wkData, status)
		if "table" ~= type(wkData) or not status then
			printdbg("Invalid arguments")
			return false
		end
		
		if not wkData.rn or not wkData.da or not wkData.ti then
			printdbg("Not found mandatory parameters for WakeUp answer!")
			return false
		end
		
		wkData.status = status or "UR"
		
		local rs, err = __send_wkAnswer(wkData)
		if not rs then
			print(err)
		end
		return rs
	end
	
	--- Processing wakeUp event as sending message to PBX interface
	-- @param wkData	A wakeUp data with WK time and room number
	-- @return boolean value followed by error message if present
	o.ring = function(wkData)
		if "table" ~= type(wkData) then
			printdbg("Invalid wakeUp data")
			return false
		end
		
		if not wkData.rn or not wkData.wkId or not wkData.ifn then
			printdbg("Couldn't get wakeUp data from manager request")
			return false
		end
		
		-- get room info
		local roomInfo = o.roomProvider.getRoom(wkData.rn, wkData.ifn)
		if not roomInfo then
			print("Not found information for room ID %q", wkData.rn)
			return false
		end
		local roomIdent = roomInfo and roomInfo.ident
		local wkId		= wkData.wkId
		
		-- get wakeUp defines
		local wkContext = o.dataProvider.getWakeUpContext()
		local wkExten   = o.dataProvider.getWakeUpExten()
		local wkTimeout = tonumber(o.dataProvider.getWakeUpTimeout())
		if not wkExten or not wkTimeout or not wkContext then
			print("Define variables are empty!")
			return false
		end
		
		-- handling some ring data variables
		local ringChannel = roomInfo.exten or wkExten
		if ringChannel then
			ringChannel = "SIP/"..ringChannel
		end
		local ringTimeout = (wkTimeout or 60) * 1000 -- converting to ms
		
		local ringData = {
			channel	  = ringChannel,
			exten     = wkExten,
			context   = wkContext,
			timeout   = ringTimeout,
			callerid  = "WHotel WakeUp <"..roomInfo.exten..">",
			variable  = "WHotelWakeup=WHOTEL-WAKEUP:"..roomIdent.."!"..wkId,
			userfield = "CDR(UserField)=WHOTEL-WAKEUP"
		}
		
		local rs, err = __send_wkRing(ringData)
		if not rs then
			print(err)
		else
			print("Making call for WakeUp in room %q to extension %q ...", roomInfo.exten, wkExten)
		end
		
		return rs  
	end
	
	--- Sync WakeUps for rooms with loading from AST storage and create in manager
	-- @param roomsList		A list of the rooms idents that must be synced
	-- @return boolean value
	o.sync = function(roomsList)
		if "table" ~= type(roomsList) then
			return false
		end
		
		-- get wakeUp defines
		local wkTries   = o.dataProvider.getWakeUpTries()
		local wkTimeout = tonumber(o.dataProvider.getWakeUpTimeout())
		if not wkTimeout or not wkTries then
			print("We can't make WakeUps sync. Define variables are empty!")
			return false
		end
		wkTimeout = wkTimeout + 60 	-- increase timeout for waiting CDR event
		
		-- start sync procedure
		local syncRoomWKCnt = 0
		o.dataProvider.setWakeUpSyncFlag()
		for _, roomIdent in ipairs(roomsList) do
			local wkMap = o.dataProvider.getAstWakeUpMap(roomIdent) or ""
			if "" ~= wkMap then
				o.dataProvider.setWakeUpMap(roomIdent, wkMap, true)
    			for wkId in wkMap:gmatch("(%d+)!?") do
    				local wkTime = o.dataProvider.getAstWakeUpData(roomIdent, wkId)
    				o.dataProvider.setWakeUpData(roomIdent, wkId, wkTime, true)
        			
        			-- push WK request to manager
        			local tag = roomIdent.."!"..wkId
        			__exec_wkService("SET", {tag = tag, timeval = wkTime, tries = wkTries, timeout = wkTimeout})
    			end
    			syncRoomWKCnt = 1 + syncRoomWKCnt
    		end
		end
		o.dataProvider.delWakeUpSyncFlag()
		printdbg("Synced WakeUps for %d rooms", syncRoomWKCnt)
	end
	
	--- Check data sync process for WakeUps 
	o.getSyncFlag = function()
		return o.dataProvider.getWakeUpSyncFlag() 
	end
end

return WakeUpController