--- memcache simple driver for raw access
-- 
-- @file	memcSimpleRawInterface.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/23

--- @module memcSimpleRawInterface
local memcSimpleRawInterface = {
	_instance  = nil,
}

----------------------------------------------------------------
-- singleton implementation
----------------------------------------------------------------
function memcSimpleRawInterface:getInstance(cfg)
    if self._instance then
    	return self._instance
    end
    
	self._instance = memcSimpleRawInterface
	local o         = self._instance
		  o.sock    = nil
		  o.refcnt  = 0
		  o.key_ttl = 0		-- expiration time in seconds, 0 mean no delay
	
	if "table" == type(cfg) then
		o.host   = cfg.host
		o.port   = tonumber(cfg.port)
		o.prefix = cfg.prefix
	end

	--- Connect to the data storage
	o.connect = function(host, port, prefix)
		if o.sock then
			o.refcnt = o.refcnt + 1 
			return true
		end
		
		local prefix = prefix or o.prefix
		local host   = host or o.host
		local port   = tonumber(port) or o.port
		if not host or not port or not prefix then
			return false, "Invalid input parameters"
		end
		local err = nil
		repeat
			o.sock, err = require"socketConnector":new({host = host, port = port})
			if not o.sock then
				err = string.format("MemcSimple: could not load socket evironment")
				break
			end
			_, err = o.sock:connect()
			if err  then
				err = string.format("MemcSimple: could not connect to %s:%s (%s) ", host, port, err or "")
				break
			end
		until true

		return err == nil, err
	end

	--- Disconnect from the data storage
	o.disconnect = function()
		if o.sock then
			o.refcnt = o.refcnt - 1
			if 0 == o.refcnt then
				o.sock:close()
			end
		end
		return false
	end

	--- Put data to storage
	o.put = function(key, val, ttl)
		if not o.sock then
			return false, "Memcache connection not found"
		end
        
		local query = string.format("set %s 0 %d %d\r\n%s\r\n", o.prefix.."/"..key, ttl or o.key_ttl, #(tostring(val)), val)
		local rv, err = o.sock:send(query)
		if rv then
			rv, err = o.sock:receive()
			if rv and rv:find("^STORED") then
				return true
			end
		end
	
		return false, err
	end

	--- Get data from storage 
	o.get = function(key)
		if not o.sock then
			return false, "Memcache connection not found"
		end
		
		local query = string.format("get %s\r\n", o.prefix.."/"..key)
		local rv, err = o.sock:send(query)
		if rv then
			rv, err = o.sock:receive()
			if rv then
				local _, i, len = rv:find("^VALUE%s+[^%s]+%s+%d+%s(%d+)")
				if len and "0" ~= len then
					return rv:sub(i+2, i+len+1)
				end 
			end
		end
		return false, err
	end

	--- delete data from storage
	o.del = function(key)
		if not o.sock then
			return false, "Memcache connection not found"
		end
		
		local query = string.format("delete %s\r\n", o.prefix.."/"..key)
		local rv, err = o.sock:send(query)
		if rv then
			rv, err = o.sock:receive()
			if rv and rv:find("^DELETED") then
				return true
			end
		end
		return false, err
	end
	
	--- increment/decrement value associated with key
	o.auto = function(act, key)
		if "incr" ~= act and "decr" ~= act or not key or "" == key then
			return false, "Invalid arguments"
		end
		if not o.sock then
			return false, "Memcache connection not found"
		end
		
		local amount  = 1 -- the amount to increase/decrease the item
		local query   = string.format("%s %s %d\r\n", act, o.prefix.."/"..key, amount)
		local rv, err = o.sock:send(query)
		if rv then
			rv, err = o.sock:receive()
			if rv then
				if "NOT_FOUND" == rv then
					rv = nil
				end
			end
		end
		return rv, err
	end
	
	-- @see auto method for incrementing counter
	o.incr = function(key)
		return o.auto("incr", key)
	end
	
	-- @see auto method for decrementing counter
	o.decr = function(key)
		return o.auto("decr", key)
	end
	
    return o
end

return memcSimpleRawInterface
