--- Sqlite raw interface
-- 
-- @file	sqliteRawInterface.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/05/23

--- @module sqliteRawInterface
sqliteRawInterface = {
	_instance  = nil,
}


----------------------------------------------------------------
-- singleton implementation
----------------------------------------------------------------
function sqliteRawInterface:getInstance(fn)
    if self._instance then
    	return self._instance
    end

	self._instance  = sqliteRawInterface
	local o         = self._instance
		  o.refcnt  = 0
		  o.conn    = nil
		  o.env     = nil
		  o.file    = fn

	--- Connect to the data storage
	-- @param fn		database filename
	o.connect = function(fname)
		if o.conn then
			o.refcnt = o.refcnt + 1 
			return true
		end
		local fn = fname or o.file
		if not fn then
			return false, "Invalid input parameters"
		end

		local err = nil
		repeat
			o.env = require"luasql.sqlite3".sqlite3()
			if not o.env then
				err = "Sqlite: couldn't load driver"
				break
			end

			o.conn = o.env:connect(fn)
			if not o.conn then
				err = string.format("Sqlite: could not connect to %s", fn)
				break
			end
		until true
		
		return err and false or true, err
	end
	
	--- Disconnect from the data storage
	o.disconnect = function()
		if o.conn then
			o.refcnt = o.refcnt - 1
			if 0 == o.refcnt then
				o.conn:close()
				if o.env then
					o.env:close()
				end
				o.conn = nil
				o.env  = nil
				return true
			end
		end
		return false
	end
 
 	--- Sqlite one-step execution function
	o.exec = function(query)
		if not o.conn or not query then
			return false
		end
		
		local attempts = 5
		local err      = nil
		repeat
			local rv = nil
			rv, err  = o.conn:execute(query)
			if "userdata" == type(rv) or "number" == type(rv) then
				return rv -- can be cursor (userdata) or affected rows count (number)  
			end

			if err then	-- processing error
				err = err:match("LuaSQL: (.*)")
				if err:find("no such table") or err:find("SQL logic error or missing database") then
					attempts = 0
				else
					if 0 < attempts then
						require"socket".sleep(0.1)	-- sleep 100 ms
					end
				end
			end
			
			attempts = attempts - 1
		until 0 >= attempts
		
		return false, string.format("Sqlite: %s (%s)", err or "", query)
	end
 
 	--- Retrieve Sqlite results (use unpack or associative array)
	o.steps = function(cursor)
		if not cursor or "userdata" ~= type(cursor) then
			return false, "Sqlite: cursor object not found"
		end

		local colnames = cursor:getcolnames()
		if not colnames or "table" ~= type(colnames) then
			cursor:close()
			return false, "Sqlite: error in column names checking"
		end
		
		local colnum = #colnames
		local rt	 = {}
		local function fetch_result()
			local recordData = {cursor:fetch()}
			if recordData[1] then
				if 1 < colnum then	-- set name values of the retreived data in result
					for idx, name in ipairs(colnames) do
						recordData[name] = recordData[idx]
					end
				end
	    		table.insert(rt, recordData)
	    		return fetch_result()
	    	end
			cursor:close()
		end
		fetch_result()

		if 0 == #rt then
			return false
		end
		if 1 == #rt then
			if 1 == colnum then
				return rt[1][1]
			else
				return rt[1]
			end
		else
			local ridx = 0
			return function()
				ridx = 1 + ridx
				if rt[ridx] then
					return unpack(rt[ridx], 1, colnum)
				end		
			end
		end
	end
 
    return o
end

return sqliteRawInterface

