--- Data storage library
-- 
-- @file	DataStorage.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/04/06

--[[
   data can be stored in:
   *  "mem"   - stored in 'memcache', temporary, key->value
   *  "hotel" - stored on disk in database, persistant, any data sets related to DB schema
   *  "ast"   - stored in AsteriskDB via AMI interface, persistant, key->value
]]
--- @module DataStorage
local DataStorage = {}

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local storage = { backends={} }        -- meta class

--- Simplify access to a memDB
function storage:putTemp(key, val)
	return self:put("mem",key,val);
end

--- Put data to a storage
-- @param dst           type of backend for write into
-- @param key           data key
-- @param val           data value
-- @return boolean status of operation
function storage:put(dst,key,val)
	if not self.backends[dst] then
		return false, string.format("Storage backend %q does not registered", dst) 
	end
	return self.backends[dst]:put(key, val)
end

--- Simplify access to a memDB
function storage:getTemp(key)
	return self:get("mem", key);
end

--- Get data from storage
-- @param dst           type of backend for read from
-- @param key           data key
-- @return status,value; in case "no errors" status should be 'true'
function storage:get(dst,key)
	if not self.backends[dst] then
		return false, string.format("Storage backend %q does not registered", dst)
	end
	return self.backends[dst]:get(key)
end

--- Simplify access to a memDB
function storage:delTemp(key)
	return self:del("mem",key);
end

--- Delete data from storage
-- @param dst           type of backend
-- @param key           data key
function storage:del(dst,key)
	if not self.backends[dst] then
		return false, string.format("Storage backend %q does not registered", dst)
	end
	return self.backends[dst]:del(key)
end


--[[
 config:
   obj={}
   obj["mem"]   = {ident="memc/sql/memc_simple",options={mode="persist/runtime",file="",host="",port=XXX}}
   obj["hotel"] = {file="",mode="persist/runtime"}
   obj["ast"]   = {prefix=""}
--]]
function DataStorage:new(cfg)
	if "table" ~= type(cfg) then
		return false, "Configuration format error"
	end

	local obj = {}
	setmetatable(obj, {__index = storage})

	local factory = require"AccessorFactory":new(cfg)
	local err 	  = nil	
	for k,v in pairs(cfg) do
		obj.backends[k], err = factory:instantiate(k)
	end
	
	if err then
		return false, err
	end
	return obj
end

return DataStorage
