--- Sqlite driver
-- 
-- @file	sqliteDriver.lua 
-- @author	Pavel Yadvichuk
-- @created 2015/04/07

--- @module sqliteDriver
local sqliteDriver = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------

--- Prepare SQL query for request
local function __query_build(qtype, key, val)
	if not qtype or "table" ~= type(key) or (val and "table" ~= type(val)) then
		return false, "Sqlite: wrong query format"
	end
	if not key.table or "" == key.table then
		return false, "Sqlite: table was not specified"
	end
	
    -- set where part
    local where = ""
    if key.where and "" ~= key.where then
    	where = string.format("WHERE %s", key.where)
    end
    
    -- set additional postfix part
    local postfix = ""
    if key.postfix and "" ~= key.postfix then
        postfix = key.postfix
    end
	
	local query = nil
	if "new" == qtype then
		if "table" ~= type(key.fields) or "table" ~= type(val) then
			return false, "Sqlite(NEW): wrong query values"
		end
		-- normilize query values
		for k, v in pairs(val) do
			if "string" == type(v) then
				val[k] = "'" .. v .. "'" 
			end
		end
		query = string.format("INSERT INTO %s(%s) VALUES(%s)", key.table, table.concat(key.fields, ','), table.concat(val, ','))
	elseif "del" == qtype then
		query = string.format("DELETE FROM %s %s", key.table, where)
	elseif "upd" == qtype then
		if "table" ~= type(key.fields) or "table" ~= type(val) then
			return false, "Sqlite(UPD): wrong query values"
		end
		-- normilize query variables
		local updVars = {}
		for idx, column in pairs(key.fields) do
			if val[idx] then
				local updVal = val[idx]
				if "string" == type(updVal) then
					updVal = "'" .. updVal .. "'"
				end
				table.insert(updVars, column .. "=" .. updVal)
			end
		end
		query = string.format("UPDATE %s SET %s %s", key.table, table.concat(updVars, ','), where)
	elseif "get" == qtype then
		if "table" ~= type(key.fields) then
			return false, "Sqlite(GET): wrong query values"
		end
		query = string.format("SELECT %s FROM %s %s %s", table.concat(key.fields, ','), key.table, where, postfix)
	else
		return false, "Sqlite: unknown query method"
	end
	
	return query
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local driver = require"IDriver":new()        -- meta class
driver.file  = nil	-- Sqlite database location

--- @see description in IDriver
function driver:put(key, val)
	if not self.connected then
		return false
	end
	
	local qtype		 = key.act or "new"
	local query, err = __query_build(qtype, key, val)
	if not query then
		return false, err
	end
	
	return self.raw.exec(query)
end

--- @see description in IDriver
function driver:get(key)
	if not self.connected then
		return false
	end
	
	local err = nil
	
	local query, err = __query_build("get", key)
	if not query then
		return false, err
	end
	
	local rv         = nil
	local cursor,err = self.raw.exec(query)
	if cursor then
		rv = self.raw.steps(cursor)
	end
	return rv
end

--- @see description in IDriver
function driver:del(key)
	if not self.connected then
		return false
	end
	
	local query, err = __query_build("del", key)
	if not query then
		return false, err
	end
	
	return self.raw.exec(query)
end

--- Connect to the data storage
function driver:connect()
	local err = nil
	if not self.connected then
		self.connected,err = self.raw.connect(self.file)
	end
	return self.connected, err
end

--- Disconnect from the data storage
function driver:disconnect()
	self.connected = self.raw.disconnect()
end


--- Constructor
function sqliteDriver:new(cfg)
	if not cfg.file or "" == cfg.file then
		return false, "Sqlite: wrong configuration parameters"
	end
	
	local obj = {}
	setmetatable(obj, {__index = driver})
	
	obj.file = cfg.file
	obj.raw  = require("sqliteRawInterface"):getInstance()
	
	return obj
end

return sqliteDriver
