--- Convertor for different date/time formats
-- 
-- @file    dateConverter.lua 
-- @author  Vyacheslav Khudan
-- @created 2015/02/25

--- @module dateConverter
local dateConverter = {}

--- Convertor from FIAS date/time to UNIX one in local timezone
-- @param date  A FIAS date
-- @param time	A FIAS time
-- @return UNIX time format
function dateConverter.fias2unix(date, time)
	local rv = 0
	if not date then
		return rv
	end
	time = time or "000000"
	
	local hour, min, sec = time:match("(%d%d)(%d%d)(%d%d)")
	if not hour then
		sec 	  = "00"
		hour, min = time:match("(%d%d):(%d%d)")
	end
	if not hour or not min or not sec then
		return rv
	end
	
	local year, month, day  = date:match("(%d%d)(%d%d)(%d%d)")
	if not year or not month or not day then
		return rv
	end
	
	year = "20" .. year		-- year value recieved only last two digit from FIAS
	local conv_time = {
		["hour"]	= hour,
		["min"]		= min,
		["sec"]		= sec,
		["year"] 	= year,
		["month"] 	= month,
		["day"]		= day
	}
	
	return os.time(conv_time)
end

--- Convertor from FIAS date/time to UNIX one in UTC timezone
-- @see fias2unix function
-- @note idea got from http://lua-users.org/lists/lua-l/2008-03/msg00051.html
function dateConverter.fias2unixUTC(date, time)
	local rtime = dateConverter.fias2unix(date, time)
	
	if 0 ~= rtime then
		-- fix up for UTC timezone
    	local gtime = os.date("!*t", rtime)							-- find out what UTC t was converted to
    	local ltime = os.date("*t", rtime)							-- find out if daylight savings was applied
    	gtime.isdst = ltime.isdst				 					-- apply DST to this time if necessary
    	rtime 		= rtime + os.difftime(rtime, os.time(gtime))
    end 
	return rtime
end

--- Convert sencods value to FIAS representation like HHMMSS
-- @param seconds	A senconds value 
-- @return FIAS time represenation
function dateConverter.second2fias(seconds)
	if not tonumber(seconds) then
		seconds = 0
	end
	
	return os.date("!%H%M%S", seconds)
end

--- Convert from string UNIX to FIAS date representation 
-- @param date_str	A string representation of the date
-- @return FIAS date and time
function dateConverter.unix2fias(date_str)
	if not date_str or 0 == date_str then
		return "000000", "000000" 
	end
	
	return os.date("%y%m%d", date_str), os.date("%H%M%S", date_str)	
end

--- Convert from string date representation to UNIX
-- @param date_str	A string representation of the date
-- @return UNIX time format
function dateConverter.string2unix(date_str)
	local rv = 0
	if not date_str or "" == date_str then
		return rv
	end
	
	local year, month, day, hour, min, sec = date_str:match("^(%d+)-?/?(%d+)-?/?(%d+)%s+(%d+):(%d+):?(%d*)$")
	if not year or not month or not day or not hour or not min then
		return rv
	end
	
	local sec = sec or 0
	local conv_time = {
		["hour"]	= hour,
		["min"]		= min,
		["sec"]		= sec,
		["year"] 	= year,
		["month"] 	= month,
		["day"]		= day
	}
	
	return os.time(conv_time) or 0
end

--- Convert from string date to FIAS representation
-- @param date_str	A string representation of the date
-- @return FIAS time format
function dateConverter.string2fias(date_str)
	local unixDate = dateConverter.string2unix(date_str)
	return dateConverter.unix2fias(unixDate)
end

return dateConverter