--- Memcache queue implementation
-- 
-- @file	memcQueue.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/05/25
-- 
-- base idea comes from		https://github.com/abhinavsingh/memq/blob/master/memq.class.php

--- @module memcQueue
local memcQueue = {}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
local function __get_raw_memc()
	local memc		 = require"memcRawInterface"
	if memc._instance then
		return memc:getInstance()
	end
	
	local memcSimple = require"memcSimpleRawInterface"
	if memcSimple._instance then
		return memcSimple:getInstance()
	end
	
	return false
end

local function __exec(raw, functor, args)
	local err = nil
	local rv  = nil
	
	if raw.request then	-- used memc via manager
		local key, val = unpack(args)
		local act = "string" == type(functor) and functor:upper()
		rv, err = raw.request(nil, {act = act, key = key, val = val})
	else	-- used memcSimple
		repeat
			rv, err = raw.connect()
			if not rv then break end
			
			rv, err = raw[functor](unpack(args))
			if not rv then break end
		until true
		raw.disconnect()
	end
    
	return rv, err
end

local function __get_top_elm(raw, prefix, getter, is_extract)
	if not raw or not prefix or not getter then
		return false
	end
	
	local id  = nil 
	local err = nil
	if is_extract then
		id, err = __exec(raw, "incr", {prefix.."_head"})
	else
		id, err = __exec(raw, "get", {prefix.."_head"})
	end
	if err or not id then
		return false, err
	end
	
	local tail, err = __exec(raw, "get", {prefix.."_tail"})
	if not tail then
		return false, err
	end

	tail = tonumber(tail)	
	id   = tonumber(id)
	if is_extract then
		id = id - 1
	end
	if id > tail then
		if is_extract then
			_, err = __exec(raw, "decr", {prefix.."_head"})
		end
		return false, err
	end
	
	local val, err = __exec(raw, "get", {prefix.."_"..id})
	if not val then
		return false, err
	end
	if is_extract then
	   __exec(raw, "del", {prefix.."_"..id})
	end
	
	return id, getter(val)
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local queue  = {} -- meta class
queue.raw	 = nil	-- interface to memcache
queue.setter = nil  -- callback function for writing value to queue
queue.getter = nil  -- callback function for reading value from queue
queue.prefix = ""   -- queue name in memcache

--- Initialize queue system variables
-- @return boolean value about initialization results
function queue:init()
    local id, err = __exec(self.raw, "get", {self.prefix.."_tail"})
    if err then
        return false, err
    end
    
    if id then
        return true
    end
    
    local rs, err = __exec(self.raw, "put", {self.prefix.."_tail", 0})
    if not rs then
        return false, err
    end
    rs, err = __exec(self.raw, "put", {self.prefix.."_head", 1})
    if not rs then
        return false, err
    end
    
    return true
end

--- Push value to memchache queue
-- @param value 	An value that we will push to queue
-- @return pushed value id otherwise FALSE followed error message 
function queue:push(val)
	local val, err = self.setter(val)
	if not val then
		return false, err
	end

	local id, err = __exec(self.raw, "incr", {self.prefix.."_tail"})
	if err then
		return false, err
	end
	
	local rs = false	
	if not id then
	   -- required queue initialization. Removed by issue
	   -- with race-condition during auto-initialization.
	   return false, "queue not initialized"
	end
	
	rs, err = __exec(self.raw, "put", {self.prefix.."_"..id, val})
	if not rs then
		return rs, err
	end
	
	return id
end

--- Extract value from memcache queue
-- @return some element value with internal id otherwise FALSE followed error message
-- @note during multi-threading processing try to pop empty element. 
--       Do not forget to check id before processing in the related functions.
function queue:pop()
	return __get_top_elm(self.raw, self.prefix, self.getter, true)
end

-- Check queue pool by zero element count
function queue:isEmpty()
	local head, err = __exec(self.raw, "get", {self.prefix.."_head"})
	if err then
		return false, err
	end
	
	local tail, err = __exec(self.raw, "get", {self.prefix.."_tail"})
	if err then
		return false, err
	end
	
	head = tonumber(head)
	tail = tonumber(tail)
	if not head or not tail or head > tail then
		return true
	end
	
	return false
end

--- Get top element from queue (like pop, but without value extracting)
-- @return top element value with internal id otherwise FALSE followed error message 
function queue:top()
	return __get_top_elm(self.raw, self.prefix, self.getter)
end

--- Private method to retrieve mutex ident bound up with thread/state id
function queue:__getLockID()
	local rv = __exec(self.raw, "get", {self.prefix.."_mutex"})
	return tonumber(rv)
end

--- Check the lock status
-- @return boolean value about lock status
function queue:isLocked()
	if self:__getLockID() then
		return true
	end
	return false
end

--- Try to set up the lock with timeout. Also skipped executing
--  if this thread make the lock before !
-- @param time  an expiration time in seconds (default 1 second)
-- @return boolean value as result for lock procedure
function queue:tryTimedLock(time)
	local lockID = self:__getLockID()
	if lockID then
		if _G.StateUID == lockID then
			return true
		end
		return false
	end

	time = time or 1
	return __exec(self.raw, "put", {self.prefix.."_mutex", _G.StateUID, time})
end

--- Executing unlock if the lock was maked by this thread
-- @return boolean value as result for the operation
function queue:unlock()
	if _G.StateUID == self:__getLockID() then
		return __exec(self.raw, "del", {self.prefix.."_mutex"})
	end
	return false
end

--- Constructor
function memcQueue:new(prefix, setter, getter)
	if "function" ~= type(setter) or "function" ~= type(getter) then
		return false, "Unable to set callback functions!"
	end
	if not prefix or "" == prefix then
		return false, "Invalid queue prefix!"
	end
	
	local obj  = {}
	setmetatable(obj, {__index = queue})
	
	-- init object variables
	obj.setter = setter
	obj.getter = getter
	obj.prefix = prefix
	obj.raw    = __get_raw_memc()
	
	return obj
end

return memcQueue