--- Connector for TCP or UDP sockets (only client sockets)
-- 
-- @file	socketConnector.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/05/15

--- @module socketConnector
local socketConnector = {}
--TODO: add implementation for UDP connections

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Checking configuration table
local function __check_cfg(cfg)
	if "table" ~= type(cfg) then
		return false
	end
	if not cfg.host or "" == cfg.host then
		return false
	end
	if not cfg.port or "" == cfg.port then
		return false
	end
	if cfg.timeout and not tonumber(cfg.timeout) then
		return false
	end
	if cfg.type and "TCP" ~= cfg.type and "UDP" ~= cfg.type then
		return false
	end
	
	return cfg
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
local socket = {
	connection = nil,
	cfg 	   = {
		host 	= nil,
		port 	= nil,
		type 	= "TCP",
		timeout = 1000
	} 
}

--- Send data to socket
-- @param data	A sending data
-- @return total number of bytes sent otherwise nil followed by an error message
function socket:send(data)
	if not self.connection then
		return false
	end
	
	return self.connection:send(data)
end

--- Receive data from the socket
-- @return data from the socket otherwise nil followed by an error message
function socket:receive()
	if not self.connection then
		return false
	end

	local data, err = self.connection:receive()
	local rv 		= data
	if rv then
		self:settimeout(0.1)
	end
	while data do
		data, err = self.connection:receive()
		if data then
			rv = rv .. "\n" .. data
		end
	end
	
	if rv then
		self:settimeout(self.cfg.timeout)
		err = nil
	end
	
	return rv, err
end

--- Set timeout for TCP sockets
-- @param timeout	A timeout for I/O socket operations in seconds
function socket:settimeout(timeout)
	if not self.connection then
		return false
	end

	if "TCP" ~= self.cfg.type then
		return false
	end
	local timeout = tonumber(timeout)
	if not timeout then
		return false
	end
	
	return self.connection:settimeout(timeout)
end

--- Open socket connection
-- @return boolean TRUE value otherwise FALSE value with a string describing the error
function socket:connect()
	local err = nil
	if not self.connection then
		local m_socket = require"socket"
		self.connection, err = m_socket.connect(self.cfg.host, self.cfg.port)
		if self.connection then
			self:settimeout(self.cfg.timeout)
		end
	end
	
	return self.connection, err
end

--- Close socket connection
function socket:close()
	if not self.connection then
		return false
	end
	
	self.connection:close()
	self.connection = nil
end


--- Constructor
function socketConnector:new(cfg)
	if not cfg then
		return false, "configuration error"
	end
	
	local obj = {}
	setmetatable(obj, {__index = socket})
	
	-- init configuration
	obj.cfg.port 		= cfg.port
	obj.cfg.host 		= cfg.host
	if cfg.type then
		obj.cfg.type 	= cfg.type
	end
	if cfg.timeout then
		obj.cfg.timeout = cfg.timeout
	end

	return obj
end

return socketConnector