--- Get/set WMS user data into LDAP
-- 
-- @file	wmsUser.lua 
-- @author	Vyacheslav Khudan
-- @created 2015/07/02

-- load additional modules
local modPathList = {
     "/usr/local/share/lua/5.1/common/?.lua"
    ,"/usr/local/share/lua/5.1/common/userFeatures/?.lua"
}
package.path  = package.path .. ";" .. table.concat(modPathList, ";")

--- @module wmsUser
local wmsUser = {
    mAttrs = {      -- user attributes source
         name    = "LDAP"
        ,surname = "LDAP"
        ,lang    = "LDAP"
        ,status  = "STATUS"
    }
}

----------------------------------------------------------------
-- Implementation details
----------------------------------------------------------------
--- Quiet module loading
local function __requireQuiet(modName)
	if not modName then
		return false
	end

    for _, modPath in ipairs(modPathList) do
        if modPath then
            local modFile = string.gsub(modPath, "?", modName)
            local file    = io.open(modFile)
            if file then
                io.close(file)
                return require(modName)
            end
        end
    end
	
	return false, "WMS User module file not found in controllers directory"
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Update user information with secified extension
-- @param exten		A user extension
-- @param userData	An user data for updating
-- @return boolean TRUE otherwise FALSE followed by error message
function wmsUser:setInfo(exten, userData)
	if not exten or "table" ~= type(userData) then
		return false, "Invalid input arguments"
	end

	local userLdapController, err = __requireQuiet("UserController")
	if not userLdapController then
		return false, err
	end
    local userStatusController, err = __requireQuiet("userStatus")
    if not userStatusController then
        return false, err
    end
	
	-- processing user data
    local pushDataList = {}
    for attr, val in pairs(userData) do
        if attr and val then
            local sName         = self.mAttrs[attr]
            pushDataList[sName] = "table" == type(pushDataList[sName]) and pushDataList[sName] or {}
            if not pushDataList[sName][attr] then
                pushDataList[sName][attr] = val
            end
        end
    end

     -- fix up for surname by linking with name attribute
    if pushDataList.LDAP and pushDataList.LDAP.surname then
        pushDataList.LDAP.name = string.format(string.format("%s %s",
                                    pushDataList.LDAP.name or "", pushDataList.LDAP.surname or ""))
    end

    -- sending data to controller
    for sName, val in pairs(pushDataList) do
        if "table" == type(val) then
            if "LDAP" == sName then
                userLdapController:setAttrByExten(exten, val)
            elseif "STATUS" == sName then
                userStatusController:setDataToJabber(exten, {status = "DND" == val.status and "dnd" or "available"})
            end
        end
    end

    return true
end

--- Retreive WMS user information
-- @param exten		An user extension 
-- @return table with user attributes otherwise FALSE followed by error message
function wmsUser:getInfo(exten)
	if not exten then
		return false
	end
	
	local userLdapController, err = __requireQuiet("UserController")
	if not userLdapController then
		return false, err
	end

    local userStatusController, err = __requireQuiet("userStatus")
    if not userStatusController then
        return false, err
    end

    local rv = userLdapController:getAttrByExten(exten)
    if rv then
        rv.status = userStatusController:getFromCache(exten)
    end
	
	return rv
end

return wmsUser