--- Processing project file of the Bticino Hotel Solution 
-- 
-- @file    xmlprj.lua 
-- @author  Vyacheslav Khudan
-- @created 2014/11/20

module("xmlprj", package.seeall)

----------------------------------------------------------------
-- Private interface
----------------------------------------------------------------

-- Read project file to memory  
local function __xmlfile_read(xml_file)
	if not xml_file or not (xml_file:find("%.hsprj$") or xml_file:find("%.xml$")) then
		return false, string.format("Wront project file path: %s", xml_file or "null")
	end
	
	local cont = false
	local file, err = io.open(xml_file)
	if file then
		cont = file:read("*all")
		file:close()
		return cont
	end
	
	return false, err
end

-- Parse project file via callbacks
local function __xmlfile_parse(xmlfile_cont)
	if not xmlfile_cont then
		return false, "Project file content not found for loading"
	end
	
	local st 		= {}	-- section table	
	local rt 		= {}	-- result table
	local data_link = rt
	
	-- markers for included section like <rooms><room> ... </room></rooms>
	local name_tag	= nil
	
	--- search new name if current one has present
	local function _get_tag_name(name, tags_tbl)
		if not name or "" == name or not tags_tbl or "table" ~= type(tags_tbl) then
			return name
		end
		
		if not tags_tbl[name] then
			return name
		end
		
		local match_name, id = name:match("(.*)_(%d+)$")
		id = (tonumber(id) or 1) + 1
		name = match_name or name
		name = name .. "_" .. id
		return _get_tag_name(name, tags_tbl)
	end
	
	local callbacks = {		-- callback functions
		StartElement  = function(parser, elm_name)
			if name_tag then -- create section for previous element. It does not have value!
				name_tag            = _get_tag_name(name_tag, data_link)
				data_link[name_tag] = {}
				data_link 			= data_link[name_tag]
				table.insert(st, 1, name_tag)	-- add element to sections structure
			end
			name_tag = elm_name
		end;
		
		EndElement 	  = function(parser, elm_name)
			if elm_name then
    			name_tag = nil
    			if elm_name == st[1] or (st[1] or ""):find("^" .. elm_name .. "_%d+$") then	-- close section
    				table.remove(st, 1)	-- remove element from sections structure
    				data_link = rt
    				for i = #st, 1, -1 do	-- reverse serctions table. Find parent link!
    					data_link = data_link[st[i]]
    				end
    			end
    		end
		end;
		
		CharacterData = function(parser, str)
			if str and str:find("%w") and name_tag then	-- skip new line characters. Processing data after start section
				data_link[name_tag] = str	-- it is element with value
				name_tag = nil	-- mark tag as used
			end
		end;
	}
	
	-- start parsing by callbacks
	local p = require"lxp".new(callbacks)
	local rs, err_msg, err_line, err_col, err_pos = p:parse(xmlfile_cont)
	p:parse() 		-- the document is closed (but the parser still has to be closed)
	if not rs then
		local emsg = string.format("XML parsing error: %s (line - %s, col - %s, pos - %s)", err_msg, err_line, err_col, err_pos)
		return false, emsg
	end
	
	p:close()
	collectgarbage()
	
	return rt
end

--- Formatted print project parsed table. Used for debugging
local function __xmlfile_tprint(xmlfile_tbl, custom_name, ident)
	local offset = 4
	if not ident then
		ident = ident and ident or offset
		local custom_name = custom_name or "xml"
		printdbg("%s = {", custom_name)
	end
	
	local rep  = string.rep(" ", ident)
	local table_keys = {}
		
	for key, value in pairs(xmlfile_tbl) do
		if "table" ~= type(value) then
			rep = string.rep(" ", ident)
			printdbg("%s%s = %s", rep, key, value)
		else
			table.insert(table_keys, key)
		end
	end
	
	for _, key in ipairs(table_keys) do
		printdbg("%s%s = {", rep, key)
   		__xmlfile_tprint(xmlfile_tbl[key], nil, ident + offset)
   		printdbg("%s}", rep)
	end
	
	if offset == ident then
		printdbg("}")
	end
end

----------------------------------------------------------------
-- Public interface
----------------------------------------------------------------
--- Processing project file and push data to a table  
function xmlprj_load(fn)
	local text = __xmlfile_read(fn)
	return __xmlfile_parse(text)
end

--- Processing worksatation mapping file and push this data to database
function wsmap_load(fn)
	local text = __xmlfile_read(fn)
	return __xmlfile_parse(text)
end

--- Processing virtual rooms file and save this data to memory
function vrlist_load(fn)
    local text = __xmlfile_read(fn)
    return __xmlfile_parse(text)
end
